/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.axis.example.loanApproval;

import java.net.URL;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPBodyElement;
import javax.xml.soap.SOAPMessage;

import junit.framework.AssertionFailedError;

import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.axis.AxisDeployAction;
import org.ow2.orchestra.axis.test.perf.AxisPerfTestCase;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.SOAPUtil;
import org.w3c.dom.Element;
import org.w3c.dom.Text;

/**
 * @author Charles Souillard
 */
public class LoanApprovalTest extends AxisPerfTestCase {

  protected String getProcessName() {
    return "loanService";
  }
  protected String getProcessNamespace() {
    return "http://orchestra.ow2.org/loanApproval/loanService";
  }
  private String endpoint = getDefaultEndpoint("loanServicePort");
  private String adminServletURL = getAdminServerURL();
  private URL approvalUndeployWsdd = LoanApprovalTest.class.getResource("approval/undeploy.wsdd");
  private URL riskUndeployWsdd = LoanApprovalTest.class.getResource("riskAssessment/undeploy.wsdd");

  public void deploy() {
    URL approvalDeployWsdd = LoanApprovalTest.class.getResource("approval/deploy.wsdd");
    URL riskDeployWsdd = LoanApprovalTest.class.getResource("riskAssessment/deploy.wsdd");

    AxisDeployAction.deployWS(adminServletURL, approvalDeployWsdd);
    AxisDeployAction.deployWS(adminServletURL, riskDeployWsdd);

    deploy(LoanApprovalTest.class.getResource(getProcessName() + ".bpel"),
        LoanApprovalTest.class.getResource(getProcessName() + ".wsdl")
    );

    //specificObjects = new Object[] { endpoint, adminServletURL, approvalUndeployWsdd, riskUndeployWsdd };
  }

  public void undeploy() {
    undeploy(new QName(getProcessNamespace(), getProcessName()));
    AxisDeployAction.deployWS((String) adminServletURL, (URL) approvalUndeployWsdd);
    AxisDeployAction.deployWS((String) adminServletURL, (URL) riskUndeployWsdd);
  }

  public long launch() {
    long endTime = -1;
    int amount = BpelUtil.getRandomInt(0, 60000);
    SOAPMessage response = SOAPUtil.call(getRequest("firstName", "Dupont", Integer.toString(amount)), endpoint);

    endTime = System.currentTimeMillis();
    if (amount >= 30000) {
      assertResult(response, "no");
    } else {
      assertResult(response, "yes");
    }
    return endTime;
  }

  public void testLoanApproval() {
    try {
      deploy();
      // amount < 10000 => assess => risk=low => yes
      assertResult(SOAPUtil.call(getRequest("firstName", "Hash", "9999"),
          endpoint), "yes");
      // amount < 10000 => assess => risk!=low => approve => risk = medium &&
      // (10000<30000) => yes
      assertResult(SOAPUtil.call(getRequest("firstName", "Dupont", "9999"),
          endpoint), "yes");
      // amount < 10000 => assess => risk!=low => approve => risk = high &&
      // (9999<10000) => yes
      assertResult(SOAPUtil.call(getRequest("firstName", "Martin", "9999"),
          endpoint), "yes");
      // amount >= 10000 => approve => amount<50000 => yes
      assertResult(SOAPUtil.call(getRequest("firstName", "Hash", "10000"),
          endpoint), "yes");
      // amount >= 10000 => approve => amount>=50000 => no
      assertResult(SOAPUtil.call(getRequest("firstName", "Hash", "50000"),
          endpoint), "no");
      // amount >= 10000 => approve => risk = medium && amount>=30000 => no
      assertResult(SOAPUtil.call(getRequest("firstName", "Dupont", "30000"),
          endpoint), "no");
      // amount >= 10000 => approve => risk = medium && amount <30000 => yes
      assertResult(SOAPUtil.call(getRequest("firstName", "Dupont", "20000"),
          endpoint), "yes");
      // amount >= 10000 => approve => risk = high && amount>=10000 => no
      assertResult(SOAPUtil.call(getRequest("firstName", "Martin", "10000"),
          endpoint), "no");

      undeploy();
    } catch (Exception e) {
      throw new OrchestraException(e);
    }
  }

  private void assertResult(SOAPMessage response, String expected) {
    if (response != null) {
      SOAPBodyElement operationElement = null;
      try {
        operationElement = getSOAPBodyElement(response);

        assertNotNull(operationElement);
        assertEquals(XmlUtil.toString(operationElement), "request", operationElement.getLocalName());
        List<Element> elements = BpelUtil.getElements(operationElement);
        assertEquals(1, elements.size());
        Element partElement = elements.get(0);
        assertEquals("accept", partElement.getLocalName());
        List<Text> texts = BpelUtil.getTexts(partElement);
        assertEquals(1, texts.size());
        assertEquals(expected, texts.get(0).getData());
      } catch (Exception e) {
        throw new OrchestraException(e);
      } catch (AssertionFailedError e) {
        System.err.println("AssertionFailed, operationElement = "
            + BpelUtil.getNodeAsString(operationElement));
        throw e;
      }
    } else {
      assertTrue("No return available", false);
    }
  }

  private SOAPMessage getRequest(String firstName, String name, String amount) {
    return SOAPUtil.buildDocumentSOAPMessage(
        "http://orchestra.ow2.org/loanApproval/loanService",
        "<request xmlns='http://orchestra.ow2.org/loanApproval/common'>"
            + "<firstName>" + firstName + "</firstName>" + "<name>" + name
            + "</name>" + "<amount>" + amount + "</amount>" + "</request>");
  }

  public String getAlias() {
    return "axis.loan";
  }
}
