/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.axis;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Operation;
import javax.wsdl.Output;
import javax.wsdl.Part;
import javax.wsdl.PortType;
import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.jbpm.env.EnvironmentFactory;
import org.jbpm.tx.Transaction;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.env.EnvFactoryRepository;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.MessageCarrier;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.SOAPUtil;
import org.ow2.orchestra.var.Message;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public final class AxisWSImpl {

  private static Logger log = Logger
      .getLogger(AbstractActivity.class.getName());

  private AxisWSImpl() {
  }

  public static Element[] call(Element[] bodies, QName processQName,
      QName portTypeQName, String operationName, String operationStyle) {

    if (log.isLoggable(Level.FINE)) {
      log.fine("Entering WS Implementation, process = " + processQName
          + ", PT = " + portTypeQName + ", op = " + operationName
          + ", opStyle = " + operationStyle);
      for (Element body : bodies) {
        log.fine("body = " + body);
      }
    }
    Element[] elementsReturned = null;
    Environment environment = null;
    Message incomingMessage = null;
    Operation operation = null;

    OperationKey operationKey = new OperationKey(processQName, portTypeQName,
        operationName);
    try {
      EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
      environment = environmentFactory.openEnvironment();
      Repository repository = environment.get(Repository.class);
      BpelProcess bpelProcess = repository.getProcess(processQName);
      WsdlsInfos wsdlsInfos = bpelProcess.getWsdlInfos();

      operation = getOperation(wsdlsInfos.getPortType(portTypeQName),
          operationName);
      javax.wsdl.Message inputMessage = bpelProcess.getWsdlInfos().getMessage(
          operation.getInput().getMessage().getQName());
      if (operationStyle.equals("document")) {
        incomingMessage = SOAPUtil.buildMessageFromDocumentSOAPBodyElement(
            inputMessage.getParts().values(), bodies[0]);
      } else {
        incomingMessage = SOAPUtil.buildMessageFromRpcSOAPBodyElement(
            inputMessage.getParts().values(), bodies[0]);
      }

    } catch (Exception ex) {
      ex.printStackTrace();
      if (environment != null) {
        Transaction t = environment.getTransaction();
        if (t != null) {
          t.setRollbackOnly();
        }
      }
      throw new OrchestraException(ex);
    } finally {
      if (environment != null) {
        environment.close();
      }
    }

    if (log.isLoggable(Level.FINE)) {
      log.fine("incomingMessage = " + incomingMessage);
    }
    MessageCarrier messageCarrier = null;
    BpelExecution instance = null;

    if (operation.getOutput() != null) {
      // it is a 2 ways operation. We have to get an identifier.
      messageCarrier = new MessageCarrier();
    }
    if (log.isLoggable(Level.FINE)) {
      log.fine("Doing call to receiver, messageCarrier = " + messageCarrier);
    }
    instance = Receiver.handle(incomingMessage, operationKey,
        EnvFactoryRepository.get(), messageCarrier);
    if (log.isLoggable(Level.FINE)) {
      log.fine("call to receiver done. Instance = " + instance);
    }
    if (operation.getOutput() != null) {
      if (log.isLoggable(Level.FINE)) {
        log.fine("This operation has an output, waiting for messageCarrier...");
      }
      Message responseMessage = messageCarrier.getMessage();
      if (operationStyle.equals("document")) {
        if (log.isLoggable(Level.FINE)) {
          log.fine("This operation is using 'document' style.");
        }
        // We make the response");
        List<Part> parts = getOutputParts(operation);
        elementsReturned = new Element[parts.size()];
        int i = 0;
        for (Part part : parts) {
          elementsReturned[i] = responseMessage.getPartValue(part.getName());
          i++;
        }
      } else {
        if (log.isLoggable(Level.FINE)) {
          log.fine("This operation is using 'rpc' style.");
        }
        // it is a rpc style!!
        // We make the response");
        Document rpcDocument = SOAPUtil.buildRpcDocument(responseMessage, operationName);
        elementsReturned = new Element[1];
        elementsReturned[0] = rpcDocument.getDocumentElement();
      }
    }
    if (log.isLoggable(Level.FINE)) {
      log.fine("elementsReturned = " + Arrays.toString(elementsReturned));
    }
    if (instance != null) {
      instance = instance.getProcessInstance();
      boolean removeMutex = false;
      synchronized (MutexRepository.getInstanceMutex(instance)) {
        try {
          environment = EnvFactoryRepository.get().openEnvironment();
          instance = environment.get(Repository.class).getUpdatedExecution(
              instance);
          if (instance != null && instance.isEnded()) {
            environment.get(Repository.class).removeInstance(instance);
            removeMutex = true;
          }
        } catch (Exception ex) {
          ex.printStackTrace();
          if (environment != null) {
            Transaction t = environment.getTransaction();
            if (t != null) {
              t.setRollbackOnly();
            }
          }
          throw new OrchestraException(ex);
        } finally {
          if (environment != null) {
            environment.close();
          }
        }
      }
      if (removeMutex) {
        MutexRepository.removeInstanceMutex(instance.getUuid());
        if (log.isLoggable(Level.FINE)) {
          log.fine("removing instance = " + instance.getUuid());
        }
      }
    }
    return elementsReturned;
  }

  private static Operation getOperation(PortType portType, String operationName) {
    List<Operation> operations = portType.getOperations();
    for (Operation operation : operations) {
      if (operation.getName().equals(operationName)) {
        return operation;
      }
    }
    return null;
  }

  private static List<Part> getOutputParts(final Operation operation) {
    Output output = operation.getOutput();
    if (output != null) {
      return getMessageParts(output.getMessage());
    }
    return new ArrayList<Part>();
  }

  @SuppressWarnings("unchecked")
  private static List<Part> getMessageParts(final javax.wsdl.Message message) {
    Map< ? , Part> partMap = (Map< ? , Part>) message.getParts();
    Collection<Part> parts = partMap.values();
    return new ArrayList<Part>(parts);
  }
}
