/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.axis.test.perf;


import java.util.HashMap;

import javax.xml.soap.SOAPBodyElement;
import javax.xml.soap.SOAPMessage;

import junit.framework.AssertionFailedError;

import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.SOAPUtil;
import org.ow2.orchestra.var.Message;
import org.w3c.dom.Element;
import org.w3c.dom.Text;

/**
 * @author Charles Souillard
 */
public class MarketplaceTest extends AxisPerfTestCase {

  protected String getProcessName() {
    return "marketplace";
  }
  protected String getProcessNamespace() {
    return "http://orchestra.ow2.org/marketplace";
  }

  public void deploy() {
    deploy(
        org.ow2.orchestra.test.perf.marketplace.MarketplaceTest.class.getResource(getProcessName() + ".bpel"),
        org.ow2.orchestra.test.perf.marketplace.MarketplaceTest.class.getResource(getProcessName() + ".wsdl"));
  }

  public long launch() {
    String item = "item" + Thread.currentThread().getName();
    int buyerPrice = BpelUtil.getRandomInt(0, 1000);
    int sellerPrice = BpelUtil.getRandomInt(-45, 1000);
    BuyerThread buyerThread = new BuyerThread(item, buyerPrice);
    SellerThread sellerThread = new SellerThread(item, sellerPrice);
    buyerThread.start();
    sellerThread.start();

    try {
      buyerThread.join();
      sellerThread.join();
    } catch (Exception e) {
      throw new OrchestraException("Exception while joining buyer and seller threads : " + e.getMessage(), e);
    }

    long endTime = System.currentTimeMillis();

    SOAPMessage buyerResponse = buyerThread.getReply();
    SOAPMessage sellerResponse = sellerThread.getReply();

    String buyerOutcome = getOutcome(buyerResponse,"buyer");
    String sellerOutcome = getOutcome(sellerResponse, "seller");

    org.ow2.orchestra.test.perf.marketplace.MarketplaceTest.chekOutcomes(buyerPrice, buyerOutcome, sellerPrice, sellerOutcome);
    return endTime;
  }

  private String getOutcome(SOAPMessage response, String operationName) {
    SOAPBodyElement result = getSOAPBodyElement(response);
    try {
      assertNotNull(result);
      assertEquals(operationName, result.getElementName().getLocalName());
      assertNotNull(result.getFirstChild());
      assertTrue(result.getFirstChild() instanceof Element);
      Element st = (Element) result.getFirstChild();
      assertEquals("outcome", st.getLocalName());
      assertNotNull(st.getFirstChild());
      assertTrue(st.getFirstChild() instanceof Text);
      Text text = (Text) st.getFirstChild();
      return text.getData();
    } catch (Exception e) {
      throw new OrchestraException("Exception while getting outcome", e);
    } catch (AssertionFailedError e) {
      System.err.println("result = " + BpelUtil.getNodeAsString(result));
      throw e;
    }
  }

  public abstract class MPThread extends java.lang.Thread {
    protected String item;
    protected int price;
    protected SOAPMessage response;

    public MPThread(String item, int price) {
      super();
      this.item = item;
      this.price = price;
    }

    public SOAPMessage getReply() {
      return this.response;
    }
    public abstract HashMap<String, Element> getParts();
    public abstract String getOperationName();
    public abstract String getSOAPAction();
    public abstract String getPort();
    public void run() {
      try {
        Message message = new Message(getParts());
        SOAPMessage request = SOAPUtil.buildRpcSOAPMessage(getSOAPAction(), message, getOperationName());
        String endpoint = getDefaultEndpoint(getPort());
        this.response = SOAPUtil.call(request, endpoint);
        if (response != null) {
          SOAPBodyElement result = (SOAPBodyElement) response.getSOAPBody().getChildElements().next();
          assertNotNull(result);
          assertFalse(result.toString(), "Fault".equals(result.getElementName().getLocalName()));
        } else {
          assertTrue("No return available", false);
        }
      } catch (Exception e) {
        e.printStackTrace();
      }

    }
  }

  public class BuyerThread extends MPThread {
    public BuyerThread(String item, int price) {
      super(item, price);
    }
    public HashMap<String, Element> getParts() {
      HashMap<String, Element> parts = new HashMap<String, Element>();
      parts.put("item", BpelXmlUtil.createElementWithContent(this.item));
      parts.put("offer", BpelXmlUtil.createElementWithContent(Integer.toString(this.price)));
      return parts;
    }
    public String getOperationName() {
      return "buyer";
    }
    public String getPort() {
      return "buyerPort";
    }
    public String getSOAPAction() {
      return "http://orchestra.ow2.org/marketplace/buyer";
    }
  }

  public class SellerThread extends MPThread {
    public SellerThread(String item, int price) {
      super(item, price);
    }
    public HashMap<String, Element> getParts() {
      HashMap<String, Element> parts = new HashMap<String, Element>();
      parts.put("inventoryItem", BpelXmlUtil.createElementWithContent(this.item));
      parts.put("askingPrice", BpelXmlUtil.createElementWithContent(Integer.toString(this.price)));
      return parts;
    }
    public String getOperationName() {
      return "seller";
    }
    public String getPort() {
      return "sellerPort";
    }
    public String getSOAPAction() {
      return "http://orchestra.ow2.org/marketplace/seller";
    }
  }

  public void testMarketplace() {
    deploy();
    launch();
    undeploy();
  }

  public String getAlias() {
    return "axis.market";
  }
}


