/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.ExternalActivity;
import org.jbpm.pvm.SignalDefinition;
import org.ow2.orchestra.definition.element.Catch;
import org.ow2.orchestra.definition.element.Link;
import org.ow2.orchestra.definition.element.Source;
import org.ow2.orchestra.definition.element.Target;
import org.ow2.orchestra.exception.BpelFaultException;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.exception.faults.JoinFailure;
import org.ow2.orchestra.lang.JoinCondition;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.FlowRuntime;
import org.ow2.orchestra.runtime.LinkRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.var.Message;

/**
 * common parts of all the bpel activities.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public abstract class AbstractActivity implements ExternalActivity {

  /** log. */
  private static Logger log = Logger
  .getLogger(AbstractActivity.class.getName());

  /** For persistence */
  protected long dbid;

  /** enclosing scope. */
  protected Scope enclosingScope;

  /** name attribute present for all activities. */
  protected String name = null;

  /** suppressJoinFailure attribute present for all activities. */
  protected boolean suppressJoinFailure;

  /**
   * targetsJoinCondition attribute set if this activity has target(s) and the
   * joinCondition attribute is set on targets element.
   */
  protected JoinCondition targetsJoinCondition = null;

  /** list of all sources present for all activities. */
  protected List<Source> sources = null;

  /** list of all targets present for all activities. */
  protected List<Target> targets = null;

  protected String uuid;


  public long getDbid() {
    return dbid;
  }

  protected AbstractActivity() {
    super();
  }

  protected AbstractActivity(String uuid) {
    this.uuid = uuid;
  }

  public enum TargetBehaviour {
    /**
     * all targets are evaluated and transitionCondition is true
     */
    execute,
    /**
     * All targets are not yet evaluated
     */
    unresolved,
    /**
     * all targets are evaluated and transitionCondition is false
     */
    notexecute
  }

  /**
   * Execute method called by process virtual machine.
   *
   * @param execution
   *            current execution
   */
  public void execute(Execution execution) {
    BpelExecution bpelExecution = (BpelExecution) execution;
    TargetBehaviour targetBehaviour = targetsOK(bpelExecution);
    if (targetBehaviour == TargetBehaviour.execute) {
      if (log.isLoggable(Level.FINE)) {
        log.fine("### Executing " + execution.getNode().getName()
            + " in execution " + execution);
      }
      try {
        executeActivity(bpelExecution);
      } catch (Exception e) {
        log.finest("ERROR CATCHED AbstractActivity: " + this);
        boolean find = false;

        // trace the corresponding fault Handler
        Iterator<ScopeRuntime> it = bpelExecution.getScopeRuntimeIterator();
        ScopeRuntime scopeRuntime = null;

        myWhile:while (it.hasNext()) {
          scopeRuntime = it.next();

          // get the fault handler of this scope
          List<Catch> catchs = scopeRuntime.getScopeDefinition().getCatchs();


          if (catchs != null) {
            Message faultMsg = null;
            for (Catch catch1 : catchs) {
              // match the fault
              if ((faultMsg = catch1.match(bpelExecution, e)) != null) {
                // place the execution on the good scope
                ScopeRuntime currentScope = bpelExecution.getScopeRuntime();
                while (scopeRuntime != currentScope) {
                  currentScope = bpelExecution.popScopeRuntime();
                }

                if (currentScope != bpelExecution.getScopeRuntime()) {
                  bpelExecution.pushScopeRuntime(currentScope);
                }
                log.finest("catch matched = " + catch1);
                if (log.isLoggable(Level.FINE)) {
                  log.fine("Execution of catch: " + this);
                }

                if (catch1.getFaultVariable() != null) {
                  VariableRuntime variableRuntime = currentScope.createVariable(
                      bpelExecution, catch1.getFaultVariable());
                  variableRuntime.setValue(faultMsg);
                }
                execution.execute(catch1.getActivity());
                find = true;
                break myWhile;
              }
            }

          }
        }
        if (!find) {
          if (e instanceof BpelFaultException) {
            throw (BpelFaultException)e;
          }
          throw new OrchestraException(e);
        }
      }
    } else if (targetBehaviour == TargetBehaviour.notexecute) {
      //execution.proceed();
    } else {
      execution.waitForSignal();
    }
  }

  /**
   * Signal method.
   *
   * @param execution
   *            execution
   * @param signal
   *            signal
   */
  public void signal(Execution execution, String signal, Map<String, Object> parameters) {
    if (signal == null || !signal.equals("target")) {
      signal((BpelExecution) execution, signal, parameters);
    }
  }

  /**
   * Signal method.
   *
   * @param execution
   *            execution
   * @param signal
   *            signal
   */
  public void signal(BpelExecution execution, String signal, Map<String, Object> parameters) {
    throw new OrchestraException("Signal not supported.");
  }

  public Set<SignalDefinition> getSignals(Execution execution) throws Exception {
    return null;
  }

  /**
   * Returns true if all target are evaluated and if joinCondition is true.
   *
   * @param execution
   *            execution
   * @return true if all target are evaluated and if joinCondition is true.
   */
  private TargetBehaviour targetsOK(BpelExecution execution) {
    // verify that all links have been evaluated
    if (targets == null || targets.isEmpty()) {
      return TargetBehaviour.execute;
    }
    boolean tjcLogicalOR = false;
    for (Target target : this.targets) {
      synchronized (MutexRepository.getInstanceMutex(execution)) {
        String linkName = target.getLinkName();
        FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
        Boolean linkStatus = fr.getLinkStatus(linkName);
        if (linkStatus == null) {
          // a link is not already evaluated
          if (log.isLoggable(Level.FINE)) {
            log.fine(this.getClass() + "." + getName()
                + " targetsOK, waiting for link : " + linkName);
          }
          // add this execution as a waiting execution
          fr.addWaitingExecution(linkName, execution);
          return TargetBehaviour.unresolved;
        } else {
          tjcLogicalOR = tjcLogicalOR || linkStatus;
        }
      }
    }
    boolean tjc = tjcLogicalOR;
    if (targetsJoinCondition != null) {
      try {
        tjc = targetsJoinCondition.getEvaluator().evaluateBoolean(execution);
      } catch (Exception e) {
        // TODO throw joinFailure
        e.printStackTrace();
        setAllUnsetLinks(execution);
        throw new JoinFailure("bpel:joinFailure : "
            + targetsJoinCondition.getText()
            + ", exception while evaluating the targetJoinCondition");
      }
    }

    if (!tjc) {
      if (log.isLoggable(Level.FINE)) {
        log.fine(this.getClass() + ".targetsOK : " + false);
      }
      if (!suppressJoinFailure) {
        if (targetsJoinCondition != null) {
          throw new JoinFailure("bpel:joinFailure : "
              + targetsJoinCondition.getText()
              + ", targetJoinCondition is false and suppressJoinFailure is no");
        } else {
          throw new JoinFailure("bpel:joinFailure : targetJoinCondition is false and suppressJoinFailure is no");
        }
      } else {
        if (this.sources != null) {
          for (Source source : this.sources) {
            synchronized (MutexRepository.getInstanceMutex(execution)) {
              String linkName = source.getLinkName();
              FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
              LinkRuntime lr = fr.getLink(linkName);
              if (lr.getStatus() == null) {
                lr.setStatus(false);
              }
              BpelExecution execToWakeUp = fr.getWaitingExecution(linkName);
              if (execToWakeUp != null) {
                execToWakeUp.execute(execToWakeUp.getNode());
              }
            }
          }
        }
        return TargetBehaviour.notexecute;
      }
      // TODO
      // set outgoingLinks (sources) status to false according to links semantic
      // (in same scope or not...)
      // see transitionCondition
      // notify waitingExecutions
      // if suppressjoinFailure : nothing (do not execute activity)
      // if not suppressJoinFailure : throw bpel:joinfFailure
    }
    if (log.isLoggable(Level.FINE)) {
      log.fine(this.getClass() + "." + getName() + " targetsOK : " + true);
    }
    return TargetBehaviour.execute;
  }

  /**
   * Executes sources of this activity : called when this activity is finished.
   *
   * @param execution
   *            execution
   */
  public void executeSources(BpelExecution execution) {
    // Sources are executed once the enclosing activity is completed.
    // In the special case of an enclosing scope, this is not required to have
    // a successful completion. In that case (enclosing scope abnormally
    // completed)
    // sources must be executed only if there is a faultHandler associated to
    // the
    // enclosing scope and if this faultHandler has terminated without throwing
    // a fault.

    // if an error occurs during the TC evaluation
    // This error does not prevent enclosing activity to be set at "completed".
    // This error must be managed by the enclosing scope.
    // If the target activity is inside the source activity enclosing scope,
    // UNSET must be set on the concerned link status. (nothing to do)
    // -> done
    // If the target activity is outside the source activity enclosing scope,
    // false must be set on the concerned link status.
    // -> done
    // If source activity is a scope, the error must be propagated to its
    // father.
    // All remaining outgoing links (sources) having target activity in source
    // activity enclosing scope
    // must not have their transitionCondition evaluated (nothing to do)
    // -> done
    // All remaining outgoing links (sources) having target activity outside
    // source activity enclosing scope
    // must not have their status set to false
    // -> done

    if (log.isLoggable(Level.FINE)) {
      log.fine("Executing sources of activity : " + this.getClass().getName()
          + "." + name);
    }
    if (sources == null) {
      return;
    }
    for (Source source : this.sources) {
      String linkName = source.getLinkName();
      boolean tc = true;
      if (source.getTransitionCondition() != null) {
        try {
          tc = source.getTransitionCondition().getEvaluator()
          .evaluateBoolean(execution);
        } catch (Exception e) {
          // TODO : throw error to enclosing scope : warning if the source
          // activity is a scope, propagate to parent
          setAllUnsetLinks(execution);
          e.printStackTrace();
        }
      }
      synchronized (MutexRepository.getInstanceMutex(execution)) {
        execution.setLinkStatus(linkName, tc);
        if (log.isLoggable(Level.FINE)) {
          log.fine("executeSources, notifying executions of the link : "
              + linkName + "(" + tc + ")");
        }
        FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
        BpelExecution execToWakeUp = fr.getWaitingExecution(linkName);
        if (execToWakeUp != null) {
          execToWakeUp.execute(execToWakeUp.getNode());
          // execToWakeUp.execute(execToWakeUp.getNode(), "target");
        }
      }
    }
  }

  /**
   * For every LinkRuntime having status UNSET, set it to false if target
   * activity is not defined in source activity enclosing scope.
   *
   * @param execution
   *            execution
   */
  private void setAllUnsetLinks(BpelExecution execution) {
    if (this.sources != null) {
      for (Source source : this.sources) {
        String linkName = source.getLinkName();
        FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
        LinkRuntime lr = fr.getLink(linkName);
        if (lr.getStatus() == null) {
          Link link = source.getLink();
          Scope targetActivityEnclosingScope = link.getTargetActivity().getEnclosingScope();
          Scope sourceActivityEnclosingScope = link.getSourceActivity().getEnclosingScope();
          if (!(targetActivityEnclosingScope == sourceActivityEnclosingScope)) {
            lr.setStatus(false);
          }
        }
      }
    }
  }

  /**
   * This method must be implemented by every activity. This is the business
   * behaviour of the activity.
   *
   * @param execution
   *            execution
   */
  protected abstract void executeActivity(BpelExecution execution);

  public Scope getEnclosingScope() {
    return enclosingScope;
  }

  public void setEnclosingScope(Scope enclosingScope) {
    this.enclosingScope = enclosingScope;
  }

  public String getName() {
    return this.name;
  }

  public void setName(String name) {
    this.name = name;
  }

  public boolean isSuppressJoinFailure() {
    return this.suppressJoinFailure;
  }

  public void setSuppressJoinFailure(boolean suppressJoinFailure) {
    this.suppressJoinFailure = suppressJoinFailure;
  }

  public List<Source> getSources() {
    return this.sources;
  }

  public void setSources(List<Source> sources) {
    this.sources = sources;
  }

  public List<Target> getTargets() {
    return this.targets;
  }

  public void setTargets(List<Target> targets) {
    this.targets = targets;
  }

  public JoinCondition getTargetsJoinCondition() {
    return this.targetsJoinCondition;
  }

  public void setTargetsJoinCondition(JoinCondition targetsJoinCondition) {
    this.targetsJoinCondition = targetsJoinCondition;
  }

  public String getUuid() {
    return uuid;
  }

  public int hashCode() {
    return uuid.hashCode();
  }

  public boolean equals(Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof AbstractActivity)) {
      return false;
    }
    AbstractActivity other = (AbstractActivity) obj;
    return other.getUuid().equals(uuid);
  }
}
