/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.CorrelationSetRuntime;
import org.ow2.orchestra.runtime.PartnerLinkRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.itf.Invoker;
import org.ow2.orchestra.var.Message;

/** the runtime behaviour of the invoke activity.
 * @author Charles Souillard, Tom Baeyens
 */
public class Invoke extends AutomaticActivity {
  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;
  private static Logger log = Logger.getLogger(Invoke.class.getName());


  /**
   * <xsd:element ref="catch" minOccurs="0" maxOccurs="unbounded"/>
   * <xsd:element ref="catchAll" minOccurs="0"/>
   * <xsd:element ref="compensationHandler" minOccurs="0"/>
   */
  /** partnerLink attribute of this invoke element. */
  protected String partnerLink;
  /** portType attribute of this invoke element. */
  protected QName portTypeQName;
  /** operation attribute of this invoke element. */
  protected String operationName;
  /** inputVariable attribute of this invoke element. */
  protected String inputVariable;
  /** outputVariable attribute of this invoke element. */
  protected String outputVariable;

  /**
   * list of correlations having pattern "request" or "request-response"
   * of this invoke element.
   */
  protected List<Correlation> requestCorrelations = null;
  /**
   * list of correlations having pattern "response" or "request-response"
   * of this invoke element.
   */
  protected List<Correlation> responseCorrelations = null;

  protected boolean isOneWay;
  protected List<Copy> outCopies;
  protected List<Copy> inCopies;
  protected Variable anonymousOutboundMessageVariable;
  protected Variable anonymousInboundMessageVariable;
  public static final String OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME = "ORCHESTRA_4_OUTBOUND_WSDL_MESSAGE";
  public static final String INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME = "ORCHESTRA_4_INBOUND_WSDL_MESSAGE";

  protected Invoke() {
    super();
  }
  public Invoke(String uuid) {
    super(uuid);
  }

  /**
   * business logic.
   * @param execution execution
   */
  public void executeBusinessLogic(BpelExecution execution) {
    Invoker serviceInvoker = (Invoker) Environment.getCurrent().get(Invoker.class);
    PartnerLinkRuntime partnerLinkRuntime = execution.getPartnerLinkRuntime(this.partnerLink);
    OperationKey operationKey = new OperationKey(
        execution.getProcessDefinition().getQName(),
        portTypeQName,
        operationName);

    ScopeRuntime scopeRuntime = execution.getScopeRuntime();
    VariableRuntime anonymousOutboundMessageVariableRuntime =
      scopeRuntime.createVariable(execution, this.anonymousOutboundMessageVariable);

    // Perform copy operations
    for (Copy copy : this.outCopies) {
      copy.execute(execution);
    }

    if (requestCorrelations != null) {
      //we verify correlation
      for (Correlation correlation : requestCorrelations) {
        // Check correlation on outgoing message
        CorrelationSetRuntime correlationSetRuntime =
          execution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
        correlationSetRuntime.checkCSValues(
            execution,
            (Message) anonymousOutboundMessageVariableRuntime.getValue(),
            anonymousOutboundMessageVariable.getMessageType(),
            correlation.getInitiate());
      }
    }
    scopeRuntime.removeVariable(execution, this.anonymousOutboundMessageVariable);

    if (log.isLoggable(Level.FINE)) {
      log.fine("Invoke, inputMessage " + anonymousOutboundMessageVariableRuntime.getValue());
    }
    Message returnedMessage =
      serviceInvoker.invoke(operationKey, this.name, partnerLinkRuntime, anonymousOutboundMessageVariableRuntime);

    if (log.isLoggable(Level.FINE)) {
      log.fine("Invoke, returnedMessage " + returnedMessage);
    }

    if (!this.isOneWay) {
      // Receive response message before correlation validation
      VariableRuntime anonymousInboundMessageVariableRuntime =
        scopeRuntime.createVariable(execution, this.anonymousInboundMessageVariable);
      anonymousInboundMessageVariableRuntime.setValue(returnedMessage);

      for (Copy copy : this.inCopies) {
        copy.execute(execution);
      }

      scopeRuntime.removeVariable(execution, this.anonymousInboundMessageVariable);

      // validate correlations
      if (responseCorrelations != null) {
        for (Correlation correlation : responseCorrelations) {
          CorrelationSetRuntime correlationSetRuntime =
            execution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
          correlationSetRuntime.checkCSValues(execution, returnedMessage, anonymousInboundMessageVariable.getMessageType(),
              correlation.getInitiate());
        }
      }
    }

  }

  public String getInputVariable() {
    return this.inputVariable;
  }
  public void setInputVariable(String inputVariable) {
    this.inputVariable = inputVariable;
  }
  public String getOperation() {
    return this.operationName;
  }
  public void setOperation(String operation) {
    this.operationName = operation;
  }
  public String getOutputVariable() {
    return this.outputVariable;
  }
  public void setOutputVariable(String outputVariable) {
    this.outputVariable = outputVariable;
  }
  public String getPartnerLink() {
    return this.partnerLink;
  }
  public void setPartnerLink(String partnerLink) {
    this.partnerLink = partnerLink;
  }
  public QName getPortTypeQName() {
    return this.portTypeQName;
  }
  public void setPortTypeQName(QName portType) {
    this.portTypeQName = portType;
  }
  public List<Correlation> getRequestCorrelations() {
    return requestCorrelations;
  }
  public void setRequestCorrelations(List<Correlation> correlations) {
    this.requestCorrelations = correlations;
  }
  public List<Correlation> getResponseCorrelations() {
    return responseCorrelations;
  }
  public void setResponseCorrelations(List<Correlation> correlations) {
    this.responseCorrelations = correlations;
  }
  public boolean isOneWay() {
    return isOneWay;
  }
  public void setOneWay(boolean oneWay) {
    this.isOneWay = oneWay;
  }
  public List<Copy> getOutCopies() {
    return outCopies;
  }
  public void setOutCopies(List<Copy> outCopies) {
    this.outCopies = outCopies;
  }
  public List<Copy> getInCopies() {
    return inCopies;
  }
  public void setInCopies(List<Copy> inCopies) {
    this.inCopies = inCopies;
  }
  public Variable getAnonymousOutboundMessageVariable() {
    return anonymousOutboundMessageVariable;
  }
  public void setAnonymousOutboundMessageVariable(
      Variable anonymousOutboundMessageVariable) {
    this.anonymousOutboundMessageVariable = anonymousOutboundMessageVariable;
  }
  public Variable getAnonymousInboundMessageVariable() {
    return anonymousInboundMessageVariable;
  }
  public void setAnonymousInboundMessageVariable(
      Variable anonymousInboundMessageVariable) {
    this.anonymousInboundMessageVariable = anonymousInboundMessageVariable;
  }
}
