/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.CorrelationSetRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.var.Message;

/**
 * the runtime behaviour of the receive activity.
 *
 * @author Charles Souillard, Tom Baeyens
 * @version $Id: Receive.java 1409 2008-04-08 09:39:42Z porcherg $
 */
public class Receive extends AbstractActivity implements ReceivingElement,
    InboundMessageElement {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /** log. */
  private static Logger log = Logger.getLogger(Receive.class.getName());

  /** partnerLink attribute of this receive element. */
  protected String partnerLinkName = null;

  /** operation attribute of this receive element. */
  protected String operationName = null;

  /** portType attribute of this receive element. */
  protected QName portTypeQName = null;

  /** variable attribute of this receive element. */
  protected String variable = null;

  /** createInstance attribute of this receive element. */
  protected boolean createInstance = false;

  /** messageExchange attribute of this receive element. */
  protected String messageExchange = "";

  /** list of correlations of this receive. */
  protected List<Correlation> correlations = null;

  /** fromParts management. */
  protected List<Copy> inCopies;

  protected Variable anonymousInboundMessageVariable;

  public static final String INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME = "ORCHESTRA_4_INBOUND_WSDL_MESSAGE";

  protected Receive() {
    super();
  }

  public Receive(String uuid) {
    super(uuid);
  }

  /**
   * business logic.
   *
   * @param execution
   *            execution
   */
  public void executeActivity(BpelExecution execution) {
    execution.getProcessInstance().addWaitingExecution(execution);

    if (log.isLoggable(Level.FINE)) {
      OperationKey operationKey = new OperationKey(execution.getProcessDefinition()
          .getQName(), portTypeQName, operationName);
      log.fine("Receive " + operationKey + " : waiting...");
    }
    execution.waitForSignal();
  }

  @Override
  public void signal(BpelExecution execution, String signal, Map<String, Object> parameters) {
    if (!execution.hasPendingMessage()) {
      throw new OrchestraException("Receive activity " + this
          + " signalled but no message available.");
    }
    execution.setHasPendingMessage(false);
    executeSources(execution);
  }

  /**
   * This message is to be called from the receive implementation (above) in
   * case the message arrived in the engine before the execution arrived in the
   * receive. in case the message arrives later, the receive will be invoked by
   * the correlator.
   */
  public void receive(Message message, BpelExecution execution) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Receiving message : " + message);
    }
    // initiate correlationSets
    initiateCs(execution, message);
    if (variable != null) {
      // we store the variable
      VariableRuntime variableRuntime = execution.getVariableByName(variable);
      variableRuntime.setValue(message);
    } else if (inCopies != null && !inCopies.isEmpty()) {
      ScopeRuntime scopeRuntime = execution.getScopeRuntime();
      VariableRuntime anonymousOutboundMessageVariableRuntime = scopeRuntime
          .createVariable(execution, this.anonymousInboundMessageVariable);

      anonymousOutboundMessageVariableRuntime.setValue(message);
      for (Copy copy : this.inCopies) {
        copy.execute(execution);
      }
      scopeRuntime.removeVariable(execution,
          this.anonymousInboundMessageVariable);
    }
    /*
     *
     * QName messageType =
     * execution.getPortType(getPortTypeQName()).getOperation(getOperation(),
     * null, null) .getInput().getMessage().getQName(); if (correlations !=
     * null) { //manage correlations for (Correlation correlation :
     * correlations) { CorrelationSetRuntime correlationSetRuntime =
     * execution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
     * correlationSetRuntime.checkValues(execution, message, messageType,
     * correlation.getInitiate()); } }
     */
  }

  private void initiateCs(BpelExecution bpelExecution, Message incomingMessage) {
    if (correlations != null) {
      synchronized (MutexRepository.getInstanceMutex(bpelExecution)) {
        // we calcul the values for each correlationSet
        for (Correlation correlation : correlations) {
          CorrelationSetRuntime correlationSetRuntime =
            bpelExecution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
          QName messageType = bpelExecution.getPortType(
              this.getPortTypeQName()).getOperation(this.getOperation(),null, null)
              .getInput().getMessage().getQName();
          correlationSetRuntime.checkCSValues(bpelExecution, incomingMessage, messageType,
              correlation.getInitiate());
        }
      }
    }
  }

  public String getPartnerLinkName() {
    return partnerLinkName;
  }

  public String getOperation() {
    return operationName;
  }

  public QName getPortTypeQName() {
    return this.portTypeQName;
  }

  public void setPortTypeQName(QName portType) {
    this.portTypeQName = portType;
  }

  public String getVariable() {
    return this.variable;
  }

  public void setVariable(String variable) {
    this.variable = variable;
  }

  public void setOperation(String operation) {
    this.operationName = operation;
  }

  public void setPartnerLinkName(String partnerLinkName) {
    this.partnerLinkName = partnerLinkName;
  }

  public List<Correlation> getCorrelations() {
    return this.correlations;
  }

  public void setCorrelations(List<Correlation> correlations) {
    this.correlations = correlations;
  }

  public boolean isCreateInstance() {
    return this.createInstance;
  }

  public void setCreateInstance(boolean createInstance) {
    this.createInstance = createInstance;
  }

  public String getMessageExchange() {
    return this.messageExchange;
  }

  public void setMessageExchange(String messageExchange) {
    this.messageExchange = messageExchange;
  }

  public Variable getAnonymousInboundMessageVariable() {
    return anonymousInboundMessageVariable;
  }

  public void setAnonymousInboundMessageVariable(
      Variable anonymousInboundMessageVariable) {
    this.anonymousInboundMessageVariable = anonymousInboundMessageVariable;
  }

  public List<Copy> getInCopies() {
    return inCopies;
  }

  public void setInCopies(List<Copy> inCopies) {
    this.inCopies = inCopies;
  }

  public OperationKey getOperationKey(QName processQName) {
    return new OperationKey(processQName, portTypeQName, operationName);
  }

  public List<ReceivingElement> getReceivingElements() {
    List<ReceivingElement> list = new ArrayList<ReceivingElement>();
    list.add(this);
    return list;
  }
}
