/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.CorrelationSetRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.runtime.BpelExecution.ScopeRuntimeIterator;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReplierKey;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.var.Message;

/**
 * the runtime behaviour of the reply activity.
 *
 * @author Charles Souillard, Tom Baeyens
 * @version $Id: Reply.java 1409 2008-04-08 09:39:42Z porcherg $
 */
public class Reply extends AutomaticActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /** partnerLink attribute of this Reply. */
  protected String partnerLinkName = null;

  /** operation attribute of this Reply. */
  protected String operationName = null;

  /** portType attribute of this Reply. */
  protected QName portTypeQName = null;

  /** variable attribute of this Reply. */
  protected String variable = null;

  /** faultName attribute of this Reply. */
  protected QName faultQName = null;

  /** messageExchange attribute of this Reply. */
  protected String messageExchange = "";

  /** list of correlations of this Reply. */
  protected List<Correlation> correlations = null;

  protected List<Copy> outCopies;

  protected Variable anonymousOutboundMessageVariable;

  public static final String OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME = "ORCHESTRA_4_OUTBOUND_WSDL_MESSAGE";

  protected Reply() {
    super();
  }

  public Reply(String uuid) {
    super(uuid);
  }

  /**
   * business logic.
   *
   * @param execution
   *            execution
   */
  public void executeBusinessLogic(BpelExecution execution) {
    // first find the replier
    OperationKey operationKey = new OperationKey(execution.getProcessDefinition()
        .getQName(), portTypeQName, operationName);
    ScopeRuntime scopeRuntime = execution.getScopeRuntime();
    Replier replier = null;

    Iterator<ScopeRuntime> it = new ScopeRuntimeIterator(execution);
    ScopeRuntime sr = null;
    while (replier == null && it.hasNext()) {
      sr = it.next();
      ReplierKey replierKey = new ReplierKey(operationKey, messageExchange, sr);
      replier = Environment.getCurrent().get(Repository.class).removeReplier(
          replierKey);
    }

    if (replier == null) {
      throw new OrchestraException("couldn't find a replier for " + this
          + " on operation " + operationKey);
    }

    // get the reply message
    Message reply = null;
    if (variable != null) {
      reply = (Message) execution.getVariableValue(variable);
    } else if (outCopies != null && !outCopies.isEmpty()) {
      scopeRuntime = execution.getScopeRuntime();
      VariableRuntime anonymousOutboundMessageVariableRuntime = scopeRuntime
          .createVariable(execution, this.anonymousOutboundMessageVariable);
      // toparts
      for (Copy copy : this.outCopies) {
        copy.execute(execution);
      }
      reply = (Message) anonymousOutboundMessageVariableRuntime.getValue();
    }
    // check correlations
    if (correlations != null) {
      //we verify correlation
      for (Correlation correlation : correlations) {
        // Check correlation on outgoing message
        CorrelationSetRuntime correlationSetRuntime =
          execution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
        correlationSetRuntime.checkCSValues(
            execution,
            reply,
            anonymousOutboundMessageVariable.getMessageType(),
            correlation.getInitiate());
      }
    }
    // do the reply
    replier.reply(reply);
  }

  public String getOperation() {
    return this.operationName;
  }

  public void setOperation(String operation) {
    this.operationName = operation;
  }

  public String getPartnerLinkName() {
    return this.partnerLinkName;
  }

  public void setPartnerLinkName(String partnerLinkName) {
    this.partnerLinkName = partnerLinkName;
  }

  public QName getPortTypeQName() {
    return this.portTypeQName;
  }

  public void setPortTypeQName(QName portType) {
    this.portTypeQName = portType;
  }

  public String getVariable() {
    return this.variable;
  }

  public void setVariable(String variable) {
    this.variable = variable;
  }

  public List<Correlation> getCorrelations() {
    return this.correlations;
  }

  public void setCorrelations(List<Correlation> correlations) {
    this.correlations = correlations;
  }

  public QName getFaultQName() {
    return this.faultQName;
  }

  public void setFaultQName(QName faultName) {
    this.faultQName = faultName;
  }

  public String getMessageExchange() {
    return this.messageExchange;
  }

  public void setMessageExchange(String messageExchange) {
    this.messageExchange = messageExchange;
  }

  public void setOutCopies(List<Copy> outCopies) {
    this.outCopies = outCopies;
  }

  public void setAnonymousOutboundMessageVariable(
      Variable anonymousOutboundMessageVariable) {
    this.anonymousOutboundMessageVariable = anonymousOutboundMessageVariable;
  }
}
