/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.Node;
import org.jbpm.pvm.impl.ObservableElementImpl;
import org.ow2.orchestra.definition.element.Catch;
import org.ow2.orchestra.definition.element.CompensationHandler;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.EventHandler;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ScopeRuntime;

/**
 * the runtime behaviour of the scope activity.
 *
 * @author Charles Souillard, Tom Baeyens
 * @version $Id: Scope.java 1437 2008-04-15 13:29:33Z porcherg $
 */
public class Scope extends AbstractActivity {
  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(Scope.class.getName());

  /**
   * List of variable name that must be initialized at the beginning of the scope.
   * (variable with an inline from-spec).
   */
  protected List<String> listOfVariableToBeInitialized;
  /** isolated attribute of this scope. */
  protected boolean isolated = false;
  /** exitOnStandardFault attribute of this scope. */
  protected boolean exitOnStandardFault;
  /** list of variables of this scope. */
  protected Map<String, Variable> variables;
  /** list of eventHandlers of this scope. */
  protected List<EventHandler> eventHandlers;
  /** list of faultHandlers of this scope. */
  protected CompensationHandler compensationHandler;
  /** list of partnerLinks of this scope. */
  protected Map<String, PartnerLink> partnerLinks;
  /** list of messageExchanges of this scope. */
  protected Set<String> messageExchanges;
  /** list of correlationSets of this scope. */
  protected Map<String, CorrelationSet> correlationSets;

  protected List<Catch> catchs;
  /*
    <xsd:element ref="compensationHandler" minOccurs="0"/>
    <xsd:element ref="terminationHandler" minOccurs="0"/>
   */

  protected Scope() {
    super();
  }
  public Scope(String uuid) {
    super(uuid);
    this.setName("Scope " + uuid);
  }

  /**
   * business logic.
   * @param execution execution
   */
  @Override
  public void executeActivity(BpelExecution execution) {
    ScopeRuntime sr = new ScopeRuntime(this, execution, UUID.randomUUID().toString());
    execution.pushScopeRuntime(sr);
    if (listOfVariableToBeInitialized != null) {
      for (String variableName : listOfVariableToBeInitialized) {
        execution.getVariableByName(variableName).initializeXmlType(execution);
      }
    }

    // Add these compensations handlers to the parent scope
    execution.getScopeRuntime().addCompensationHandlerInCurrentScope(this.getCompensationHandler());

    // get the current sequence node
    Node node = execution.getNode();
    Node firstNode = node.getNodes().get(0);
    execution.execute(firstNode);
  }

  /**
   * Signal method.
   * @param execution execution
   * @param signal signal
   */
  @Override
  public void signal(BpelExecution execution, String signal, Map<String, Object> parameters) {
    executeSources(execution);
    ObservableElementImpl parent = execution.getNode().getParent();
    if (parent != null && parent instanceof Node) {

      List<CompensationHandler> chsChilds = null;
      // we're done and the sequence is completed
      ScopeRuntime scopeRuntime = execution.popScopeRuntime(); 
      if (scopeRuntime != null) {
        chsChilds = scopeRuntime.getCompensationHandlers();
      }

      scopeRuntime = execution.getScopeRuntime();
      // Add these compensations handlers to the parent scope
      if ((scopeRuntime != null) && (chsChilds != null)) {
        scopeRuntime.addCompensationHandlersInCurrentScope(chsChilds);
      }
    } else {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("top level scope ends the bpel process execution");
      }
      execution.end();
    }
  }

  /**
   * Searched for a correlation having the given name.
   * Search first in this scope and if nothing is found, search in father scopes...
   * @param correlationSetName correlationSetName
   * @return the found correlation
   */
  public CorrelationSet findCorrelationSet(String correlationSetName) {
    if (correlationSets != null && correlationSets.containsKey(correlationSetName)) {
      return correlationSets.get(correlationSetName);
    } else if (enclosingScope != null) {
      return enclosingScope.findCorrelationSet(correlationSetName);
    }
    //TODO : throw an exception
    return null;
  }
  /**
   * Searched for a partnerLink having the given name.
   * Search first in this scope and if nothing is found, search in father scopes...
   * @param partnerLinkName partnerLinkName
   * @return the found partnerLink
   */
  public PartnerLink findPartnerLink(String partnerLinkName) {
    if (partnerLinks != null && partnerLinks.containsKey(partnerLinkName)) {
      return partnerLinks.get(partnerLinkName);
    } else if (enclosingScope != null) {
      return enclosingScope.findPartnerLink(partnerLinkName);
    }
    //TODO : throw an exception
    return null;
  }
  /**
   * Searched for a variable having the given name.
   * Search first in this scope and if nothing is found, search in father scopes...
   * @param variableName variableName
   * @return the found variable
   */
  public Variable findVariable(String variableName) {
    if (variables != null && variables.containsKey(variableName)) {
      return variables.get(variableName);
    } else if (enclosingScope != null) {
      return enclosingScope.findVariable(variableName);
    }
    //TODO : throw an exception
    return null;
  }
  /**
   * Add a variable in the list of variables that the scope needs to initialize.
   * @param variableName variableName
   */
  public void addVariableToBeInitialized(String variableName) {
    if (this.listOfVariableToBeInitialized == null) {
      this.listOfVariableToBeInitialized = new ArrayList<String>();
    }
    this.listOfVariableToBeInitialized.add(variableName);
  }

  public Variable getVariable(String variableName) {
    return this.variables.get(variableName);
  }
  public Map<String, PartnerLink> getPartnerLinks() {
    return partnerLinks;
  }
  public Map<String, Variable> getVariables() {
    return this.variables;
  }
  public List<EventHandler> getEventHandlers() {
    return this.eventHandlers;
  }
  public void setEventHandlers(List<EventHandler> eventHandlers) {
    this.eventHandlers = eventHandlers;
  }
  public boolean isExitOnStandardFault() {
    return this.exitOnStandardFault;
  }
  public void setExitOnStandardFault(boolean exitOnStandardFault) {
    this.exitOnStandardFault = exitOnStandardFault;
  }
  public boolean isIsolated() {
    return this.isolated;
  }
  public void setIsolated(boolean isolated) {
    this.isolated = isolated;
  }
  public void setPartnerLinks(Map<String, PartnerLink> partnerLinks) {
    this.partnerLinks = partnerLinks;
  }
  public void setVariables(Map<String, Variable> variables) {
    this.variables = variables;
  }
  public Set<String> getMessageExchanges() {
    return this.messageExchanges;
  }
  public void setMessageExchanges(Set<String> messageExchanges) {
    this.messageExchanges = messageExchanges;
  }
  public Map<String, CorrelationSet> getCorrelationSets() {
    return this.correlationSets;
  }
  public void setCorrelationSets(Map<String, CorrelationSet> correlationSets) {
    this.correlationSets = correlationSets;
  }

  public CompensationHandler getCompensationHandler() {
    return compensationHandler;
  }
  public void setCompensationHandler(CompensationHandler compensationHandler) {
    this.compensationHandler = compensationHandler;
  }



  @Override
  public String toString() {
    return Scope.class.getName()
        + "[listOfVariableToBeInitialized: " + listOfVariableToBeInitialized
        + ", isolated: " + isolated
        + ", exitOnStandardFault: " + exitOnStandardFault
        + ", variables: " + variables
        + ", eventHandlers: " + eventHandlers
        + ", catchs: " + catchs
        + ", compensationHandlers: " + compensationHandler
        + ", partnerLinks: " + partnerLinks
        + ", messageExchanges: " + messageExchanges
        + ", correlationSets: " + correlationSets
        + "]";
  }

  public List<Catch> getCatchs() {
    return catchs;
  }

  public void setCatchs(List<Catch> catchs) {
    this.catchs = catchs;
  }

}
