/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.deployment;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.jbpm.xml.Parse;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.parsing.BpelParser;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.Namespaces;
import org.ow2.orchestra.util.StyleSheetRepository;
import org.ow2.orchestra.wsdl.WsdlDefinitions;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.xml.sax.InputSource;

/**
 * takes {@link Deployment}s as input, parses the contents into a
 * {@link BpelProcess}, stores the {@link BpelProcess} into the
 * {@link Repository} and publishes the services with a {@link Publisher}.
 *
 * <p>
 * processRepository and servicePublisher have to be injected by the user (or an
 * IoC framework) before the deploy method is invoked.
 * </p>
 *
 * <p>
 * This class is thread safe and can be shared between multiple threads.
 * </p>
 *
 * @author Tom Baeyens, Goulven Le Jeune
 */
public class Deployer {

  private static Logger log = Logger.getLogger(Deployer.class.getName());

  /** parser to use to parse wsdl files. */
  //private BpelWsdlParser bpelWsdlParser = null;

  /** parser to use to parse bpel files. */
  private BpelParser bpelParser = null;

  /**
   * Default constructor.
   */
  public Deployer() {
    //bpelWsdlParser = new BpelWsdlParser();
    bpelParser = new BpelParser();
  }

  /**
   * Thanks to the deployment object given in parameter, parse all wsdl files,
   * bpel file. Then store the process in processRepository and publish bpel
   * process services.
   *
   * @param bpelDeployment
   *            populated bpelDeployment
   * @param environment
   *            environment
   * @return created BpelProcess
   */
  public BpelProcess deploy(Deployment bpelDeployment, Environment environment) {
    if (log.isLoggable(Level.INFO)) {
      log.info("Deploying BPEL from URL: " + bpelDeployment.getBpelURL());
    }

    Parse parse = bpelParser.createParse();
    
    WsdlDefinitions wsdlDefinitions = bpelDeployment.getWsdlDefinitions();
    parse.pushObject(wsdlDefinitions);

    Namespaces namespaces = new Namespaces(parse);
    parse.pushObject(namespaces);
    parse.pushObject(environment);

    BpelProcess bpelProcess = new BpelProcess(bpelDeployment.getBpelURL());
    StyleSheetRepository styleSheetRepository = new StyleSheetRepository();
    bpelProcess.setStyleSheetRepository(styleSheetRepository);
    parse.pushObject(bpelProcess);

    WsdlsInfos wsdlInfos = new WsdlsInfos();
    bpelProcess.setWsdlInfos(wsdlInfos);

    // now the bpel process will be parsed.
    InputSource bpel = bpelDeployment.getBpel();
    if (bpel != null) {
      // make the bpel process available to the bpel parser by pushing it
      // on the object stack in the parse.
      parse.pushObject(bpelProcess);
      try {
        if (parse.setInputSource(bpel).execute().getDocumentObject() != null) {
          bpelProcess.staticAnalysisCheck(parse);
        }
      } finally {
        parse.popObject();
        parse.checkProblems("bpel file");
      }
    }
    Repository repository = environment.get(Repository.class);
    Publisher servicePublisher = environment.get(Publisher.class);
    repository.storeProcess(bpelProcess);
    servicePublisher.publishServices(bpelProcess, environment);

    return bpelProcess;
  }

  public void undeploy(QName processQName, Environment environment) {
    Repository repository = environment.get(Repository.class);
    Publisher servicePublisher = environment.get(Publisher.class);
    BpelProcess bpelProcess = repository.removeProcess(processQName);
    servicePublisher.unpublishServices(bpelProcess, environment);
  }

}
