package org.ow2.orchestra.env;

import java.util.HashMap;
import java.util.Map;

import org.jbpm.env.EnvironmentFactory;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.services.impl.DbRepository;
import org.ow2.orchestra.services.impl.PublisherImpl;
import org.ow2.orchestra.services.impl.UnsupportedInvoker;
import org.ow2.orchestra.services.itf.Invoker;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;

/**
 * Generates a default environment configuration for running Orchestra.
 *
 * @author Guillaume Porcher, Charles Souillard
 *
 */
public class DefaultEnvXMLGenerator {

  private Map<String, String> applicationVariables;

  private Map<String, String> blockVariables;

  private Class< ? extends Repository> repositoryType = null;

  private Class< ? extends Publisher> publisherType = null;

  private Class< ? extends Invoker> invokerType = null;

  public DefaultEnvXMLGenerator() {
    applicationVariables = new HashMap<String, String>();
    blockVariables = new HashMap<String, String>();
    this.setPublisherType(PublisherImpl.class);
    this.setInvokerType(UnsupportedInvoker.class);
    
    String hibernateConfigXml = "";
    hibernateConfigXml += "    <hibernate-session-factory>\n";
    hibernateConfigXml += "      <properties resource='hibernate.properties' />\n";
    hibernateConfigXml += "      <mappings resource='bpel.hibernate.mappings.xml' />\n";
    hibernateConfigXml += "      <cache-configuration resource='bpel.cache.xml' usage='read-write' />\n";
    hibernateConfigXml += "    </hibernate-session-factory>";
    this.setApplicationEntry("hibernate-session-factory", hibernateConfigXml);

    String commandServiceConfigXml = "    <standard-command-service>\n";
    commandServiceConfigXml += "      <retry-interceptor />\n";
    commandServiceConfigXml += "      <environment-interceptor />\n";
    commandServiceConfigXml += "      <transaction-interceptor />\n";
    commandServiceConfigXml += "    </standard-command-service>\n";
    this.setApplicationEntry("standard-command-service", commandServiceConfigXml);
    this.setApplicationEntry("job-executor",
          "    <job-executor threads='1' auto-start='false' />");

    this.setRepositoryType(DbRepository.class);

    this.setBlockEntry("transaction", "    <standard-transaction />");
    this.setBlockEntry("session", "    <hibernate-session />");

    // TODO: change name
    this.setBlockEntry("persistence_service",
        "    <hibernate-bpel-persistence-service />");

    this.setBlockEntry("timer_session", "    <timer-session />");
    this.setBlockEntry("job-session", "    <job-session />");
  }

  /**
   * Generates a xml for the given environment configuration.
   *
   * @return
   */
  public String createEnvironmentXml() {
    if (invokerType == null) {
      throw new OrchestraException(
          "Error while creating environment xml : invoker is null");
    }
    if (publisherType == null) {
      throw new OrchestraException(
          "Error while creating environment xml : publisher is null");
    }
    if (repositoryType == null) {
      throw new OrchestraException(
          "Error while creating environment xml : repository is null");
    }
    String result = "<environment-definition>\n\n"
        + "  <application>\n";
    for (String entry : applicationVariables.values()) {
      result += entry + "\n";
    }
    result += "    <repository class='" + repositoryType.getName() + "' />\n"
        + "    <publisher class='" + publisherType.getName() + "' />\n"
        + "    <invoker class='" + invokerType.getName() + "' />\n"
        + "  </application>\n\n" + "  <block>\n";
    for (String entry : blockVariables.values()) {
      result += entry + "\n";
    }
    result += "  </block>\n\n" + "</environment-definition>\n";
    return result;
  }

  /**
   * Creates a new EnvironmentFactory from the list of entries.
   */
  public EnvironmentFactory createEnvironmentFactory() {
    EnvironmentFactory environmentFactory = BpelEnvironmentParser
        .parseEnvironmentFactoryFromXmlString(createEnvironmentXml());
    return environmentFactory;
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key
   *            objectName
   * @param classToUse
   *            classToUse at runtime
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDecl(String key, Class classToUse) {
    return "    <object name='" + key + "' class='" + classToUse.getName()
        + "' init='eager'/>";
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key
   *            objectName
   * @param classToUse
   *            classToUse at runtime
   * @param fieldKeys
   *            fields name
   * @param refKeys
   *            the keys of objects to reference in fields declaration
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDeclWithFields(String key, Class classToUse,
      String[] fieldKeys, String[] refKeys) {
    if (fieldKeys.length != refKeys.length) {
      throw new RuntimeException(
          "Error : fieldKeys and refKeys don't have the same number of elements");
    }
    String result = "    <object name='" + key + "' class='"
        + classToUse.getName() + "' init='eager'>\n";
    for (int i = 0; i < fieldKeys.length; i++) {
      result += "      <field name='" + fieldKeys[i] + "'><ref object='"
          + refKeys[i] + "' /></field>\n";
    }
    result += "    </object>";
    return result;
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param xml
   *            declaration of this entry
   * @return previous declaration for this name
   */
  public String setApplicationEntry(String name, String xml) {
    return applicationVariables.put(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @return previous declaration for this name
   */
  public String setApplicationEntry(String name, Class< ? > classToUse) {
    String xml = getObjectDecl(name, classToUse);
    return setApplicationEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @param fieldKeys
   *            list of fields to set
   * @param refKeys
   *            value to set to the fields
   * @return previous declaration for this name
   */
  public String setApplicationEntry(String name, Class< ? > classToUse,
      String[] fieldKeys, String[] refKeys) {
    String xml = getObjectDeclWithFields(name, classToUse, fieldKeys, refKeys);
    return setApplicationEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param xml
   *            declaration of this entry
   * @return previous declaration for this name
   */
  public String setBlockEntry(String name, String xml) {
    return blockVariables.put(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @return previous declaration for this name
   */
  public String setBlockEntry(String name, Class< ? > classToUse) {
    String xml = getObjectDecl(name, classToUse);
    return setBlockEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @param fieldKeys
   *            list of fields to set
   * @param refKeys
   *            value to set to the fields
   * @return previous declaration for this name
   */
  public String setBlockEntry(String name, Class< ? > classToUse,
      String[] fieldKeys, String[] refKeys) {
    String xml = getObjectDeclWithFields(name, classToUse, fieldKeys, refKeys);
    return setBlockEntry(name, xml);
  }

  public void setRepositoryType(Class< ? extends Repository> repositoryType) {
    this.repositoryType = repositoryType;
  }

  public void setPublisherType(Class< ? extends Publisher> publisherType) {
    this.publisherType = publisherType;
  }

  public void setInvokerType(Class< ? extends Invoker> invokerType) {
    this.invokerType = invokerType;
  }
}
