/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.xml.namespace.QName;

import org.apache.tools.ant.Project;
import org.apache.tools.ant.taskdefs.Delete;
import org.jbpm.env.Environment;
import org.jbpm.env.EnvironmentFactory;
import org.jbpm.tx.Transaction;
import org.jbpm.xml.Parse;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.deployment.Deployer;
import org.ow2.orchestra.deployment.Deployment;
import org.ow2.orchestra.env.EnvFactoryRepository;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.parsing.BpelImportParser;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;

public class RemoteDeployer implements RemoteDeployerMBean {

  public QName deploy(Deployment bpelDeployment) {
    QName processQName;
    EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    if (environmentFactory == null) {
      return null;
    }
    Environment environment = null;
    try {
      environment = environmentFactory.openEnvironment();
      Deployer deployer = new Deployer();

      BpelProcess bpelProcess = deployer.deploy(bpelDeployment, environment);
      if (bpelProcess == null) {
        throw new OrchestraException("Process is null");
      }
      processQName = bpelProcess.getQName();
    } catch (Exception e) {
      if (environment != null) {
        Transaction t = environment.getTransaction();
        if (t != null) {
          t.setRollbackOnly();
        }
      }
      throw new OrchestraException(e);
    } finally {
      if (environment != null) {
        environment.close();
      }
    }

    return processQName;

  }

  public boolean undeploy(QName processQName) {
    EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    if (environmentFactory == null) {
      return false;
    }
    Environment environment = null;
    boolean result = false;
    try {
      environment = environmentFactory.openEnvironment();
      Deployer deployer = new Deployer();
      deployer.undeploy(processQName, environment);
      result = true;
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
    return result;
  }

  public List<QName> listProcess() {
    EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    List<QName> result = new ArrayList<QName>();
    if (environmentFactory == null) {
      return result;
    }
    Environment environment = null;
    try {
      environment = environmentFactory.openEnvironment();
      Repository repository = environment.get(Repository.class);
      for (BpelProcess process : repository.getProcesses()) {
        result.add(process.getQName());
      }
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
    return result;
  }

  public ProcessInfo getProcess(QName processQName) {
    EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    ProcessInfo recProcess = null;
    if (environmentFactory == null) {
      return recProcess;
    }
    Environment environment = null;
    try {
      environment = environmentFactory.openEnvironment();
      Repository repository = environment.get(Repository.class);
      //TODO: add others info for the process !
      if (repository.getProcess(processQName) != null) {
        recProcess = new ProcessInfo();
        recProcess.setQname(repository.getProcess(processQName).getQName());
      }
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
    return recProcess;
  }

  public String printInstances(BpelExecution execution, int depth) {
    String s = "-";
    for (int i = 0; i < depth; i++) {
      s += "--";
    }
    s += ">" + execution.getUuid() + " (" + execution.getState() + "), node= "
        + execution.getNode() + "\n";
    if (execution.getExecutions() != null) {
      for (org.jbpm.pvm.Execution child : execution.getExecutions()) {
        s += printInstances((BpelExecution) child, depth + 1);
      }
    }
    return s;
  }

  public InstanceInfo getInstanceRecord(BpelExecution execution) {
    InstanceInfo instRec = new InstanceInfo();
    instRec.setInstanceUId(execution.getUuid());
    instRec.setNodeName(execution.getNode().getName());
    instRec.setState(execution.getState());
    return instRec;
  }

  public Map<QName, List<String>> listInstanceStatus() {
    EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    Map<QName, List<String>> result = new HashMap<QName, List<String>>();
    if (environmentFactory == null) {
      return result;
    }
    Environment environment = null;
    try {
      environment = environmentFactory.openEnvironment();
      Repository repository = environment.get(Repository.class);
      for (BpelProcess process : repository.getProcesses()) {
        result.put(process.getQName(), new ArrayList<String>());
        List<BpelExecution> instances = repository.getInstances(process
            .getQName());
        if (instances != null) {
          for (BpelExecution execution : instances) {
            result.get(process.getQName()).add(printInstances(execution, 0));
          }
        }
      }
    } finally {
      environment.close();
    }
    return result;
  }

  public Map<QName, List<InstanceInfo>> listInstance() {
    EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    Map<QName, List<InstanceInfo>> result = new HashMap<QName, List<InstanceInfo>>();
    if (environmentFactory == null) {
      return result;
    }
    Environment environment = null;
    try {
      System.out.println("*** RemoteDeplyer: opening env.");
      environment = environmentFactory.openEnvironment();
      Repository repository = environment.get(Repository.class);
      for (BpelProcess process : repository.getProcesses()) {
        System.out.println("*** RemoteDeplyer: process name= " + process.getQName());
        result.put(process.getQName(), new ArrayList<InstanceInfo>());
        List<BpelExecution> instances = repository.getInstances(process.getQName());
        if (instances != null) {
          for (BpelExecution execution : instances) {
            System.out.println("*** RemoteDeplyer: instance uid = " + getInstanceRecord(execution).getInstanceUId());
            result.get(process.getQName()).add(getInstanceRecord(execution));
          }
        }
      }
    } finally {
      environment.close();
    }
    return result;
  }

  public QName deployBar(byte[] barFile) {
    try {
      // Create temp dir
      new File(System.getProperty("java.io.tmpdir")).mkdirs();
      File tmpDir = File.createTempFile("orch", null, null);
      tmpDir.delete();
      if (!tmpDir.mkdirs()) {
        throw new IOException("Cannot create the temporary directory '"
            + tmpDir + "'.");
      }

      // Unzip file
      InputStream in = new ByteArrayInputStream(barFile);
      ZipInputStream zipInput = new ZipInputStream(in);
      ZipEntry zipFile;
      while ((zipFile = zipInput.getNextEntry()) != null) {
        // write the file to the disk
        // Create file
        OutputStream outputFile = new FileOutputStream(tmpDir.getAbsolutePath()
            + File.separator + zipFile.getName());
        try {
          // Copy file content
          int c;
          while ((c = zipInput.read()) != -1) {
            outputFile.write(c);
          }
        } finally {
          outputFile.close();
        }
      }
      zipInput.close();
      in.close();

      // Read properties from the unzipped directory:
      Properties props = new Properties();
      InputStream propsIs = new FileInputStream(tmpDir.getAbsolutePath() + File.separator
          + "bar.properties");
      try {
        props.load(propsIs);
      } finally {
        propsIs.close();
      }

      String bpelFileName = props.getProperty("bpel");
      URL bpelFile = new File(tmpDir.getAbsolutePath() + File.separator
          + bpelFileName).toURL();

      String wsdlsCmd = props.getProperty("wsdls");
      List<URL> wsdlUrls = new ArrayList<URL>();
      if (wsdlsCmd != null) {
        for (String wsdl : wsdlsCmd.split(",")) {
          wsdlUrls.add(new File(wsdl).toURL());
        }
      }
      Deployment deployment = BpelImportParser.parseBpelFromUrl(bpelFile,
          wsdlUrls);

      // Deploy the process
      QName processName = deploy(deployment);

      // Delete temp dir
      Delete d = new Delete();
      d.setTaskName("delete");
      d.setProject(new Project());
      d.setDir(tmpDir);
      d.execute();

      return processName;
    } catch (Exception e) {
      // TODO: Handle exceptions.
      throw new OrchestraException("Unable to deploy bar", e);
    }
  }

  public void initialize() {
    EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    if (environmentFactory == null) {
      return;
    }
    Environment environment = null;
    try {
      environment = environmentFactory.openEnvironment();
      Publisher publisher = environment.get(Publisher.class);
      Repository repository = environment.get(Repository.class);
      for (BpelProcess process : repository.getProcesses()) {
        publisher.republishServices(process, environment);
      }
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
  }

}
