/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the JBPM BPEL PUBLIC LICENSE AGREEMENT as
 * published by JBoss Inc.; either version 1.0 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package org.ow2.orchestra.lang.jaxen;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jaxen.BaseXPath;
import org.jaxen.Context;
import org.jaxen.ContextSupport;
import org.jaxen.FunctionContext;
import org.jaxen.JaxenException;
import org.jaxen.SimpleFunctionContext;
import org.jaxen.XPathFunctionContext;
import org.jaxen.dom.DocumentNavigator;
import org.jaxen.expr.CommentNodeStep;
import org.jaxen.expr.LocationPath;
import org.jaxen.expr.NameStep;
import org.jaxen.expr.Predicate;
import org.jaxen.expr.ProcessingInstructionNodeStep;
import org.jaxen.expr.Step;
import org.jaxen.expr.TextNodeStep;
import org.jaxen.saxpath.Axis;
import org.ow2.orchestra.exception.BpelFaultException;
import org.ow2.orchestra.lang.function.DoXslTranformationFunction;
import org.ow2.orchestra.lang.function.GetVariablePropertyFunction;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * @author Alejandro Guizar
 * @version $Revision: 1.9 $ $Date: 2007/09/04 06:42:26 $
 */
abstract class XPathEvaluator extends BaseXPath {

  private static final FunctionContext EMPTY_FUNCTION_CONTEXT = new SimpleFunctionContext();

  private static final Log LOG = LogFactory.getLog(XPathEvaluator.class);

  private static final long serialVersionUID = 1L;

  XPathEvaluator(String text) throws JaxenException {
    super(text, DocumentNavigator.getInstance());
  }

  protected static List selectOrCreateNodes(LocationPath location,
      Context context) throws JaxenException {
    List contextNodes = context.getNodeSet();
    // empty context nodeset -> empty selection
    if (contextNodes.isEmpty()) {
      return Collections.EMPTY_LIST;
    }
    ContextSupport support = context.getContextSupport();

    // absolute path?
    if (location.isAbsolute()) {
      // start from the root node
      Object rootNode = support.getNavigator().getDocumentNode(
          contextNodes.get(0));
      contextNodes = Collections.singletonList(rootNode);
    }

    Iterator stepIter = location.getSteps().iterator();
    while (stepIter.hasNext()) {
      // prepare the context for the current step
      Context stepContext = new Context(support);
      stepContext.setNodeSet(contextNodes);
      // evaluate the step, capture the selected nodes
      Step step = (Step) stepIter.next();
      contextNodes = step.evaluate(stepContext);
      // no node was selected?
      if (contextNodes.isEmpty()) {
        // try to create the missing node
        Node newNode = createNode(step, stepContext);
        // create a new list, since the existing empty list is immutable
        contextNodes = Collections.singletonList(newNode);
      }
    }
    return contextNodes;
  }

  private static Node createNode(Step step, Context context) {
    List nodeset = context.getNodeSet();
    if (nodeset.size() != 1) {
      LOG
          .error("cannot create node for context node set of size other than one: "
              + nodeset);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }

    Object contextNode = nodeset.get(0);
    if (!(contextNode instanceof Element)) {
      LOG.error("cannot create node for non-element context node: "
          + contextNode);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }

    Element contextElem = (Element) contextNode;
    Node newNode;

    switch (step.getAxis()) {
      case Axis.ATTRIBUTE:
        newNode = createAttribute(step, context, contextElem);
        break;
      case Axis.CHILD:
        if (step instanceof NameStep) {
          newNode = createElement((NameStep) step, context, contextElem);
        } else {
          newNode = createNonElementChild(step, contextElem);
        }
        contextElem.appendChild(newNode);
        break;
      default:
        LOG.error("cannot create node on the specified axis: " + step);
        throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }
    return newNode;
  }

  private static Attr createAttribute(Step step, Context context,
      Element contextElem) {
    if (!step.getPredicates().isEmpty()) {
      LOG.error("cannot create attribute for step with predicates: " + step);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }

    if (!(step instanceof NameStep)) {
      LOG.error("cannot create attribute for non-name node test: " + step);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }

    NameStep nameStep = (NameStep) step;
    String localName = nameStep.getLocalName();

    if ("*".equals(localName)) {
      LOG.error("cannot create attribute for any-name node test: " + nameStep);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }

    // BPEL-191: preserve source prefix
    String prefix = nameStep.getPrefix();
    String namespaceURI = context.translateNamespacePrefixToUri(prefix);

    Attr attribute = contextElem.getOwnerDocument().createAttributeNS(
        namespaceURI,
        namespaceURI != null ? prefix + ':' + localName : localName);
    contextElem.setAttributeNodeNS(attribute);

    return attribute;
  }

  private static Element createElement(NameStep step, Context context,
      Element contextElem) {
    String localName = step.getLocalName();
    if ("*".equals(localName)) {
      LOG.error("cannot create node for any-name node tests: " + step);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }

    // BPEL-191: preserve source prefix
    String prefix = step.getPrefix();
    String namespaceURI = context.translateNamespacePrefixToUri(prefix);

    List predicates = step.getPredicates();
    switch (predicates.size()) {
      case 0:
        break;
      case 1: {
        // prepare the context for the predicate
        List contextNodes = getElements(contextElem, namespaceURI, localName);
        Context predicateContext = new Context(context.getContextSupport());
        predicateContext.setNodeSet(contextNodes);

        try {
          // evaluate the predicate, capture the result
          Predicate predicate = (Predicate) predicates.get(0);
          Object result = predicate.evaluate(predicateContext);

          if (!(result instanceof Number)) {
            LOG
                .error("cannot create element for step with non-numeric predicate: "
                    + step);
            throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
          }

          if (((Number) result).intValue() != contextNodes.size() + 1) {
            LOG.error("cannot create element for step with numeric predicate "
                + "beyond the next position: " + step);
            throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
          }
        } catch (JaxenException e) {
          LOG.error("predicate evaluation failed", e);
          throw new BpelFaultException(
              XmlConstants.FAULT_SUB_LANGUAGE_EXECUTION);
        }
        break;
      }
      default:
        LOG.error("cannot create element for step with multiple predicates: "
            + step);
        throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }

    String qualifiedName = namespaceURI != null ? prefix + ':' + localName
        : localName;
    return contextElem.getOwnerDocument().createElementNS(namespaceURI,
        qualifiedName);
  }

  private static Node createNonElementChild(Step step, Element contextElem) {
    if (!step.getPredicates().isEmpty()) {
      LOG.error("cannot create node for step with predicates: " + step);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }

    Document contextDoc = contextElem.getOwnerDocument();

    if (step instanceof TextNodeStep) {
      return contextDoc.createTextNode("");
    }

    if (step instanceof ProcessingInstructionNodeStep) {
      ProcessingInstructionNodeStep processingStep = (ProcessingInstructionNodeStep) step;
      return contextDoc.createProcessingInstruction(processingStep.getName(),
          "");
    }

    if (step instanceof CommentNodeStep) {
      return contextDoc.createComment("");
    }

    LOG.error("cannot create node for any-node tests on the child axis: "
        + step);
    throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
  }

  private static List getElements(Element parent, String namespaceURI,
      String localName) {
    ArrayList elements = new ArrayList();
    for (Node child = parent.getFirstChild(); child != null; child = child
        .getNextSibling()) {
      if (BpelXmlUtil.nodeNameEquals(child, namespaceURI, localName)) {
        elements.add(child);
      }
    }
    return elements;
  }

  protected static Node narrowToSingleNode(List nodeset) {
    LOG.debug("narrowing to single node: " + nodeset);
    if (nodeset == null || nodeset.size() != 1) {
      LOG.error("selection of size other than one: " + nodeset);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }
    Object singleResult = nodeset.get(0);
    if (!(singleResult instanceof Node)) {
      LOG.error("selection is not a node: " + singleResult);
      throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
    }
    return (Node) singleResult;
  }

  protected static FunctionContext readFunctionLibrary() {
    SimpleFunctionContext fc = new XPathFunctionContext();
    fc.registerFunction(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE,
        "getVariableProperty", new GetVariablePropertyFunction());
    fc.registerFunction(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE,
        "doXslTransform", new DoXslTranformationFunction());
    return fc;
  }
}
