/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the JBPM BPEL PUBLIC LICENSE AGREEMENT as
 * published by JBoss Inc.; either version 1.0 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package org.ow2.orchestra.lang.jaxen;

import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.Duration;
import javax.xml.datatype.XMLGregorianCalendar;

import org.jaxen.Context;
import org.jaxen.ContextSupport;
import org.jaxen.FunctionContext;
import org.jaxen.JaxenException;
import org.jaxen.expr.Expr;
import org.jaxen.expr.FilterExpr;
import org.jaxen.expr.FunctionCallExpr;
import org.jaxen.expr.LiteralExpr;
import org.jaxen.expr.LocationPath;
import org.jaxen.expr.NumberExpr;
import org.jaxen.expr.Predicate;
import org.jaxen.expr.PredicateSet;
import org.jaxen.expr.VariableReferenceExpr;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.BpelFaultException;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.lang.evaluator.ExpressionEvaluator;
import org.ow2.orchestra.lang.evaluator.InvalidExpressionValue;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.var.Message;
import org.w3c.dom.Node;

/**
 * Evaluator for general expressions. When XPath 1.0 is used as an expression
 * language, there is no context node available.
 *
 * @author Alejandro Guizar
 * @version $Revision: 1.10 $ $Date: 2007/09/12 23:20:16 $
 */
class XPathExpressionEvaluator extends XPathEvaluator implements
    ExpressionEvaluator {

  // Create only one instance.
  // TODO: Is it thread safe? Not sure! Check!
  private static DatatypeFactory datatypeFactory;

  static {
    try {
      datatypeFactory = DatatypeFactory.newInstance();
    } catch (DatatypeConfigurationException e) {
      throw new ExceptionInInitializerError(e);
    }
  }

  /**
   * The jBPM configuration object name for the resource describing the function
   * library available to expressions.
   */

  private static final long serialVersionUID = 1L;

  private static final Logger LOG = Logger.getLogger(XPathExpressionEvaluator.class.getName());

  private static final FunctionContext FUNTION_LIBRARY = readFunctionLibrary();

  XPathExpressionEvaluator(String text) throws JaxenException {
    super(text);
  }

  /**
   * Evaluates this expression. The context of XPath evaluation is:
   * <ul>
   * <li>the context node, position and size are unspecified</li>
   * <li>the function library, namespace declaration and variable bindings are
   * obtained by a call to {@link #getContext(Object)} with the given
   * <code>contextInfo</code> as argument</li>
   * <ul>
   */
  public synchronized Object evaluate(BpelExecution bpelExecution) {
    try {
      List nodeSet = selectNodes(bpelExecution);
      if (nodeSet != null) {
        switch (nodeSet.size()) {
          case 0:
            break;
          case 1:
            return nodeSet.get(0);
          default:
            return nodeSet;
        }
      }
      return null;
    } catch (JaxenException e) {
      LOG.log(Level.SEVERE, "expression evaluation failed", e);
      throw new BpelFaultException(XmlConstants.FAULT_SUB_LANGUAGE_EXECUTION);
    }
  }

  public synchronized void assign(BpelExecution bpelExecution, Object value) {
    Expr rootExpr = getRootExpr();
    Context context = getContext(bpelExecution);
    // direct variable assignment?
    if (rootExpr instanceof VariableReferenceExpr) {
      assignVariable((VariableReferenceExpr) rootExpr, context, value);
    } else {
      try {
        new ExprAssigner().assign(rootExpr, context, value);
      } catch (JaxenException e) {
        LOG.log(Level.SEVERE, "expression assignment failed", e);
        throw new BpelFaultException(XmlConstants.FAULT_SUB_LANGUAGE_EXECUTION);
      }
    }
  }

  protected FunctionContext createFunctionContext() {
    return FUNTION_LIBRARY;
  }

  /**
   * Gets a context for XPath evaluation. The context for an expression
   * expression is:
   * <ul>
   * <li>the variable bindings are the variables in the scope of the given
   * execution</li>
   * <li>the function library is {@link #FUNTION_LIBRARY}</li>
   * <li>the namespace declarations are taken from the snippet</li>
   * </ul>
   *
   * @param execution
   */
  protected Context getContext(Object o) {
    if (!(o instanceof BpelExecution)) {
      return super.getContext(o);
    }
    BpelExecution bpelExecution = (BpelExecution) o;
    ContextSupport support = getContextSupport();
    support.setVariableContext(new BpelVariableContext(bpelExecution));

    /*
     * ContextSupport support = new ContextSupport(getNamespaceContext(),
     * getFunctionContext(), new BpelVariableContext(bpelExecution),
     * getNavigator());
     */

    return new Context(support);
  }

  private static void assignVariable(VariableReferenceExpr rootExpr,
      Context context, Object value) {
    String variableName = rootExpr.getVariableName();
    BpelVariableContext variableContext = (BpelVariableContext) context
        .getContextSupport().getVariableContext();
    // look for a dot in the variable name, indicating a message part access
    int dotIndex = variableName.indexOf('.');
    if (dotIndex == -1) {
      // dotless name => variable must be defined as schema type or element
      // find variable definition
      Variable variableDefinition = variableContext
          .findVariableDefinition(variableName);
      if (variableDefinition == null) {
        throw new OrchestraException("variable not found: " + variableName);
      }

      // extract variable value
      Object valueForAssign = variableContext.getValueForAssign(variableName);

      // prevent direct access to a message variable
      if (variableDefinition.getMessageType() != null) {
        throw new OrchestraException("illegal access to message variable: "
            + variableName);
      }

      // assign variable value
      variableContext.setValue(variableName, value);
    } else {
      // dotted name => variable must be defined as messsage
      String messageName = variableName.substring(0, dotIndex);

      // find variable definition
      Variable variableDefinition = variableContext
          .findVariableDefinition(messageName);
      if (variableDefinition == null) {
        throw new OrchestraException("variable not found: " + messageName);
      }

      // extract variable value
      Object valueForAssign = variableContext.getValueForAssign(messageName);

      // prevent access to part of non-message variable
      if (!(variableDefinition.getMessageType() != null)) {
        throw new OrchestraException(
            "illegal access to part of non-message variable: " + variableName);
      }

      // assign part value
      Message messageValue = (Message) valueForAssign;
      String partName = variableName.substring(dotIndex + 1);
      messageValue.setPart(partName, value);
    }
  }

  private static class ExprAssigner extends ExprVisitor {

    private Context myContext;

    private Object result;

    private JaxenException jaxenException;

    public void assign(Expr pathExpr, Context context, Object value)
      throws JaxenException {
      this.myContext = context;
      visit(pathExpr);

      if (jaxenException != null) {
        throw jaxenException;
      }
      Node node = null;
      if (result instanceof List) {
        node = narrowToSingleNode((List) result);
      } else {
        node = (Node) result;
      }
      BpelXmlUtil.setObjectValue(node, value, false);
    }

    public void visit(FilterExpr filterExpr) {
      visit(filterExpr.getExpr());
      // if result is not a list, then there is nothing to filter
      if (result instanceof List) {
        try {
          result = evaluatePredicates(filterExpr.getPredicateSet(),
              (List< ? >) result, myContext.getContextSupport());
        } catch (JaxenException e) {
          jaxenException = e;
        }
      }
    }

    public void visit(LocationPath locationPath) {
      if (result instanceof List) {
        myContext.setNodeSet((List< ? >) result);
      } else {
        myContext.setNodeSet(Collections.singletonList(result));
      }
      try {
        result = selectOrCreateNodes(locationPath, myContext);
      } catch (JaxenException e) {
        jaxenException = e;
      }
    }

    public void visit(VariableReferenceExpr varExpr) {
      String variableName = varExpr.getVariableName();
      BpelVariableContext variableContext = (BpelVariableContext) myContext
          .getContextSupport().getVariableContext();
      // look for a dot in the variable name, indicating a message part access
      int dotIndex = variableName.indexOf('.');
      if (dotIndex == -1) {
        // dotless name => variable must be defined as schema type or element
        // find variable definition
        Variable variableDefinition = variableContext
            .findVariableDefinition(variableName);
        if (variableDefinition == null) {
          throw new OrchestraException("variable not found: " + variableName);
        }
        // variable value for assign
        result = variableContext.getValueForAssign(variableName);
      } else {
        // dotted name => variable must be defined as messsage
        String messageName = variableName.substring(0, dotIndex);

        // find variable definition
        Variable definition = variableContext
            .findVariableDefinition(messageName);
        if (definition == null) {
          throw new OrchestraException("variable not found: " + messageName);
        }
        // prevent access to a non-existent part
        if (definition.getMessageType() == null) {
          throw new OrchestraException(
              "illegal access to part of non-message variable: " + variableName);
        }
        // part value for assign
        Message messageValue = (Message) variableContext
            .getValueForAssign(messageName);
        String partName = variableName.substring(dotIndex + 1);
        result = messageValue.getPartValue(partName);
      }
    }

    public void visit(LiteralExpr literalExpr) {
      try {
        result = literalExpr.evaluate(myContext);
      } catch (JaxenException e) {
        jaxenException = e;
      }
    }

    public void visit(NumberExpr numberExpr) {
      try {
        result = numberExpr.evaluate(myContext);
      } catch (JaxenException e) {
        jaxenException = e;
      }
    }

    public void visit(FunctionCallExpr callExpr) {
      try {
        result = callExpr.evaluate(myContext);
      } catch (JaxenException e) {
        jaxenException = e;
      }
    }

    @SuppressWarnings("unchecked")
    private List< ? > evaluatePredicates(PredicateSet predicateSet,
        List< ? > nodes, ContextSupport support) throws JaxenException {
      List<Predicate> predicates = (List<Predicate>) predicateSet
          .getPredicates();
      for (Predicate predicate : predicates) {
        nodes = predicateSet.applyPredicate(predicate, nodes, support);
      }
      return nodes;
    }
  }

  public boolean evaluateBoolean(BpelExecution bpelExecution) throws InvalidExpressionValue {
    try {
      //return booleanValueOf(evaluate(bpelExecution));
      // TODO: why can't we use the booleanValueOf() method here? Unclear to me!
      return (Boolean) evaluate(bpelExecution);
    //} catch (JaxenException e) {
    } catch (ClassCastException e) {
      throw new InvalidExpressionValue(e);
    }
  }

  public XMLGregorianCalendar evaluateDeadline(BpelExecution bpelExecution)
    throws InvalidExpressionValue {
    try {
      final String s = stringValueOf(evaluate(bpelExecution));
      return datatypeFactory.newXMLGregorianCalendar(s);
    } catch (JaxenException e) {
      throw new InvalidExpressionValue(e);
    } catch (IllegalArgumentException iae) {
      throw new InvalidExpressionValue(iae);
    }
  }

  public Duration evaluateDuration(BpelExecution bpelExecution)
    throws InvalidExpressionValue {
    try {
      final String s = stringValueOf(evaluate(bpelExecution));
      return datatypeFactory.newDuration(s);
    } catch (JaxenException e) {
      throw new InvalidExpressionValue(e);
    } catch (IllegalArgumentException iae) {
      throw new InvalidExpressionValue(iae);
    }
  }

  public int evaluateUnsignedInt(BpelExecution bpelExecution)
    throws InvalidExpressionValue {
    try {
      final Number nb = numberValueOf(evaluate(bpelExecution));
      final int n = nb.intValue();
      if (n < 0) {
        throw new InvalidExpressionValue("Should be an Unsigned Integer expression that conforms to xsd:unsignedInt");
      }
      return n;
    } catch (JaxenException e) {
      throw new InvalidExpressionValue(e);
    }
  }
}

