/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 *
 * DefaultConfigurationHandler.java
 *
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet De Boudemange
 *
 * Created on : Mar 8, 2007
 *
 * $Id: BasicHandler.java 1453 2008-04-21 12:13:58Z porcherg $
 **/
package org.ow2.orchestra.parsing;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Locale;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
 *
 * BasicHandler. Generates an object using SAX events during the parsing of the
 * xml file. For each element "name", the corresponding field in the target
 * object is set by the method "setName"
 *
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet
 *         De Boudemange, Guillaume Porcher
 */
public class BasicHandler extends DefaultHandler {

  /**
   * conf.
   */
  private Object result;

  /**
   * The name of the current element, with its first letter capitalized. (to
   * enable us to find the matching setter method in the target class).
   */
  private String currentElementCapitalizedName;

  /**
   * The value of the element being parsed.
   */
  private String currentElementValue;

  /**
   * Set up a the target to parse the xml file.
   */
  public BasicHandler(Object target) {
    super();
    this.result = target;
  }

  /**
   * returns the target.
   *
   * @return object updated with values defined in the xml
   */
  public Object getResult() {
    return this.result;
  }

  /**
   * This method overrides the default characters() method of the SAX parser. It
   * sets the value of the current element.
   *
   * @param buf
   *            The characters.
   * @param offset
   *            The start position in the character array.
   * @param len
   *            The number of characters to use from the character array.
   */
  public void characters(final char[] buf, final int offset, final int len)
    throws SAXException {
    String value = new String(buf, offset, len);
    this.currentElementValue = value;
  }

  /**
   * This method overrides the default endElement() method of the SAX parser. It
   * calls the appropriate setter method in the Configuration class and set the
   * attributes which name matches the name of the element.
   *
   * @param namespaceURI
   *            namespace
   * @param sName
   *            localname
   * @param qName
   *            qualified name
   * @throws SAXException
   *             any SAX exception
   */
  public void endElement(final String namespaceURI, final String sName,
      final String qName) throws SAXException {
    try {
      // Capitalize the first letter of the element name for the setter method
      currentElementCapitalizedName = qName.substring(0, 1).toUpperCase(Locale.getDefault());
      currentElementCapitalizedName += qName.substring(1);

      // Find the setter method of the element (only if it is present in the
      // configuration file)
      if (currentElementValue != null
          && this.currentElementValue.trim().length() > 0) {
        Method setterMethod = result.getClass().getMethod(
            "set" + this.currentElementCapitalizedName,
            new Class[] { java.lang.String.class });
        setterMethod.invoke(result, this.currentElementValue);
      }
      // Set the attributes to null to avoid problems
      currentElementCapitalizedName = null;
      currentElementValue = null;

    } catch (NoSuchMethodException e) {
      // If no setter method has been found, it means that the element
      // does not enclose any value and that it can be safely skipped.
      // (This is the normal behaviour of top level elements for instance.)
      currentElementCapitalizedName = null;
      currentElementValue = null;
    } catch (IllegalAccessException e) {
      // TODO Handle errors appropriately
      e.printStackTrace();
    } catch (InvocationTargetException e) {
      // TODO Handle errors appropriately
      e.printStackTrace();
    }

  }

  @Override
  public void error(SAXParseException e) throws SAXException {
    throw e;
  }

}
