/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing;

import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Definition;
import javax.wsdl.Operation;
import javax.wsdl.PortType;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.jbpm.PvmException;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Problem;
import org.jbpm.xml.XmlUtil;
import org.ow2.novabpm.util.Misc;
import org.ow2.orchestra.deployment.Deployment;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.wsdl.WsdlDefinitions;
import org.ow2.orchestra.wsdl.WsdlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
*
* @author Goulven Le Jeune
*
*/
public final class BpelImportParser {

  private BpelImportParser() { }

  private static final Logger LOG = Logger.getLogger(BpelImportParser.class.getName());
  private static DocumentBuilderFactory documentBuilderFactory = null;

  private static DocumentBuilderFactory getDocumentBuilderFactory() {
    if (documentBuilderFactory == null) {
      documentBuilderFactory = DocumentBuilderFactory.newInstance();
      documentBuilderFactory.setNamespaceAware(true);
    }
    return documentBuilderFactory;
  }

  public static Deployment parseBpelFromUrl(URL bpelUrl, List<URL> wsdls) {
    List<Problem> problems = new ArrayList<Problem>();
    if (LOG.isLoggable(Level.INFO)) {
      LOG.info("Parsing BPEL from URL: " + bpelUrl + " using WSDL list: " + wsdls);
    }
    Deployment deployment = new Deployment();
    deployment.setBpelUrl(bpelUrl);
    WsdlDefinitions wsdlDefinitions = new WsdlDefinitions();

    if (wsdls != null) {
      for (URL wsdl : wsdls) {
        Definition wsdlDef = WsdlUtil.readWsdl(wsdl);
        wsdlDefinitions
            .addWsdlDefinition(wsdlDef.getTargetNamespace(), wsdlDef);
      }
    }

    DocumentBuilderFactory factory = getDocumentBuilderFactory();
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = factory.newDocumentBuilder();
    } catch (Exception e) {
      problems.add(new Problem("couldn't get new document builder", e, Problem.SEVERITY_WARNING));
      return null;
    }
    Document document = null;
    try {
      InputStream inStream = bpelUrl.openStream();
      document = documentBuilder.parse(inStream);
    } catch (Exception e) {
      problems.add(new Problem("couldn't parse xml document", e, Problem.SEVERITY_WARNING));
      return null;
    }
    NodeList imports = document.getElementsByTagNameNS(
        XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, "import");
    for (int i = 0; i < imports.getLength(); i++) {
      parseImportElement((Element) imports.item(i), problems, bpelUrl,
          wsdlDefinitions, deployment);
    }
    if (! problems.isEmpty()) {
      StringBuffer errorMsg = new StringBuffer();
      for (Problem p : problems) {
        errorMsg.append(Misc.LINE_SEPARATOR);
        errorMsg.append("  ");
        errorMsg.append(p.toString());
      }
      if (errorMsg!=null) {
        throw new OrchestraException("errors during parsing of bpel imports: "+errorMsg);
      }
    }
    return deployment;
  }

  /**
   * Parse Bpel process import element.
   */
  private static void parseImportElement(Element importElement, List<Problem> problems,
      URL bpelURL, WsdlDefinitions definitions, Deployment deployment) {
    String namespace = null;
    if (importElement.hasAttribute("namespace")) {
      namespace = importElement.getAttribute("namespace");
    }
    String location = null;
    if (importElement.hasAttribute("location")) {
      location = importElement.getAttribute("location");
    }

    String importType = null;
    if (importElement.hasAttribute("importType")) {
      importType = importElement.getAttribute("importType");
    }

    if (importType == null) {
      problems.add(new Problem("missing importType attribute in import element "
          + XmlUtil.toString(importElement), null, Problem.SEVERITY_WARNING));
    } else if (importType.equals(XmlConstants.XMLNS_WSDL)) {
      // Import a WSDL
      if (location != null) {
        // build url relative to the bpel url
        URL wsdlURL = null;
        String bpelPath = bpelURL.toExternalForm();
        try {
          if (location.startsWith("http://") || location.startsWith("https://")
              || location.startsWith("file:")) {
            wsdlURL = new URL(location);
          } else if (bpelURL.toURI().isOpaque() && bpelPath.startsWith("jar:")) {
            int lastSlash = bpelPath.lastIndexOf("!") + 1;
            String packageBase = bpelPath.substring(0, lastSlash);
            String bpelBase = bpelPath.substring(lastSlash);

            URI wsdlLocationInPackage = URI.create(bpelBase).resolve(location);

            wsdlURL = new URL(packageBase + wsdlLocationInPackage);
          } else {
            wsdlURL = bpelURL.toURI().resolve(location).toURL();
          }
        } catch (Exception e) {
          throw new OrchestraException("Exception while building wsdlURL : ", e);
        }
        // parse definition
        Definition wsdlDefinition = WsdlUtil.readWsdl(wsdlURL);
        // perform static analysis
        if (namespace == null && wsdlDefinition.getTargetNamespace() != null) {
          problems.add(new Problem("invalid namespace in imported wsdl (SA00012) :"
              + " no target namespace expected, got \""
              + wsdlDefinition.getTargetNamespace() + "\"", null, Problem.SEVERITY_WARNING));
        } else if (namespace != null
            && !namespace.equals(wsdlDefinition.getTargetNamespace())) {
          problems.add(new Problem("invalid namespace in imported wsdl (SA00011) :"
              + " expected \"" + namespace + "\", got \""
              + wsdlDefinition.getTargetNamespace() + "\"", null, Problem.SEVERITY_WARNING));
        }

        // Static Analysis
        Map<QName, PortType> portTypes = wsdlDefinition.getAllPortTypes();
        for (PortType pt : portTypes.values()) {
          List<Operation> operations = pt.getOperations();
          for (int index = 0; index < operations.size(); index++) {
            Operation op = operations.get(index);
            String name = op.getName();
            int cmpt = 1;
            for (int i = index + 1; i < operations.size(); i++) {
              if (operations.get(i).getName().equals(name)) {
                cmpt++;
              }
            }
            if (cmpt > 1) {
              // SA00002
              String message = StaticAnalysisFault.getMessage("SA00002",
                  "  PartnerLink " + pt.getQName().toString() + " has " + cmpt + " operations with name " + name);
              problems.add(new Problem(message, new StaticAnalysisException("SA00002"), Problem.SEVERITY_WARNING));
            }
          }
        }

        deployment.addWsdlDefinition(namespace, wsdlDefinition);
      } else {
        Set<Definition> definitionsForNS = null;
        if (definitions != null) {
          definitionsForNS = definitions.getWsdlDefinitions(namespace);
        }
        if (definitionsForNS == null) {
          problems.add(new Problem("No wsdl file to import in import element "
              + XmlUtil.toString(importElement), null, Problem.SEVERITY_WARNING));
        } else {
          for (Definition def : definitionsForNS) {
            deployment.addWsdlDefinition(namespace, def);
          }
        }
      }
    } else {
      problems.add(new Problem("unsupported importType attribute in import element "
          + XmlUtil.toString(importElement), null, Problem.SEVERITY_WARNING));
    }
  }
}
