package org.ow2.orchestra.parsing;


import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.wsdl.Definition;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.jbpm.pvm.impl.NodeImpl;
import org.jbpm.xml.Bindings;
import org.jbpm.xml.Entity;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.UrlEntity;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.parsing.binding.AssignBinding;
import org.ow2.orchestra.parsing.binding.CompensateScopeBinding;
import org.ow2.orchestra.parsing.binding.EmptyBinding;
import org.ow2.orchestra.parsing.binding.ExitBinding;
import org.ow2.orchestra.parsing.binding.ExtensionActivityBinding;
import org.ow2.orchestra.parsing.binding.FlowBinding;
import org.ow2.orchestra.parsing.binding.ForEachBinding;
import org.ow2.orchestra.parsing.binding.IfBinding;
import org.ow2.orchestra.parsing.binding.InvokeBinding;
import org.ow2.orchestra.parsing.binding.PickBinding;
import org.ow2.orchestra.parsing.binding.ReceiveBinding;
import org.ow2.orchestra.parsing.binding.RepeatUntilBinding;
import org.ow2.orchestra.parsing.binding.ReplyBinding;
import org.ow2.orchestra.parsing.binding.RethrowBinding;
import org.ow2.orchestra.parsing.binding.ScopeBinding;
import org.ow2.orchestra.parsing.binding.SequenceBinding;
import org.ow2.orchestra.parsing.binding.ThrowBinding;
import org.ow2.orchestra.parsing.binding.ValidateBinding;
import org.ow2.orchestra.parsing.binding.WaitBinding;
import org.ow2.orchestra.parsing.binding.WhileBinding;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.wsdl.BpelWsdlParser;
import org.ow2.orchestra.wsdl.WsdlDefinitions;
import org.w3c.dom.Element;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/** parses the dom tree of the bpel document and stores the
 * builds the BpelProcess.   The Deployer will make the BpelProcess
 * available in the object stack of the parse.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class BpelParser extends Parser implements ErrorHandler {

  /** path to the directory containing all xsd files. */
  private static final String RESSOURCES_DIR = "";
  /** all schema resources used to parse bpel files. */
  private static final String[] SCHEMA_RESOURCES = {
    RESSOURCES_DIR + "wsdl.xsd", RESSOURCES_DIR + "xml-schema.xsd",
    RESSOURCES_DIR + "ws-bpel_abstract_common_base.xsd", RESSOURCES_DIR + "ws-bpel_executable.xsd"};
  /** all bindings used for parsing. */
  private static Bindings defaultBindings = getDefaultBindings();
  /** schema document URIs as stringsF. */
  private static String[] schemaSources = getSchemaSources();

  /** all allowed tag names for process top level activities. */
  //public static final Set<String> processActivityTagNames = getProcessActivityTagNames();

  // the default entities are initialized at the bottom of this file.
  private static Map<String, Entity> defaultEntities = getDefaultEntities();

  private BpelWsdlParser wsdlParser = null;
  /**
   * Default constructor.
   */
  public BpelParser() {
    super(defaultBindings, defaultEntities);
    getDocumentBuilderFactory().setNamespaceAware(true);
    wsdlParser = new BpelWsdlParser();
  }

  /**
   * Parses a the dom document representing the bpel file.
   * @param processElement root element of the document : the process element
   * @param parse parse
   * @return the resulting bpelProcess
   */
  public Object parseDocumentElement(Element processElement, Parse parse) {
    BpelProcess bpelProcess = parse.findObject(BpelProcess.class);

    List<Element> elements = XmlUtil.elements(processElement);
    if (elements == null || elements.isEmpty()) {
      parse.addProblem("no elements in process");
      return bpelProcess;
    }
    parseProcessAttributes(processElement, bpelProcess, parse);
    parseProcessElements(processElement, bpelProcess, parse);

    // creates a scope associated to the process : in the bindings : process is mapped to ScopeBinding
    // let's create the top level node for the bpel process.

    NodeImpl processScope = (NodeImpl) parseElement(processElement, parse, "activities");
    bpelProcess.setInitial(processScope);

    return bpelProcess;
  }

  /**
   * Parses all process specific attributes (those not common with scope).
   * @param processElement process element (dom)
   * @param bpelProcess bpelProcess
   * @param parse parse
   */
  private void parseProcessAttributes(Element processElement, BpelProcess bpelProcess, Parse parse) {
    String name = XmlUtil.attribute(processElement, "name");
    bpelProcess.setName(name);

    String targetNamespace = XmlUtil.attribute(processElement, "targetNamespace");
    bpelProcess.setTargetNamespace(targetNamespace);

    String queryLanguage = XmlUtil.attribute(processElement, "queryLanguage");
    bpelProcess.setQueryLanguage(queryLanguage);

    String expressionLanguage = XmlUtil.attribute(processElement, "expressionLanguage");
    bpelProcess.setExpressionLanguage(expressionLanguage);
  }

  /**
   * Parses all process specific children elements (those not common with scope).
   * @param elements list of all process children elements
   * @param bpelProcess bpelProcess
   * @param parse parse
   */
  private void parseProcessElements(Element processElement, BpelProcess bpelProcess, Parse parse) {
    Element extensions = XmlUtil.element(processElement, "extensions");
    if (extensions != null) {
      parse.addProblem("extensions element not yet supported on process");
    }

    List<Element> importElements = XmlUtil.elements(processElement, "import");
    if (importElements != null) {
//    parse.addProblem("import element not yet supported on process");
      for (Element importElement : importElements) {
        parseImportElement(importElement, bpelProcess, parse);
      }
    }
  }

  /**
   * Parse Bpel process import element.
   */
  private void parseImportElement(Element importElement, BpelProcess bpelProcess, Parse parse) {
    String namespace = null;
    if (importElement.hasAttribute("namespace")) {
      namespace = importElement.getAttribute("namespace");
    }
    /*String location = null;
    if (importElement.hasAttribute("location")) {
      location = importElement.getAttribute("location");
    }*/

    String importType = null;
    if (importElement.hasAttribute("importType")) {
      importType = importElement.getAttribute("importType");
    }

    if (importType == null) {
      parse.addProblem("missing importType attribute in import element " + XmlUtil.toString(importElement));
    } else if (importType.equals(XmlConstants.XMLNS_WSDL)) {
      // Import a WSDL from Deployment
      WsdlDefinitions wsdlDefinitions = parse.findObject(WsdlDefinitions.class);
      Set<Definition> definitionsForNS = null;
      if (wsdlDefinitions != null) {
        definitionsForNS = wsdlDefinitions.getWsdlDefinitions(namespace);
      }
      if (definitionsForNS == null) {
        parse.addProblem("No wsdl file to import in import element " + XmlUtil.toString(importElement));
      } else {
        for (Definition wsdlDef : definitionsForNS) {
          wsdlParser.parse(
              wsdlDef,
              parse,
              bpelProcess.getQueryLanguage(),
              bpelProcess,
              bpelProcess.getWsdlInfos());
        }
      }
    } else {
      parse.addProblem("unsupported importType attribute in import element " + XmlUtil.toString(importElement));
    }
  }

  /**
   * returns a configured documentBuilderFactory using schemaLanguage, xsd resources...
   * @return the build DocumentBuilderFactory
   */
  public synchronized DocumentBuilderFactory getDocumentBuilderFactory() {
    documentBuilderFactory = newDocumentBuilderFactory();
    documentBuilderFactory.setNamespaceAware(true);
    documentBuilderFactory.setValidating(true);
    // ignore white space can only be set if parser is validating
    documentBuilderFactory.setIgnoringElementContentWhitespace(true);
    // select xml schema as the schema language (a.o.t. DTD)
    documentBuilderFactory.setAttribute(
      "http://java.sun.com/xml/jaxp/properties/schemaLanguage",
      "http://www.w3.org/2001/XMLSchema");
    // set schema sources
    documentBuilderFactory.setAttribute(
        "http://java.sun.com/xml/jaxp/properties/schemaSource", schemaSources);
    return documentBuilderFactory;
  }

  /**
   * Returns all necessary bindings that the parser should use.
   * @return the build Bindings object
   */
  public static Bindings getDefaultBindings() {
    // dunno if you need those.
    Bindings bindings = new Bindings();
    bindings.addBinding(getBPELQName("process"), new ScopeBinding(), "activities");
    bindings.addBinding(getBPELQName("sequence"), new SequenceBinding(), "activities");
    bindings.addBinding(getBPELQName("scope"), new ScopeBinding(), "activities");
    bindings.addBinding(getBPELQName("receive"), new ReceiveBinding(), "activities");
    bindings.addBinding(getBPELQName("reply"), new ReplyBinding(), "activities");
    bindings.addBinding(getBPELQName("flow"), new FlowBinding(), "activities");
    bindings.addBinding(getBPELQName("empty"), new EmptyBinding(), "activities");
    bindings.addBinding(getBPELQName("assign"), new AssignBinding(), "activities");
    bindings.addBinding(getBPELQName("invoke"), new InvokeBinding(), "activities");
    bindings.addBinding(getBPELQName("if"), new IfBinding(), "activities");
    bindings.addBinding(getBPELQName("compensate"), new CompensateScopeBinding(), "activities");
    bindings.addBinding(getBPELQName("compensateScope"), new CompensateScopeBinding(), "activities");
    bindings.addBinding(getBPELQName("exit"), new ExitBinding(), "activities");
    bindings.addBinding(getBPELQName("forEach"), new ForEachBinding(), "activities");
    bindings.addBinding(getBPELQName("pick"), new PickBinding(), "activities");
    bindings.addBinding(getBPELQName("repeatUntil"), new RepeatUntilBinding(), "activities");
    bindings.addBinding(getBPELQName("throw"), new ThrowBinding(), "activities");
    bindings.addBinding(getBPELQName("rethrow"), new RethrowBinding(), "activities");
    bindings.addBinding(getBPELQName("validate"), new ValidateBinding(), "activities");
    bindings.addBinding(getBPELQName("wait"), new WaitBinding(), "activities");
    bindings.addBinding(getBPELQName("while"), new WhileBinding(), "activities");

    bindings.addBinding(getBPELQName("extensionActivity"), new ExtensionActivityBinding(), "activities");
    return bindings;
  }

  private static QName getBPELQName(String localName) {
    return new QName(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, localName);
  }

  /**
   * Returns all necessary entities that the parser should use.
   * @return a map containing all entities : key : url on the internet, value : local resource.
   */
  private static Map<String, Entity> getDefaultEntities() {
    Map<String, Entity> defaultSchemaCatalog = new HashMap<String, Entity>();
    ClassLoader resourceLoader = BpelParser.class.getClassLoader();
    defaultSchemaCatalog.put("http://schemas.xmlsoap.org/wsdl/",
        new UrlEntity(RESSOURCES_DIR + "wsdl.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://www.w3.org/2001/XMLSchema",
        new UrlEntity(RESSOURCES_DIR + "xml-schema.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://www.w3.org/2001/xml.xsd",
        new UrlEntity(RESSOURCES_DIR + "xml-schema.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://docs.oasis-open.org/wsbpel/2.0/process/abstract",
        new UrlEntity(RESSOURCES_DIR + "ws-bpel_abstract_common_base.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://docs.oasis-open.org/wsbpel/2.0/process/executable",
        new UrlEntity(RESSOURCES_DIR + "ws-bpel_executable.xsd", resourceLoader));
    return defaultSchemaCatalog;
  }

  /**
   * Build a string array containing all schema resources to use when parsing a bpel file.
   * @return the build string array
   */
  private static String[] getSchemaSources() {
    String[] schemas = new String[SCHEMA_RESOURCES.length];
    ClassLoader resourceLoader = BpelParser.class.getClassLoader();
    for (int i = 0; i < SCHEMA_RESOURCES.length; i++) {
      schemas[i] = resourceLoader.getResource(SCHEMA_RESOURCES[i]).toExternalForm();
    }
    return schemas;
  }

  protected void setErrorHandler(DocumentBuilder documentBuilder) {
    documentBuilder.setErrorHandler(this);
  }

  public void error(SAXParseException exception) throws SAXException {
    throw exception;
  }

  public void fatalError(SAXParseException exception) throws SAXException {
    throw exception;
  }

  public void warning(SAXParseException exception) throws SAXException {
    throw exception;
  }
}
