/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Assign;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.FromTo;
import org.ow2.orchestra.definition.element.Literal;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.definition.element.FromTo.Variant;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.lang.PropertyQuery;
import org.ow2.orchestra.lang.VariableQuery;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.StyleSheetRepository;
import org.ow2.orchestra.util.XmlConstants;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * @author Charles Souillard
 */
public class AssignBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(AssignBinding.class.getName());

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *            dom element to parse
   * @param parse
   *            parse
   * @param parser
   *            parser
   * @return the build behaviour!
   */
  public Object parse(Element element, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    Assign assign = new Assign(UUID.randomUUID().toString());
    setEnclosingScope(assign, parse);
    parseStandardAttributes(element, assign, parse);
    parseStandardElements(element, assign, parse);

    // attributes
    boolean validate = BpelUtil.booleanValue(XmlUtil.attribute(element,
        "validate"));
    if (validate) {
      parse
          .addProblem("validate attribute not yet supported on assign activity");
      assign.setValidate(validate);
    }

    // copy elements
    List<Element> copyElements = XmlUtil.elements(element, "copy");
    parseCopies(copyElements, assign, parse);

    // extensionAssignOperation elements
    List<Element> extensionAssignOperationElements = XmlUtil.elements(element,
        "extensionAssignOperation");
    if (extensionAssignOperationElements != null) {
      parse
          .addProblem("extensionAssignOperation not yet supported on assign activity");
    }

    org.jbpm.pvm.Node activityNode = createActivityNode(element, parse, parser,
        assign);
    return activityNode;
  }

  /**
   * Parses copy elements of the given assign.
   *
   * @param copyElements
   *            a the list of all copies of this assign
   * @param parse
   *            parse
   * @param assign
   *            assign to analyse
   */
  private void parseCopies(List<Element> copyElements, Assign assign,
      Parse parse) {
    List<Copy> copies = new ArrayList<Copy>();
    if (copyElements != null) {
      for (Element copyElement : copyElements) {
        copies.add(parseCopy(copyElement, parse, assign.getEnclosingScope()));
      }
    }
    assign.setCopies(copies);
  }

  /**
   * Parses a copy element.
   *
   * @param copyElement
   *            copy dom element to parse
   * @param parse
   *            parse
   * @param enclosingScope
   *            enclosingScope
   * @return the build Copy
   */
  private Copy parseCopy(Element copyElement, Parse parse, Scope enclosingScope) {
    Copy copy = new Copy();
    // attributes
    boolean keepSrcElementName = BpelUtil.booleanValue(XmlUtil.attribute(
        copyElement, "keepSrcElementName"));
    if (keepSrcElementName) {
      parse
          .addProblem("keepSrcElementName attribute not yet supported on copy element");
      copy.setKeepSrcElementName(keepSrcElementName);
    }

    boolean ignoreMissingFromData = BpelUtil.booleanValue(XmlUtil.attribute(
        copyElement, "ignoreMissingFromData"));
    if (ignoreMissingFromData) {
      parse
          .addProblem("ignoreMissingFromData attribute not yet supported on copy element");
      copy.setIgnoreMissingFromData(ignoreMissingFromData);
    }

    parseFrom(XmlUtil.element(copyElement, "from"), copy, parse, enclosingScope);
    parseTo(XmlUtil.element(copyElement, "to"), copy, parse, enclosingScope);
    return copy;
  }

  /**
   * Parses a from element.
   *
   * @param fromElement
   *            from dom element to parse
   * @param parse
   *            parse
   * @param enclosingScope
   *            enclosingScope
   * @param copy
   *            enclosing copy
   */
  protected static void parseFrom(Element fromElement, Copy copy, Parse parse,
      Scope enclosingScope) {
    From from = new From();
    // attributes
    from.setEndpointReference(XmlUtil.attribute(fromElement,
        "endpointReference"));

    // parse from and to common elements/attributes
    parseFromTo(fromElement, from, parse, enclosingScope);

    // parse literal
    Element literalElement = XmlUtil.element(fromElement, "literal");
    if (literalElement != null) {
      from.setLiteral(parseLiteral(literalElement));
      from.setVariant(Variant.LITERAL);
    }

    copy.setFrom(from);
  }

  /**
   * Parses a to element.
   *
   * @param toElement
   *            to dom element to parse
   * @param parse
   *            parse
   * @param enclosingScope
   *            enclosingScope
   * @param copy
   *            enclosing copy
   */
  private void parseTo(Element toElement, Copy copy, Parse parse,
      Scope enclosingScope) {
    To to = new To();
    // parse from and to common elements/attributes
    parseFromTo(toElement, to, parse, enclosingScope);
    copy.setTo(to);
  }

  /**
   * Parses common attributes/children elements of from and to elements.
   *
   * @param fromToElement
   *            dom element to parse (from or to)
   * @param enclosingScope
   *            enclosingScope
   * @param parse
   *            parse
   * @param fromTo
   *            build fromTo
   */
  protected static void parseFromTo(Element fromToElement, FromTo fromTo,
      Parse parse, Scope enclosingScope) {
    // attributes
    fromTo.setExpressionLanguage(XmlUtil.attribute(fromToElement,
        "expressionLanguage"));
    fromTo.setVariable(XmlUtil.attribute(fromToElement, "variable"));
    fromTo.setPart(XmlUtil.attribute(fromToElement, "part"));
    fromTo.setProperty(XmlUtil.attributeQName(fromToElement, "property"));
    fromTo.setPartnerLink(XmlUtil.attribute(fromToElement, "partnerLink"));

    BpelProcess bpelProcess = (BpelProcess) parse.findObject(BpelProcess.class);
    if (fromTo.getVariable() != null && fromTo.getProperty() == null) {
      // we are in the variable variant
      fromTo.setVariant(Variant.VAR);
      // TODO : bug here maybe the variable is defined an a scope father...
      Variable variable = enclosingScope.findVariable(fromTo.getVariable());
      Element queryElement = XmlUtil.element(fromToElement, "query");
      if (queryElement != null) {
        String queryLanguage = XmlUtil.attribute(queryElement, "queryLanguage");
        if (queryLanguage == null) {
          queryLanguage = bpelProcess.getQueryLanguage();
        }
        VariableQuery query = new VariableQuery();
        query.setLanguage(queryLanguage);
        query.setNamespaces(BpelUtil.getAllNameSpaces(queryElement));
        query.setText(queryElement.getTextContent());
        fromTo.setVariableQuery(query);
      }
    } else if (fromTo.getVariable() != null) {
      // we are in the property variant
      fromTo.setVariant(Variant.PROP);
      String queryString = XmlConstants.BPEL_FUNC_PREFIX
          + ":getVariableProperty(\"" + fromTo.getVariable() + "\",\""
          + fromTo.getProperty().getPrefix() + ":"
          + fromTo.getProperty().getLocalPart() + "\")";
      PropertyQuery query = new PropertyQuery();
      query.setLanguage(bpelProcess.getQueryLanguage());
      query.setNamespaces(BpelUtil.getAllNameSpaces(fromToElement));
      query.setText(queryString);

      fromTo.setPropertyQuery(query);
    } else if (fromTo.getPartnerLink() != null) {
      // we are in the partnerLink variant
      fromTo.setVariant(Variant.PLNK);
    } else if (fromTo instanceof To
        || (fromTo instanceof From && ((From) fromTo).getLiteral() == null)) {
      fromTo.setVariant(Variant.EXPR);
      Expression expression = new Expression();
      expression.setLanguage(bpelProcess.getExpressionLanguage());
      expression.setNamespaces(BpelUtil.getAllNameSpaces(fromToElement));
      expression.setText(fromToElement.getTextContent());
      fromTo.setExpression(expression);
      addStyleSheets(expression.getText(), bpelProcess);
      if (fromTo.getExpressionLanguage() == null) {
        fromTo.setExpressionLanguage(bpelProcess.getExpressionLanguage());
      }
    }
  }

  protected static void addStyleSheets(String expression,
      BpelProcess bpelProcess) {
    String functionNameBegin = "doXslTransform(\"";
    if (expression.contains(functionNameBegin)) {
      int indexBegin = expression.indexOf(functionNameBegin);
      int indexEnd = expression.indexOf(",", indexBegin
          + functionNameBegin.length());
      String styleSheet = expression.substring(indexBegin
          + functionNameBegin.length(), indexEnd - 1);
      URL xslURL = null;
      try {
        if (styleSheet.startsWith("http://") || styleSheet.startsWith("file:")) {
          xslURL = new URL(styleSheet);
        } else {
          URL bpelURL = bpelProcess.getBpelURL();
          String bpelPath = bpelURL.toExternalForm();
          int lastSlash = bpelPath.lastIndexOf("/");
          int lastBackSlash = bpelPath.lastIndexOf("\\");
          String bpelBase = null;
          if (lastSlash > lastBackSlash) {
            bpelBase = bpelPath.substring(0, lastSlash);
          } else {
            bpelBase = bpelPath.substring(0, lastBackSlash);
          }

          xslURL = new URL(bpelBase + "/" + styleSheet);
        }
      } catch (Exception e) {
        throw new OrchestraException("Exception while building xslURL : ", e);
      }
      StyleSheetRepository styleSheetRepository = bpelProcess
          .getStyleSheetRepository();
      styleSheetRepository.addStyleSheet(xslURL, styleSheet);
    }
  }

  /**
   * Parses literal dom element.
   *
   * @param literalElement
   *            dom element to parse
   * @return build Literal object
   */
  protected static Literal parseLiteral(Element literalElement) {
    Literal literal = new Literal();
    Element value = null;

    List<Element> elements = BpelUtil.getElements(literalElement);
    if (elements.size() > 1) {
      // TODO : maybe it is a list ?
      throw new OrchestraException(
          "Invalid literal element : it has more than one element child");
    } else if (elements.size() == 1) {
      value = elements.get(0);
    }
    if (value == null) {
      // we need to look for Texts and CDATA
      NodeList nodeChilds = literalElement.getChildNodes();
      String data = null;
      for (int i = 0; i < nodeChilds.getLength(); i++) {
        Node node = nodeChilds.item(i);
        if (node.getNodeType() == Node.TEXT_NODE
            || node.getNodeType() == Node.CDATA_SECTION_NODE) {
          data = ((Text) node).getData();
          value = BpelXmlUtil.createElementWithContent(data);
          // we only take the first
          break;
        }
      }
    }
    if (value == null) {
      // we create an empty text node
      value = BpelXmlUtil.createElementWithContent("");
    }
    literal.setValue(value);
    return literal;
  }

}
