/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.impl.NodeImpl;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.activity.Flow;
import org.ow2.orchestra.definition.element.Link;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */

public class FlowBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(FlowBinding.class.getName());

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *            dom element to parse
   * @param parse
   *            parse
   * @param parser
   *            parser
   * @return the build behaviour!
   */
  public Object parse(Element element, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    Flow flow = new Flow(UUID.randomUUID().toString());
    setEnclosingScope(flow, parse);
    parseStandardAttributes(element, flow, parse);
    parseStandardElements(element, flow, parse);

    Element linksElement = XmlUtil.element(element, "links");
    if (linksElement != null) {
      List<Element> linkElements = XmlUtil.elements(linksElement, "link");
      parseLinks(linkElements, flow, parse);
    }

    NodeImpl activityNode = createActivityNodeWithChildren(element, parse,
        parser, flow);

    checkLinks(flow, parse);

    return activityNode;
  }

  /**
   * Parses links elements of the given flow.
   *
   * @param linkElements
   *            the list of all links elements of the given flow
   * @param flow
   *            flow
   */
  private void parseLinks(List<Element> linkElements, Flow flow, Parse parse) {
    Map<String, Link> links = new HashMap<String, Link>();
    if (linkElements != null) {
      for (Element linkElement : linkElements) {
        String linkName = XmlUtil.attribute(linkElement, "name");
        if (links.containsKey(linkName)) {
          // Static Analysis SA00064
          String message = StaticAnalysisFault.getMessage("SA00064",
              "  link " + linkName + " defined twice in flow " + flow.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00064"));
        } else {
          Link link = new Link();
          link.setName(linkName);
          links.put(linkName, link);
        }
      }
    }
    flow.setLinks(links);
  }

  /**
   * Check that every link defined in a flow have a corresponding source and target
   * activities defined.
   *
   * @param flow : The flow on which links should be checked
   * @param parse : parse
   */
  private void checkLinks(Flow flow, Parse parse) {
    Map<String, Link> links = flow.getLinks();
    if (links != null) {
      for (Link link : links.values()) {
        if ((link.getSourceActivity() == null) || (link.getTargetActivity() == null)) {
          // Static Analysis SA00066
          String message = StaticAnalysisFault.getMessage("SA00066",
              "  Link " + link.getName() + " missing either source or target");
          parse.addProblem(message, new StaticAnalysisException("SA00066"));
        }
      }
    }
  }
}
