/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Operation;
import javax.xml.namespace.QName;

import org.jbpm.pvm.impl.NodeImpl;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Invoke;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Catch;
import org.ow2.orchestra.definition.element.CompensationHandler;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.FromPart;
import org.ow2.orchestra.definition.element.FromTo.Variant;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.ToPart;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */
public class InvokeBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(InvokeBinding.class.getName());

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *            dom element to parse
   * @param parse
   *            parse
   * @param parser
   *            parser
   * @return the build behaviour!
   */
  public Object parse(Element element, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    String uuid = UUID.randomUUID().toString();
    Scope scope = new Scope(uuid);
    setEnclosingScope(scope, parse);
    scope.setName("surroundingInvoke_" + uuid);
    parseStandardAttributes(element, scope, parse);

    String nodeName = getNodeName(Scope.class.getSimpleName(), "surroundingInvoke_" + uuid);
    NodeImpl scopeNode = createActivityNode(nodeName, parse, parser, scope);

    parse.pushObject(scope);
    parse.pushObject(scopeNode);
    
    try {
      Invoke invoke = new Invoke(UUID.randomUUID().toString());
      setEnclosingScope(invoke, parse);
      parseStandardAttributes(element, invoke, parse);
      parseStandardElements(element, invoke, parse);

      String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
      invoke.setPartnerLink(partnerLinkName);
      invoke.setOperation(XmlUtil.attribute(element, "operation"));
      invoke.setInputVariable(XmlUtil.attribute(element, "inputVariable"));
      invoke.setOutputVariable(XmlUtil.attribute(element, "outputVariable"));

      // get partnerLinkDef and get plnkType attribute and myRole attribute
      PartnerLink partnerLink = findPartnerLinkDefinition(parse,
          partnerLinkName);
      QName plportType = partnerLink.getPartnerRolePortTypeQName();
      invoke.setPortTypeQName(plportType);

      if (XmlUtil.attribute(element, "portType") != null) {
        if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
          // SA00005
          String message = StaticAnalysisFault.getMessage("SA00005",
              "  invoke " + invoke.getName()
              + " has specified a PortType that is different from the one specified by the partnerLink");
          parse.addProblem(message, new StaticAnalysisException("SA00005"));
        }
      }

      // fromParts
      Element fromPartsElement = XmlUtil.element(element, "fromParts");
      List<Element> fromPartElements = null;
      if (fromPartsElement != null) {
        fromPartElements = XmlUtil.elements(fromPartsElement, "fromPart");
        if (invoke.getOutputVariable() != null) {
          // SA00052
          String message = StaticAnalysisFault.getMessage("SA00052",
              "  invoke " + invoke.getName() + " has specified both an outputVariable and a fromPart");
          parse.addProblem(message, new StaticAnalysisException("SA00052"));
        }
      }

      // toParts
      Element toPartsElement = XmlUtil.element(element, "toParts");
      List<Element> toPartElements = null;
      if (toPartsElement != null) {
        toPartElements = XmlUtil.elements(toPartsElement, "toPart");
        if (invoke.getInputVariable() != null) {
          // SA00051
          String message = StaticAnalysisFault.getMessage("SA00051",
              "  invoke " + invoke.getName() + " has specified both an inputVariable and a toPart");
          parse.addProblem(message, new StaticAnalysisException("SA00051"));
        }
      }

      // one way
      invoke.setOneWay(toPartElements == null
          && invoke.getOutputVariable() == null);

      // correlations (needs one way)
      Element correlationsElement = XmlUtil.element(element, "correlations");
      if (correlationsElement != null) {
        List<Element> correlationElements = XmlUtil.elements(correlationsElement, "correlation");
        parseCorrelations(correlationElements, invoke);
      }

      BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
      Operation operation = bpelProcess.getWsdlInfos().getPortType(
          invoke.getPortTypeQName()).getOperation(invoke.getOperation(), null,
              null);
      QName inputMessageQName = operation.getInput().getMessage().getQName();
      QName outputMessageQName = null;
      if (operation.getOutput() != null) {
        outputMessageQName = operation.getOutput().getMessage().getQName();
      }
      // Deploy copy elements for the outbound message
      setOutCopies(toPartElements, invoke, inputMessageQName);
      setInCopies(fromPartElements, invoke, outputMessageQName);



      createActivityNode(element, parse, parser, invoke);

      // Analyse of compensationHandlers
      Element compensationHandlerElement = XmlUtil.element(element, "compensationHandler");
      if (compensationHandlerElement != null) {

        NodeImpl compensationHandlerActivity = this.parseChildActivity(compensationHandlerElement, parse, parser);

        CompensationHandler ch = new CompensationHandler();
        ch.setActivity(compensationHandlerActivity);
        scope.setCompensationHandler(ch);

      }

      // Analyse of Catchs
      List<Catch> catchs = parseAllCatchs(invoke, element, parse, parser);
      scope.setCatchs(catchs);

    } finally {
      parse.popObject();
      parse.popObject();
    }

    return scopeNode;
  }

  private void setOutCopies(List<Element> toPartElements, Invoke invoke,
      QName inputMessageQName) {
    Variable anonymousOutboundMessageVariable = new Variable();
    anonymousOutboundMessageVariable
    .setName(Invoke.OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
    anonymousOutboundMessageVariable.setMessageType(inputMessageQName);
    invoke
    .setAnonymousOutboundMessageVariable(anonymousOutboundMessageVariable);
    String anonymousOutboundMessageVariableName = anonymousOutboundMessageVariable
    .getName();

    String inputVariable = invoke.getInputVariable();
    List<Copy> outCopies = new ArrayList<Copy>();
    if (inputVariable != null) {
      From from = new From();
      from.setVariable(inputVariable);
      from.setVariant(Variant.VAR);
      To to = new To();
      to.setVariable(anonymousOutboundMessageVariableName);
      to.setVariant(Variant.VAR);

      Copy copy = new Copy();
      copy.setFrom(from);
      copy.setTo(to);
      outCopies.add(copy);
    } else {
      // Use the toParts elements
      List<ToPart> toParts = new ArrayList<ToPart>();
      if (toPartElements != null) {
        for (Element toPartElement : toPartElements) {
          ToPart toPart = new ToPart();
          String part = XmlUtil.attribute(toPartElement, "part");
          toPart.setPart(part);
          String fromVariable = XmlUtil
          .attribute(toPartElement, "fromVariable");
          toPart.setFromVariable(fromVariable);
          toParts.add(toPart);
        }
      }

      for (ToPart toPart : toParts) {
        From from = new From();
        from.setVariable(toPart.getFromVariable());
        from.setVariant(Variant.VAR);
        To to = new To();
        to.setVariable(anonymousOutboundMessageVariableName);
        to.setPart(toPart.getPart());
        to.setVariant(Variant.VAR);
        Copy copy = new Copy();
        copy.setFrom(from);
        copy.setTo(to);
        outCopies.add(copy);
      }
    }
    if (outCopies.size() != 0) {
      invoke.setOutCopies(outCopies);
    }
  }

  private void setInCopies(List<Element> fromPartElements, Invoke invoke,
      QName outputMessageQName) {
    String outputVariable = invoke.getOutputVariable();
    List<Copy> inCopies = new ArrayList<Copy>();
    // If the <invoke> is a request/response invocation, also deploy copy
    // elements for the inbound message
    if (!invoke.isOneWay()) {
      Variable anonymousInboundMessageVariable = new Variable();
      anonymousInboundMessageVariable
      .setName(Invoke.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
      anonymousInboundMessageVariable.setMessageType(outputMessageQName);
      invoke
      .setAnonymousInboundMessageVariable(anonymousInboundMessageVariable);
      String anonymousInboundMessageVariableName = anonymousInboundMessageVariable
      .getName();
      if (outputVariable != null) {
        From from = new From();
        from.setVariable(anonymousInboundMessageVariableName);
        from.setVariant(Variant.VAR);
        To to = new To();
        to.setVariable(outputVariable);
        to.setVariant(Variant.VAR);
        Copy copy = new Copy();
        copy.setFrom(from);
        copy.setTo(to);
        inCopies.add(copy);
      } else {
        List<FromPart> fromParts = new ArrayList<FromPart>();
        if (fromPartElements != null) {
          for (Element fromPartElement : fromPartElements) {
            FromPart fromPart = new FromPart();
            String part = XmlUtil.attribute(fromPartElement, "part");
            fromPart.setPart(part);
            String toVariable = XmlUtil
            .attribute(fromPartElement, "toVariable");
            fromPart.setToVariable(toVariable);
            fromParts.add(fromPart);
          }
        }
        // Use the fromParts elements
        for (FromPart fromPart : fromParts) {
          From from = new From();
          from.setVariable(anonymousInboundMessageVariableName);
          from.setPart(fromPart.getPart());
          from.setVariant(Variant.VAR);
          To to = new To();
          to.setVariable(fromPart.getToVariable());
          to.setVariant(Variant.VAR);
          Variable toVariable = invoke.getEnclosingScope().findVariable(
              fromPart.getToVariable());
          if (toVariable.getMessageType() != null) {
            to.setPart(fromPart.getPart());
          }
          Copy copy = new Copy();
          copy.setFrom(from);
          copy.setTo(to);
          inCopies.add(copy);
        }
      }
    }
    if (inCopies.size() != 0) {
      invoke.setInCopies(inCopies);
    }
  }

  /**
   * Parses correlation elements of the given receive.
   *
   * @param correlationElements
   *            the list of all correlation elements of the given receive
   * @param invoke
   *            invoke
   */
  private void parseCorrelations(List<Element> correlationElements,
      Invoke invoke) {
    List<Correlation> requestCorrelations = new ArrayList<Correlation>();
    List<Correlation> responseCorrelations = new ArrayList<Correlation>();
    if (correlationElements != null) {
      for (Element correlationElement : correlationElements) {
        Correlation correlation = new Correlation();
        String set = XmlUtil.attribute(correlationElement, "set");
        Scope enclosingScope = invoke.getEnclosingScope();
        correlation.setCorrelationSet(enclosingScope.findCorrelationSet(set));
        String initiate = XmlUtil.attribute(correlationElement, "initiate");
        correlation.setInitiate(initiate);
        if (invoke.isOneWay()) {
          // TODO: if pattern != null, error during static analysis
          requestCorrelations.add(correlation);
        } else {
          String pattern = XmlUtil.attribute(correlationElement, "pattern");
          // TODO: if pattern == null, error during static analysis
          if ("request".equals(pattern) || "request-response".equals(pattern)) {
            requestCorrelations.add(correlation);
          }
          if ("response".equals(pattern) || "request-response".equals(pattern)) {
            responseCorrelations.add(correlation);
          }
        }
      }
    }
    invoke.setRequestCorrelations(requestCorrelations);
    invoke.setResponseCorrelations(responseCorrelations);
  }
}
