/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.jbpm.pvm.impl.NodeImpl;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Pick;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.OnAlarm;
import org.ow2.orchestra.definition.element.OnMessage;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.w3c.dom.Element;

public class PickBinding extends ActivityBinding {
  /** log. */
  private static Logger log = Logger.getLogger(PickBinding.class.getName());

  public Object parse(Element element, Parse parse, Parser parser) {
    Pick pick = new Pick(UUID.randomUUID().toString());
    setEnclosingScope(pick, parse);
    parseStandardAttributes(element, pick, parse);
    parseStandardElements(element, pick, parse);

    pick.setCreateInstance(BpelUtil.isCreateInstance(element));

    NodeImpl pickNode = createActivityNode(element, parse, parser, pick);

    List<Element> onMessageElements = XmlUtil.elements(element, "onMessage");
    if (onMessageElements == null || onMessageElements.isEmpty()) {
      parse.addProblem("pick element " + element + "should contain at least one onMessage element.");
    }
    List<Element> onAlarmElements = XmlUtil.elements(element, "onAlarm");
    if (onAlarmElements != null && !onAlarmElements.isEmpty() && pick.isCreateInstance()) {
      String message = StaticAnalysisFault.getMessage("SA00062",
          "  pick " + pick + " has a createInstance attribute with a value of yes. It must contain only onMessage events.");
      parse.addProblem(message, new StaticAnalysisException("SA00062"));
    }
    List<OnMessage> onMessages = new ArrayList<OnMessage>();
    List<OnAlarm> onAlarms = null;
    parse.pushObject(pickNode);
    try {
      for (Element onMessageElement : onMessageElements) {
        OnMessage onMessage = parseOnMessage(onMessageElement, pickNode, pick, parse, parser);
        onMessages.add(onMessage);
      }
      if (onAlarmElements != null) {
        onAlarms = new ArrayList<OnAlarm>();
        for (Element onAlarmElement : onAlarmElements) {
          OnAlarm onAlarm = parseOnAlarm(onAlarmElement, pickNode, pick, parse, parser);
          onAlarms.add(onAlarm);
        }
      }
    } finally {
      parse.popObject();
    }
    pick.setOnMessages(onMessages);
    pick.setOnAlarms(onAlarms);
    
    return pickNode;
  }

  public OnAlarm parseOnAlarm(Element element, NodeImpl pickNode, Pick pick, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }  
    OnAlarm onAlarm = new OnAlarm();
    onAlarm.setActivity(parseChildActivity(element, parse, parser));
    
    BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    Element deadlineElement = XmlUtil.element(element, "until");
    Element durationElement = XmlUtil.element(element, "for");
    if (durationElement != null && deadlineElement != null) {
      parse.addProblem("Error in element: " + element + " only one element 'for' or 'until' should be used.");
    }
    if (durationElement != null) {
      onAlarm.setDurationExpression(parseExpression(durationElement, bpelProcess));
    } else if (deadlineElement != null) {
      onAlarm.setDeadlineExpression(parseExpression(deadlineElement, bpelProcess));
    } else {
      parse.addProblem("Error in element: " + element + " one element 'for' or 'until' should be used.");
    }
    return onAlarm;
  }
  
  public OnMessage parseOnMessage(Element element, NodeImpl pickNode, Pick pick, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    OnMessage onMessage = new OnMessage();
    onMessage.setVariable(XmlUtil.attribute(element, "variable"));
    String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    onMessage.setPartnerLinkName(partnerLinkName);
    onMessage.setOperation(XmlUtil.attribute(element, "operation"));

    String messageExchange = XmlUtil.attribute(element, "messageExchange");
    if (messageExchange != null) {
      onMessage.setMessageExchange(messageExchange);
    }

    // get partnerLinkDef and get plnkType attribute and myRole attribute
    PartnerLink partnerLink = findPartnerLinkDefinition(parse,
        partnerLinkName);
    QName plportType = partnerLink.getMyRolePortTypeQName();
    onMessage.setPortTypeQName(plportType);

    if (XmlUtil.attribute(element, "portType") != null) {
      if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
        // SA00005
        String message = StaticAnalysisFault.getMessage("SA00005",
            "  onMessage " + onMessage + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }

    // fromParts
    Element fromPartsElement = XmlUtil.element(element, "fromParts");
    if (fromPartsElement != null) {
      if (onMessage.getVariable() != null) {
        // SA00055
        String message = StaticAnalysisFault.getMessage("SA00063",
            "  onMessage " + onMessage + " has specified both a Variable and a fromPart");
        parse.addProblem(message, new StaticAnalysisException("SA00063"));        
      }
      List<Element> fromPartElements = XmlUtil.elements(fromPartsElement, "fromPart");
      onMessage.setAnonymousInboundMessageVariable(ReceiveBinding.getAnonymousVariable(onMessage, parse));
      List<Copy> inCopies = ReceiveBinding.getInCopies(fromPartElements, pick.getEnclosingScope());
      if (inCopies.size() != 0) {
        onMessage.setInCopies(inCopies);
      }
    }

    // correlations
    Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      List<Element> correlationElements = XmlUtil.elements(correlationsElement, "correlation");
      onMessage.setCorrelations(ReceiveBinding.parseCorrelations(correlationElements, pick.getEnclosingScope()));
    }
    
    BpelProcess bpelProcess = parse.findObject(BpelProcess.class);

    if (pick.isCreateInstance()) {
      OperationKey operationKey = onMessage.getOperationKey(bpelProcess.getQName());
      bpelProcess.addStartElement(operationKey, onMessage, correlationsElement != null);
    }

    onMessage.setActivity(parseChildActivity(element, parse, parser));
    
    return onMessage;
  }



  
  protected static Expression parseExpression(Element element, BpelProcess bpelProcess) {
    String expressionLanguage = XmlUtil.attribute(element, "expressionLanguage");
    if (expressionLanguage == null) {
      expressionLanguage = bpelProcess.getExpressionLanguage();
    }
    Expression expression = new Expression();
    expression.setLanguage(expressionLanguage);
    expression.setNamespaces(BpelUtil.getAllNameSpaces(element));
    expression.setText(element.getTextContent());
    return expression;    
  }
}
