/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Operation;
import javax.xml.namespace.QName;

import org.jbpm.pvm.impl.NodeImpl;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Receive;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.FromPart;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.definition.element.FromTo.Variant;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */
public class ReceiveBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(ReceiveBinding.class.getName());

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *            dom element to parse
   * @param parse
   *            parse
   * @param parser
   *            parser
   * @return the build behaviour!
   */
  public Object parse(Element element, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    Receive receive = new Receive(UUID.randomUUID().toString());
    setEnclosingScope(receive, parse);
    parseStandardAttributes(element, receive, parse);
    parseStandardElements(element, receive, parse);

    BpelProcess bpelProcess = parse.findObject(BpelProcess.class);

    receive.setVariable(XmlUtil.attribute(element, "variable"));
    String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    receive.setPartnerLinkName(partnerLinkName);
    receive.setOperation(XmlUtil.attribute(element, "operation"));
    receive.setCreateInstance(BpelUtil.booleanValue(XmlUtil.attribute(element,
        "createInstance")));
    String messageExchange = XmlUtil.attribute(element, "messageExchange");
    if (messageExchange != null) {
      receive.setMessageExchange(messageExchange);
    }

    // get partnerLinkDef and get plnkType attribute and myRole attribute
    PartnerLink partnerLink = findPartnerLinkDefinition(parse,
        partnerLinkName);
    QName plportType = partnerLink.getMyRolePortTypeQName();
    receive.setPortTypeQName(plportType);

    if (XmlUtil.attribute(element, "portType") != null) {
      if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
        // SA00005
        String message = StaticAnalysisFault.getMessage("SA00005",
            "  receive " + receive.getName()
            + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }

    // fromParts
    Element fromPartsElement = XmlUtil.element(element, "fromParts");
    if (fromPartsElement != null) {
      if (receive.getVariable() != null) {
        // SA00055
        String message = StaticAnalysisFault.getMessage("SA00055",
            "  receive " + receive.getName() + " has specified both a Variable and a fromPart");
        parse.addProblem(message, new StaticAnalysisException("SA00055"));
      }
      List<Element> fromPartElements = XmlUtil.elements(fromPartsElement, "fromPart");
      receive.setAnonymousInboundMessageVariable(getAnonymousVariable(receive, parse));
      List<Copy> inCopies = getInCopies(fromPartElements, receive.getEnclosingScope());
      if (inCopies.size() != 0) {
        receive.setInCopies(inCopies);
      }
    }

    // correlations
    Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      List<Element> correlationElements = XmlUtil.elements(correlationsElement,
          "correlation");
      receive.setCorrelations(parseCorrelations(correlationElements, receive.getEnclosingScope()));
    }

    NodeImpl receiveNode = createActivityNode(element, parse, parser, receive);
    if (BpelUtil.isCreateInstance(element)) {
      OperationKey operationKey = receive.getOperationKey(bpelProcess
          .getQName());
      bpelProcess.addStartElement(operationKey, receive,
          correlationsElement != null);
    }

    return receiveNode;
  }

  /**
   * Parses correlation elements.
   *
   * @param correlationElements
   *            the list of all correlation elements of the given receivingElement
   * @param Scope
   *            enclosing scope of the current activity
   */
  protected static List<Correlation> parseCorrelations(List<Element> correlationElements, Scope enclosingScope) {
    List<Correlation> correlations = new ArrayList<Correlation>();
    if (correlationElements != null) {
      for (Element correlationElement : correlationElements) {
        Correlation correlation = new Correlation();
        String set = XmlUtil.attribute(correlationElement, "set");
        correlation.setCorrelationSet(enclosingScope.findCorrelationSet(set));
        String initiate = XmlUtil.attribute(correlationElement, "initiate");
        correlation.setInitiate(initiate);

        correlations.add(correlation);
      }
    }
    return correlations;
  }

  /**
   * Creates anonymous inbound variable of the receiving element.
   */
  protected static Variable getAnonymousVariable(ReceivingElement receivingElement, Parse parse) {
    BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    Operation operation = bpelProcess.getWsdlInfos().getPortType(
        receivingElement.getPortTypeQName()).getOperation(receivingElement.getOperation(),
            null, null);
    QName messageQName = operation.getInput().getMessage().getQName();

    Variable anonymousInboundMessageVariable = new Variable();
    anonymousInboundMessageVariable.setName(Receive.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
    anonymousInboundMessageVariable.setMessageType(messageQName);
    return anonymousInboundMessageVariable;
  }

  /**
   * Parses inCopies elements
   * @param fromPartElements
   *    the list of all fromParts elements of the given receivingElement
   * @param enclosingScope
   *            enclosing scope of the current activity
   */
  protected static List<Copy> getInCopies(List<Element> fromPartElements, Scope enclosingScope) {
    List<Copy> inCopies = new ArrayList<Copy>();
    if (fromPartElements != null) {
      List<FromPart> fromParts = new ArrayList<FromPart>();
      for (Element fromPartElement : fromPartElements) {
        FromPart fromPart = new FromPart();
        String part = XmlUtil.attribute(fromPartElement, "part");
        fromPart.setPart(part);
        String toVariable = XmlUtil.attribute(fromPartElement, "toVariable");
        fromPart.setToVariable(toVariable);
        fromParts.add(fromPart);
      }

      // Use the fromParts elements
      for (FromPart fromPart : fromParts) {
        From from = new From();
        from.setVariable(Receive.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
        from.setPart(fromPart.getPart());
        from.setVariant(Variant.VAR);
        To to = new To();
        to.setVariable(fromPart.getToVariable());
        to.setVariant(Variant.VAR);
        Variable toVariable = enclosingScope.findVariable(
            fromPart.getToVariable());
        if (toVariable.getMessageType() != null) {
          to.setPart(fromPart.getPart());
        }
        Copy copy = new Copy();
        copy.setFrom(from);
        copy.setTo(to);
        inCopies.add(copy);
      }
    }

    return inCopies;
  }
}
