/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Operation;
import javax.xml.namespace.QName;

import org.jbpm.pvm.impl.NodeImpl;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Reply;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.ToPart;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.definition.element.FromTo.Variant;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */

public class ReplyBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(ReplyBinding.class.getName());

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *            dom element to parse
   * @param parse
   *            parse
   * @param parser
   *            parser
   * @return the build behaviour!
   */
  public Object parse(Element element, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    Reply reply = new Reply(UUID.randomUUID().toString());
    setEnclosingScope(reply, parse);
    parseStandardAttributes(element, reply, parse);
    parseStandardElements(element, reply, parse);

    reply.setVariable(XmlUtil.attribute(element, "variable"));
    String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    reply.setPartnerLinkName(partnerLinkName);
    reply.setOperation(XmlUtil.attribute(element, "operation"));
    QName faultQName = XmlUtil.attributeQName(element, "faultName");
    if (faultQName != null) {
      reply.setFaultQName(faultQName);
      parse
          .addProblem("faultName attribute not yet supported on reply activity");
    }
    String messageExchange = XmlUtil.attribute(element, "messageExchange");
    if (messageExchange != null) {
      reply.setMessageExchange(messageExchange);
    }

    // get partnerLinkDef and get plnkType attribute and myRole attribute
    PartnerLink partnerLink = findPartnerLinkDefinition(parse,
        partnerLinkName);
    QName plportType = partnerLink.getMyRolePortTypeQName();
    reply.setPortTypeQName(plportType);

    if (XmlUtil.attribute(element, "portType") != null) {
      if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
        // SA00005
        String message = StaticAnalysisFault.getMessage("SA00005",
            "  reply " + reply.getName()
            + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }

    // toParts
    Element toPartsElement = XmlUtil.element(element, "toParts");
    if (toPartsElement != null) {
      if (reply.getVariable() != null) {
        // SA00059
        String message = StaticAnalysisFault.getMessage("SA00059",
            "  reply " + reply.getName() + " has specified both a Variable and a toPart");
        parse.addProblem(message, new StaticAnalysisException("SA00059"));
      }
      List<Element> toPartElements = XmlUtil.elements(toPartsElement, "toPart");
      setOutCopies(toPartElements, reply, parse);
    }

    // correlations
    Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      List<Element> correlationElements = XmlUtil.elements(correlationsElement,
          "correlation");
      parseCorrelations(correlationElements, reply);
    }

    NodeImpl replyNode = createActivityNode(element, parse, parser, reply);
    return replyNode;
  }

  private void setOutCopies(List<Element> toPartElements, Reply reply,
      Parse parse) {
    if (toPartElements != null) {
      List<ToPart> toParts = new ArrayList<ToPart>();
      for (Element toPartElement : toPartElements) {
        ToPart toPart = new ToPart();
        String part = XmlUtil.attribute(toPartElement, "part");
        toPart.setPart(part);
        String fromVariable = XmlUtil.attribute(toPartElement, "fromVariable");
        toPart.setFromVariable(fromVariable);
        toParts.add(toPart);
      }
      BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
      Operation operation = bpelProcess.getWsdlInfos().getPortType(
          reply.getPortTypeQName()).getOperation(reply.getOperation(), null,
          null);
      QName messageQName = operation.getOutput().getMessage().getQName();

      Variable anonymousOutboundMessageVariable = new Variable();
      anonymousOutboundMessageVariable
          .setName(Reply.OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
      anonymousOutboundMessageVariable.setMessageType(messageQName);
      reply
          .setAnonymousOutboundMessageVariable(anonymousOutboundMessageVariable);
      String anonymousOutboundMessageVariableName = anonymousOutboundMessageVariable
          .getName();

      List<Copy> outCopies = null;
      for (ToPart toPart : toParts) {
        if (outCopies == null) {
          outCopies = new ArrayList<Copy>();
        }
        From from = new From();
        from.setVariable(toPart.getFromVariable());
        from.setVariant(Variant.VAR);
        To to = new To();
        to.setVariable(anonymousOutboundMessageVariableName);
        to.setPart(toPart.getPart());
        to.setVariant(Variant.VAR);
        Copy copy = new Copy();
        copy.setFrom(from);
        copy.setTo(to);
        outCopies.add(copy);
      }
      reply.setOutCopies(outCopies);
    }
  }

  private void parseCorrelations(List<Element> correlationElements, Reply reply) {
    List<Correlation> correlations = new ArrayList<Correlation>();
    if (correlationElements != null) {
      for (Element correlationElement : correlationElements) {
        Correlation correlation = new Correlation();
        String set = XmlUtil.attribute(correlationElement, "set");
        Scope enclosingScope = reply.getEnclosingScope();
        correlation.setCorrelationSet(enclosingScope.findCorrelationSet(set));
        String initiate = XmlUtil.attribute(correlationElement, "initiate");
        correlation.setInitiate(initiate);

        correlations.add(correlation);
      }
    }
    reply.setCorrelations(correlations);
  }

}
