/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.jbpm.pvm.impl.NodeImpl;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Catch;
import org.ow2.orchestra.definition.element.CompensationHandler;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.definition.element.FromTo.Variant;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.wsdl.PartnerLinkType;
import org.ow2.orchestra.wsdl.Property;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */

public class ScopeBinding extends ActivityBinding {
  /** log. */
  private static Logger log = Logger.getLogger(ScopeBinding.class.getName());

  /**
   * Parses the given dom element to build a Node and a behaviour.
   * @param element dom element to parse
   * @param parse parse
   * @param parser parser
   * @return the build behaviour!
   */
  public Object parse(Element element, Parse parse, Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    Scope scope = new Scope(UUID.randomUUID().toString());
    setEnclosingScope(scope, parse);
    parseStandardAttributes(element, scope, parse);
    parseStandardElements(element, scope, parse);

    WsdlsInfos wsdlInfos = parse.findObject(BpelProcess.class).getWsdlInfos();

    // parse attributes
    String isolatedAttr = XmlUtil.attribute(element, "isolated");
    if (isolatedAttr != null) {
      boolean isolated = BpelUtil.booleanValue(isolatedAttr);
      scope.setIsolated(isolated);
      if (isolated) {
        parse.addProblem("isolated attribute not yet supported on scope activity");
      }
    }

    boolean defaultExitOnStandardFault = false;
    if (scope.getEnclosingScope() != null) {
      defaultExitOnStandardFault = scope.getEnclosingScope().isExitOnStandardFault();
    }
    boolean exitOnStandardFault = BpelUtil.booleanValue(
        XmlUtil.attribute(element, "exitOnStandardFault"), defaultExitOnStandardFault);
    scope.setExitOnStandardFault(exitOnStandardFault);
    if (exitOnStandardFault) {
      parse.addProblem("exitOnStandardFault attribute not yet supported on scope activity");
    }

    // partnerLinks
    Element partnerLinksElement = XmlUtil.element(element, "partnerLinks");
    if (partnerLinksElement != null) {
      List<Element> partnerLinkElements = XmlUtil.elements(partnerLinksElement, "partnerLink");
      parsePartnerLinks(partnerLinkElements, wsdlInfos, scope, parse);
    }

    // messageExchanges
    Element messageExchangesElement = XmlUtil.element(element, "messageExchanges");
    if (messageExchangesElement != null) {
      List<Element> messageExchangeElements = XmlUtil.elements(messageExchangesElement, "messageExchange");
      parseMessageExchanges(messageExchangeElements, scope);
    }

    // variables
    Element variablesElement = XmlUtil.element(element, "variables");
    if (variablesElement != null) {
      List<Element> variableElements = XmlUtil.elements(variablesElement, "variable");
      parseVariables(variableElements, scope, parse);
    }

    // correlationSets
    Element correlationSetsElement = XmlUtil.element(element, "correlationSets");
    if (correlationSetsElement != null) {
      List<Element> correlationSetElements = XmlUtil.elements(correlationSetsElement, "correlationSet");
      parseCorrelationSets(correlationSetElements, scope, wsdlInfos, parse);
    }

    // terminationHandlers
    Element terminationHandlersElement = XmlUtil.element(element, "terminationHandlers");
    if (terminationHandlersElement != null) {
      parse.addProblem("terminationHandlers not yet supported on scope activity");
    }

    // eventHandlers
    Element eventHandlersElement = XmlUtil.element(element, "eventHandlers");
    if (eventHandlersElement != null) {
      parse.addProblem("eventHandlers not yet supported on scope activity");
    }

    NodeImpl scopeNode = createActivityNodeWithChildren(element, parse, parser, scope);
    try {
      parse.pushObject(scope);
      parse.pushObject(scopeNode);

      // faultHandlers
      Element faultHandlersElement = XmlUtil.element(element, "faultHandlers");
      if (faultHandlersElement != null) {
        List<Catch> catchs = parseAllCatchs(scope, faultHandlersElement, parse, parser);
        scope.setCatchs(catchs);
      }

      // compensationHandlers
      Element compensationHandlerElement = XmlUtil.element(element, "compensationHandler");
      if (compensationHandlerElement != null) {
        NodeImpl activity = this.parseChildActivity(compensationHandlerElement, parse, parser);
        CompensationHandler ch = new CompensationHandler();
        ch.setName(scope.getName());
        ch.setActivity(activity);
      }
    } finally {
      parse.popObject();
      parse.popObject();
    }

    return scopeNode;
  }

  /**
   * Parses all partnerLink elements of the fiven scope.
   * @param partnerLinkElements list of all partnerLink elements of this scope
   * @param wsdlInfos wsdlInfos
   * @param scope current scope
   */
  private void parsePartnerLinks(List<Element> partnerLinkElements, WsdlsInfos wsdlInfos, Scope scope, Parse parse) {
    Map<String, PartnerLink> partnerLinks = new HashMap<String, PartnerLink>();
    if (partnerLinkElements != null) {
      for (Element partnerLinkElement : partnerLinkElements) {
        PartnerLink partnerLink = new PartnerLink();
        String name = XmlUtil.attribute(partnerLinkElement, "name");
        if (partnerLinks.containsKey(name)) {
          // Static Analysis SA00018
          String message = StaticAnalysisFault.getMessage("SA00018",
              "  PartnerLink " + name + " defined twice in scope " + scope.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00018"));
        } else {
          partnerLink.setName(name);

          QName partnerLinkTypeName = XmlUtil.attributeQName(partnerLinkElement, "partnerLinkType");
          PartnerLinkType partnerLinkType = wsdlInfos.getPartnerLinkType(partnerLinkTypeName);

          boolean roleDef = false;
          if (partnerLinkType != null) {
            String myRoleName = XmlUtil.attribute(partnerLinkElement, "myRole");
            if (myRoleName != null) {
              QName myRolePortTypeQName = partnerLinkType.getPortTypeQName(myRoleName);
              partnerLink.setMyRolePortTypeQName(myRolePortTypeQName);
              roleDef = true;
            }
            String partnerRoleName = XmlUtil.attribute(partnerLinkElement, "partnerRole");
            if (partnerRoleName != null) {
              QName partnerRolePortTypeQName = partnerLinkType.getPortTypeQName(partnerRoleName);
              partnerLink.setPartnerRolePortTypeQName(partnerRolePortTypeQName);
              roleDef = true;
            }
            String initializePartnerRole = XmlUtil.attribute(partnerLinkElement, "initializePartnerRole");
            if (initializePartnerRole != null) {
              if (partnerRoleName != null) {
                partnerLink.setInitializePartnerRole(initializePartnerRole.equals("true"));
              } else {
                // SA00017
                String message = StaticAnalysisFault.getMessage("SA00017",
                    "  PartnerLink " + name + " has initializePartnerRole attribute used and does not have a partner role.");
                parse.addProblem(message, new StaticAnalysisException("SA00017"));
              }
            }
          }
          if (!roleDef) {
            // SA00016
            String message = StaticAnalysisFault.getMessage("SA00016",
                "  PartnerLink " + name + " should specify myRole or partnerRole or both");
            parse.addProblem(message, new StaticAnalysisException("SA00016"));
          } else {
            partnerLinks.put(name, partnerLink);
          }
        }
      }
    }
    scope.setPartnerLinks(partnerLinks);
  }

  /**
   * Parses all messageExchanges of the given scope.
   * @param messageExchangeElements list of all messageExchange elements of the given scope
   * @param scope scope
   */
  private void parseMessageExchanges(List<Element> messageExchangeElements, Scope scope) {
    Set<String> messageExchanges = new HashSet<String>();
    if (messageExchanges != null) {
      for (Element messageExchangeElement : messageExchangeElements) {
        String name = XmlUtil.attribute(messageExchangeElement, "name");
        messageExchanges.add(name);
      }
    }
    scope.setMessageExchanges(messageExchanges);
  }

  /**
   * Parses all variables of the given scope.
   * @param variableElements list of all variable elements of the given scope
   * @param scope scope
   * @param parse parse
   */
  private void parseVariables(List<Element> variableElements, Scope scope, Parse parse) {
    Map<String, Variable> variables = new HashMap<String, Variable>();
    if (variableElements != null) {
      for (Element variableElement : variableElements) {
        Variable variable = new Variable();
        String name = XmlUtil.attribute(variableElement, "name");
        if (variables.containsKey(name)) {
          // Static Analysis SA00023
          String message = StaticAnalysisFault.getMessage("SA00023",
              "  Variable " + name + " defined twice in scope " + scope.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00023"));
        } else {
          variable.setName(name);
          variable.setMessageType(XmlUtil.attributeQName(variableElement, "messageType"));
          variable.setType(XmlUtil.attributeQName(variableElement, "type"));
          variable.setElement(XmlUtil.attributeQName(variableElement, "element"));

          if (!(((variable.getElement() != null) && (variable.getMessageType() == null)
              && (variable.getType() == null)) || ((variable.getElement() == null)
                  && (variable.getMessageType() != null) && (variable.getType() == null))
                  || ((variable.getElement() == null) && (variable.getMessageType() == null)
                      && (variable.getType() != null)))) {
            // Static Analysis SA00025
            String message = StaticAnalysisFault.getMessage("SA00025",
                "  Type of variable " + name + " not defined properly.");
            parse.addProblem(message, new StaticAnalysisException("SA00025"));
          } else {

            Element fromElement = XmlUtil.element(variableElement, "from");
            if (fromElement != null) {
              Copy virtualCopy = new Copy();
              virtualCopy.setKeepSrcElementName(true);
              AssignBinding.parseFrom(fromElement, virtualCopy, parse, scope);
              variable.setVirtualCopy(virtualCopy);
              To to = new To();
              to.setVariable(name);
              to.setVariant(Variant.VAR);
              virtualCopy.setTo(to);
              scope.addVariableToBeInitialized(name);
            }
            variables.put(name, variable);
          }
        }
      }
    }
    scope.setVariables(variables);
  }

  /**
   * Parses all correlationSet of the given scope.
   * @param correlationSetElements list of all correlationSet elements of the given scope
   * @param scope scope
   * @param wsdlInfos wsdlInfos
   */
  private void parseCorrelationSets(List<Element> correlationSetElements, Scope scope, WsdlsInfos wsdlInfos, Parse parse) {
    Map<String, CorrelationSet> correlationSets = new HashMap<String, CorrelationSet>();
    if (correlationSetElements != null) {
      for (Element correlationSetElement : correlationSetElements) {
        CorrelationSet cs = new CorrelationSet();
        String name = XmlUtil.attribute(correlationSetElement, "name");
        if (correlationSets.containsKey(name)) {
          // Static Analysis SA00044
          String message = StaticAnalysisFault.getMessage("SA00044",
              "  CorrelationSet " + name + " defined twice in scope " + scope.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00044"));
        } else {
          cs.setName(name);
          String propertiesSt = XmlUtil.attribute(correlationSetElement, "properties");
          Map<QName, Property> properties = new HashMap<QName, Property>();
          String[] propertyQNames = propertiesSt.split(" ");
          for (String prop : propertyQNames) {
            prop = prop.trim();
            if (!prop.equals("")) {
              QName propertyQN = XmlUtil.getQNameFromString(correlationSetElement, prop);
              Property property = wsdlInfos.getProperty(propertyQN);
              properties.put(propertyQN, property);
            }
          }
          cs.setProperties(properties);
          correlationSets.put(name, cs);
        }
      }
    }
    scope.setCorrelationSets(correlationSets);
  }



}
