/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Message;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.jbpm.pvm.impl.ExecutionImpl;
import org.jbpm.pvm.impl.ExecutionScopeIterator;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.itf.BpelPersistenceService;
import org.ow2.orchestra.wsdl.PartnerLinkType;
import org.ow2.orchestra.wsdl.Property;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.ow2.orchestra.wsdl.WsdlsInfos;

/**
 * a runtime path of execution.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class BpelExecution extends ExecutionImpl {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /** log. */
  private static Logger log = Logger.getLogger(BpelExecution.class.getName());

  protected List<FlowRuntime> flowRuntimes;

  protected List<ScopeRuntime> scopeRuntimes;

  protected String uuid;

  protected boolean hasPendingMessage;

  protected Set<BpelExecution> waitingExecutions;

  protected BpelExecution() {
    super();
  }

  public BpelExecution(String uuid) {
    this.uuid = uuid;
  }

  protected ExecutionImpl newChildExecution() {
    return new BpelExecution(UUID.randomUUID().toString());
  }

  /**
   * Returns the rootExecution (instance).
   *
   * @return the rootExecution (instance)
   */
  public BpelExecution getProcessInstance() {
    return (BpelExecution) super.getProcessInstance();
  }

  /**
   * Search for the flowRuntime having a link declared with the given name. We
   * start searching into parent and not into this execution as links used in an
   * activity (the current node) must be declared before (in an enclosing flow).
   * So if the current node is flow, we don't want to search into its links. If
   * this parent execution flowRuntime doesn't have one, look into parent
   *
   * @param linkName
   *            linkName
   * @return the flowRuntime having a link declared with the given name.
   */
  public FlowRuntime getFlowRuntimeWithLink(String linkName) {
    Iterator<FlowRuntime> it = new FlowRuntimeIterator(this, linkName);
    if (it.hasNext()) {
      return it.next();
    }
    return null;
  }

  public FlowRuntime getFlowRuntime() {
    Iterator<FlowRuntime> it = new FlowRuntimeIterator(this);
    if (it.hasNext()) {
      return it.next();
    }
    return null;
  }

  public ScopeRuntime getTopLevelScopeRuntime() {
    Iterator<ScopeRuntime> it = new ScopeRuntimeIterator(this);
    ScopeRuntime scopeRuntime = null;
    while (it.hasNext()) {
      scopeRuntime = it.next();
    }
    return scopeRuntime;
  }

  public ScopeRuntime getScopeRuntime() {
    Iterator<ScopeRuntime> it = new ScopeRuntimeIterator(this);
    if (it.hasNext()) {
      return it.next();
    }
    return null;
  }

  public ScopeRuntimeIterator getScopeRuntimeIterator() {
    return new ScopeRuntimeIterator(this);
  }

  public ScopeRuntime getScopeRuntimeWithVariable(String variableName) {
    Iterator<ScopeRuntime> it = new ScopeRuntimeIterator(this, variableName,
        ScopeRuntimeIterator.Type.VAR);
    if (it.hasNext()) {
      return it.next();
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithCS(String csName) {
    Iterator<ScopeRuntime> it = new ScopeRuntimeIterator(this, csName,
        ScopeRuntimeIterator.Type.CS);
    if (it.hasNext()) {
      return it.next();
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithPLNK(String partnerLinkName) {
    Iterator<ScopeRuntime> it = new ScopeRuntimeIterator(this, partnerLinkName,
        ScopeRuntimeIterator.Type.PLNK);
    if (it.hasNext()) {
      return it.next();
    }
    return null;
  }

  /**
   * Set the given linkStatus to the first enclosing flow having a link with the
   * given linkName.
   *
   * @param linkName
   *            linkName of the searched link
   * @param linkStatus
   *            linkStatus to set
   */
  public void setLinkStatus(String linkName, Boolean linkStatus) {
    FlowRuntime fr = getFlowRuntimeWithLink(linkName);
    fr.setLinkStatus(linkName, linkStatus);
  }

  /**
   * Called by a flow when execute method is called by a child (parentNode
   * propagation). This means a flow child is finished. If all flow children are
   * finished, return true. Else return false.
   *
   * @return true if all flow children are finished. Else return false.
   */
  public boolean flowChildFinished() {
    synchronized (MutexRepository.getInstanceMutex(this)) {
      Iterator<FlowRuntime> it = new FlowRuntimeIterator(this);
      if (it.hasNext()) {
        FlowRuntime flowRuntime = it.next();
        if (flowRuntime.childFinished()) {
          it.remove();
          return true;
        }
      }
    }
    return false;
  }

  /**
   * toString method.
   *
   * @return String string representation of this execution.
   */
  public String toString() {
    return "BpelExecution" + this.hashCode();
  }

  /**
   * The process of this execution is guaranteed to be a {@link BpelProcess}.
   */
  @Override
  public BpelProcess getProcessDefinition() {
    return (BpelProcess) processDefinition;
  }

  /**
   * Search in currentScope (and parents) for a partnerLinkRuntime having the
   * given name.
   *
   * @param partnerLinkName
   *            partnerLinkName of the searched partnerLinkRuntime
   * @return the partnerLinkRuntime having the given partnerLinkName in the
   *         closest enclosing scope.
   */
  public PartnerLinkRuntime getPartnerLinkRuntime(String partnerLinkName) {
    ScopeRuntime scopeRuntime = getScopeRuntimeWithPLNK(partnerLinkName);
    if (scopeRuntime != null) {
      return scopeRuntime.getPartnerLink(partnerLinkName);
    }
    return null;
  }

  /**
   *
   * /** Search for a variableRuntime having the given variableName in the
   * closest scope and return its value.
   *
   * @param variableName
   *            variableName
   * @return Search for a variableRuntime having the given variableName in the
   *         closest scope and return its value.
   */
  public Object getVariableValue(String variableName) {
    VariableRuntime variable = getVariableByName(variableName);
    return variable.getValue();
  }

  /**
   * Search for a variableRuntime having the given variableName in the closest
   * scope.
   *
   * @param variableName
   *            variableName
   * @return the found variableRuntime
   */
  public VariableRuntime getVariableByName(String variableName) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Getting variable " + variableName);
    }
    ScopeRuntime currentRuntimeScope = getScopeRuntimeWithVariable(variableName);
    if (currentRuntimeScope != null) {
      return currentRuntimeScope.getVariable(variableName, this);
    }
    return null;
  }

  /**
   * Search for a correlationSetRuntime having the given csName in the closest
   * scope.
   *
   * @param csName
   *            csName
   * @return the found correlationSetRuntime
   */
  public CorrelationSetRuntime getCorrelationSetByName(String csName) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("Getting CorrelationSet " + csName);
    }
    ScopeRuntime scopeRuntime = getScopeRuntimeWithCS(csName);
    if (scopeRuntime != null) {
      return scopeRuntime.getCorrelationSet(csName);
    }
    return null;
  }

  public void addPartnerLinkType(PartnerLinkType partnerLinkType) {
    getWsdlInfos().addPartnerLinkType(partnerLinkType);
  }

  public PartnerLinkType getPartnerLinkType(QName partnerLinkTypeName) {
    return getWsdlInfos().getPartnerLinkType(partnerLinkTypeName);
  }

  public void addPortTypes(Map<QName, PortType> portTypes) {
    getWsdlInfos().addPortTypes(portTypes);
  }

  public PortType getPortType(QName portTypeQName) {
    return getWsdlInfos().getPortType(portTypeQName);
  }

  public void addMessages(Map<QName, Message> messages) {
    getWsdlInfos().addMessages(messages);
  }

  public Message getMessage(QName messageName) {
    return getWsdlInfos().getMessage(messageName);
  }

  public void addServices(Map<QName, Service> services) {
    getWsdlInfos().addServices(services);
  }

  public Service getService(QName serviceName) {
    return getWsdlInfos().getService(serviceName);
  }

  public Service getServiceOfPort(Port port) {
    return getWsdlInfos().getServiceOfPort(port);
  }

  public List<Port> getPorts(QName portTypeQName) {
    return getWsdlInfos().getPorts(portTypeQName);
  }

  public void addProperty(Property property) {
    getWsdlInfos().addProperty(property);
  }

  public Property getProperty(QName propertyQName) {
    return getWsdlInfos().getProperty(propertyQName);
  }

  public Map<String, URL> getStyleSheets() {
    return getProcessDefinition().getStyleSheetRepository()
    .getStyleSheets();
  }

  public void addPropertyAlias(PropertyAlias propertyAlias) {
    getWsdlInfos().addPropertyAlias(propertyAlias);
  }

  public PropertyAlias getPropertyAliasWithMessageType(QName propertyName,
      QName messageType) {
    return getWsdlInfos().getPropertyAliasWithMessageType(propertyName,
        messageType);
  }

  public List<PropertyAlias> getPropertyAliasesWithMessageType(QName messageType) {
    return getWsdlInfos().getPropertyAliasesWithMessageType(messageType);
  }

  public List<PropertyAlias> getPropertyAliases(CorrelationSet cs,
      QName messageType) {
    return getWsdlInfos().getPropertyAliases(cs.getProperties().values(),
        messageType);
  }

  public PropertyAlias getPropertyAliasWithElement(QName propertyName,
      QName element) {
    return getWsdlInfos().getPropertyAliasWithElement(propertyName, element);
  }

  public PropertyAlias getPropertyAliasWithType(QName propertyName, QName type) {
    return getWsdlInfos().getPropertyAliasWithType(propertyName, type);
  }

  public WsdlsInfos getWsdlInfos() {
    return getProcessDefinition().getWsdlInfos();
  }

  public void setWsdlInfos(WsdlsInfos wsdlInfos) {
    getProcessDefinition().setWsdlInfos(wsdlInfos);
  }

  private static class FlowRuntimeIterator extends ExecutionScopeIterator implements
    Iterator<FlowRuntime> {

    private String linkName;

    public FlowRuntimeIterator(BpelExecution execution, String linkName) {
      super(execution);
      this.linkName = linkName;
    }

    public FlowRuntimeIterator(BpelExecution execution) {
      super(execution);
    }

    protected boolean isWanted(Object arg0) {
      if (linkName == null) {
        return true;
      }
      FlowRuntime flowRuntime = (FlowRuntime) arg0;
      return flowRuntime.hasLink(linkName);
    }

    protected List getScopeListFromExecution() {
      return ((BpelExecution) execution).getFlowRuntimes();
    }

    public FlowRuntime next() {
      return (FlowRuntime) super.next();
    }
  }

  public void pushFlowRuntime(FlowRuntime flowRuntime) {
    if (flowRuntimes == null) {
      flowRuntimes = new ArrayList<FlowRuntime>();
    }
    flowRuntimes.add(0, flowRuntime);
  }

  public void popFlowRuntime() {
    if (flowRuntimes != null && !flowRuntimes.isEmpty()) {
      flowRuntimes.remove(0);
    }
  }

  public List<FlowRuntime> getFlowRuntimes() {
    return flowRuntimes;
  }

  public static class ScopeRuntimeIterator extends ExecutionScopeIterator
  implements Iterator<ScopeRuntime> {

    public static enum Type {
      CS, VAR, PLNK, MESSAGE_EXCHANGE
    }

    private String s;

    private Type type;

    public ScopeRuntimeIterator(BpelExecution execution) {
      super(execution);
    }

    public ScopeRuntimeIterator(BpelExecution execution, String s, Type type) {
      super(execution);
      this.s = s;
      this.type = type;
    }

    protected boolean isWanted(Object arg0) {
      if (s == null) {
        return true;
      } else if (type == Type.VAR) {
        return ((ScopeRuntime) arg0).containsVariable(s);
      } else if (type == Type.CS) {
        return ((ScopeRuntime) arg0).containsCorrelationSet(s);
      } else if (type == Type.PLNK) {
        return ((ScopeRuntime) arg0).getPartnerLink(s) != null;
      } else if (type == Type.MESSAGE_EXCHANGE) {
        return ((ScopeRuntime) arg0).getScopeDefinition().getMessageExchanges()
        .contains(s);
      }
      return false;
    }

    protected List getScopeListFromExecution() {
      return ((BpelExecution) execution).getScopeRuntimes();
    }

    public ScopeRuntime next() {
      return (ScopeRuntime) super.next();
    }
  }

  public void pushScopeRuntime(ScopeRuntime scopeRuntime) {
    if (scopeRuntimes == null) {
      scopeRuntimes = new ArrayList<ScopeRuntime>();
    }
    scopeRuntimes.add(0, scopeRuntime);
  }

  public ScopeRuntime popScopeRuntime() {
    ScopeRuntime res = null;
    if (scopeRuntimes != null && !scopeRuntimes.isEmpty()) {
      res = scopeRuntimes.remove(0);
    }
    return res;
  }

  public List<ScopeRuntime> getScopeRuntimes() {
    return scopeRuntimes;
  }

  public String getUuid() {
    return uuid;
  }

  public int hashCode() {
    return uuid.hashCode();
  }

  public boolean equals(Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof BpelExecution)) {
      return false;
    }
    BpelExecution other = (BpelExecution) obj;
    return other.getUuid().equals(uuid);
  }

  public boolean hasPendingMessage() {
    return hasPendingMessage;
  }

  public void setHasPendingMessage(boolean hasPendingMessage) {
    this.hasPendingMessage = hasPendingMessage;
  }

  public Set<BpelExecution> getWaitingExecutions() {
    return waitingExecutions;
  }

  public void addWaitingExecution(BpelExecution execution) {
    if (waitingExecutions == null) {
      waitingExecutions = new HashSet<BpelExecution>();
    }
    waitingExecutions.add(execution);
  }

}
