/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.Correlation.Initiate;
import org.ow2.orchestra.exception.CorrelationViolation;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.var.Message;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * @author Charles Souillard, Guillaume Porcher
 */
public class CorrelationSetRuntime {

  /** cs name. */
  protected String name;

  /** cs definition (the same for all instances of the process). */
  protected CorrelationSet correlationSetDefinition;

  /** map of runtime values of properties used by this cs. */
  protected Map<QName, Node> propertyValues;

  /** is cs initialized */
  protected boolean initialized = false;

  protected CorrelationSetRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param correlationSetDefinition
   *            correlationSetDefinition
   */
  public CorrelationSetRuntime(CorrelationSet correlationSetDefinition) {
    this.name = correlationSetDefinition.getName();
    this.correlationSetDefinition = correlationSetDefinition;
    this.propertyValues = new HashMap<QName, Node>();
  }

  /**
   * Return the runtime value of the property identified by the given
   * propertyName.
   *
   * @param propertyName
   *            propertyName
   * @return the runtime value of the property identified by the given
   *         propertyName.
   */
  public Node getPropertyValue(QName propertyName) {
    return propertyValues.get(propertyName);
  }

  /**
   * Set the runtime value of the property identified by the given propertyName.
   *
   * @param propertyName
   *            propertyName
   * @param value
   *            value to set
   */
  public void setPropertyValue(QName propertyName, Node value) {
    propertyValues.put(propertyName, value);
  }

  protected boolean validatePropertyValue(PropertyAlias propertyAlias, Message message, BpelExecution execution) {
    try {
      Node expectedValue = this.getPropertyValue(propertyAlias.getPropertyName());

      Element partElement = message.getPartValue(propertyAlias.getPart());
      if (partElement == null) {
        return expectedValue == null;
      }

      Node value = partElement;
      if (propertyAlias.getQuery() != null) {
        value = propertyAlias.getQuery().getEvaluator().evaluate(partElement, execution);
      }

      if (value == null) {
        return expectedValue == null;
      } else if (expectedValue == null) {
        return false;
      }
      return value.isEqualNode(expectedValue);
    } catch (Exception ex) {
      throw new OrchestraException("In Correlation: Function call exception:" + ex, ex);
    }
  }
  
  public boolean validateCSValues(BpelExecution execution, final Message message,
      final QName messageType) {
    List<PropertyAlias> propertyAliases = correlationSetDefinition.getPropertyAliasRefs(execution, messageType);

    //foreach propertyAlias, we extract the value from the message.
    for (PropertyAlias propertyAlias : propertyAliases) {
      if (!validatePropertyValue(propertyAlias, message, execution)) {
        return false;
      }
    }
    return true;
  }

  public void checkCSValues(BpelExecution execution, final Message message,
      final QName messageType, final Correlation.Initiate initiate) {
    // initialize CS
    if (Initiate.yes.equals(initiate) || !isInitialized()) {
      initiate(execution, message, messageType, initiate);
    } else {
      // check CS values
      if (!validateCSValues(execution, message, messageType)) {
        throw new CorrelationViolation("Correlation consistency constraint violated");
      }
    }
  }

  public synchronized void initiate(BpelExecution execution, final Message message,
      final QName messageType, final Correlation.Initiate initiate) {

    if (isInitialized()) {
      if (Initiate.yes.equals(initiate)) {
        throw new CorrelationViolation("CorrelationSet already initialized and initiate='yes'");
      }
      return;
    }

    if (Initiate.no.equals(initiate)) {
      throw new CorrelationViolation("CorrelationSet not initialized and initiate='no'");
    }
    
    List<PropertyAlias> propertyAliases = execution.getPropertyAliases(correlationSetDefinition, messageType);
    //TODO : see if we can have an Element type message

    //foreach propertyAlias, we extract the value from the message.
    for (PropertyAlias propertyAlias : propertyAliases) {
      try {
        Element partElement = message.getPartValue(propertyAlias.getPart());

        Node value = partElement;
        if (propertyAlias.getQuery() != null) {
          value = propertyAlias.getQuery().getEvaluator().evaluate(partElement, execution);
        }
        this.setPropertyValue(propertyAlias.getPropertyName(), value);
      } catch (Exception ex) {
        throw new OrchestraException("In Correlation: Function call exception:" + ex, ex);
      }
    }
    this.setInitialized(true);
  }

  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!obj.getClass().equals(CorrelationSetRuntime.class)) {
      return false;
    }
    CorrelationSetRuntime corr = (CorrelationSetRuntime) obj;
    return this.name.equals(corr.getName())
      && this.propertyValues.equals(corr.getPropertyValues());
  }

  public int hashCode() {
    String result = "";
    result += "name=\"" + this.name + "\" properties=\"";
    for (QName property : propertyValues.keySet()) {
      result += property;
    }
    result += "\"";

    return result.hashCode();
  }

  public Map<QName, Node> getPropertyValues() {
    return propertyValues;
  }

  public void setPropertyValues(Map<QName, Node> propertyValues) {
    this.propertyValues = propertyValues;
  }

  public String getName() {
    return name;
  }

  public void setName(String name) {
    this.name = name;
  }

  public CorrelationSet getCorrelationSetDefinition() {
    return correlationSetDefinition;
  }

  public void setCorrelationSetDefinition(
      CorrelationSet correlationSetDefinition) {
    this.correlationSetDefinition = correlationSetDefinition;
  }
  public boolean isInitialized() {
    return initialized;
  }
  public void setInitialized(boolean initialized) {
    this.initialized = initialized;
  }
}
