/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * runtime representation of one partner for a {@link PartnerLinkRole}.
 *
 * @author Tom Baeyens, Charles SOuillard
 */
public class PartnerLinkRuntime {

  /** log. */
  private static Logger log = Logger.getLogger(PartnerLinkRuntime.class
      .getName());

  protected long dbid;

  /** name of this partnerLink. */
  protected String name;

  /** definition of this partnerLink. */
  protected PartnerLink partnerLinkDefinition;

  /**
   * The value of the endpointReference associated to the partnerRole of the
   * partnerLink. It is a mechanism that allows you to select one of the
   * available services in the WSDL, or even define new services at runtime.
   */
  private Element partnerRoleEndPointReference;

  /**
   * The value of the endpointReference associated to the myRole of the
   * partnerLink. It is a mechanism that allows you to select one of the
   * available services in the WSDL, or even define new services at runtime.
   */
  private Element myRoleEndPointReference;

  protected PartnerLinkRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param partnerLinkDefinition
   *            partnerLinkDefinition of this partnerLinkRuntime.
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to.
   */
  public PartnerLinkRuntime(PartnerLink partnerLinkDefinition,
      BpelExecution execution) {
    this.name = partnerLinkDefinition.getName();
    this.partnerLinkDefinition = partnerLinkDefinition;
    if (partnerLinkDefinition.getMyRolePortTypeQName() != null) {
      initializeMyRole(execution);
    }
    if (partnerLinkDefinition.getPartnerRolePortTypeQName() != null) {
      initializePartnerRole(execution);
    }
  }

  /**
   * Initialize partnerRole endpoint reference to the defaultPort value.
   *
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to
   */
  @SuppressWarnings("unchecked")
  public void initializePartnerRole(BpelExecution execution) {
    Port defaultPort = null;
    final QName expectedPortTypeQName = partnerLinkDefinition.getPartnerRolePortTypeQName();
    WsdlsInfos wsdlsInfos = execution.getProcessDefinition().getWsdlInfos();
    final List<Port> ports = wsdlsInfos.getPorts(partnerLinkDefinition.getPartnerRolePortTypeQName());

    try {
      final Map<Port, SOAPAddress> supported = new HashMap<Port, SOAPAddress>();
      for (Port port : ports) {
        final List<ExtensibilityElement> elements = port.getExtensibilityElements();
        if (elements.size() > 0) {
          for (ExtensibilityElement element : elements) {
            if (!(element instanceof SOAPAddress)) {
              log.warning("The extensibility element: " + element.getElementType() + " is currently not supported!");
            } else if (port.getBinding().getPortType().getQName().equals(expectedPortTypeQName)) {
              supported.put(port, (SOAPAddress) element);
            }
          }
        } else {
          supported.put(port, null);
        }
      }

      SOAPAddress soapAddress = null;
      Service service = null;
      if (supported.entrySet() != null && supported.entrySet().size() > 0) {
        Map.Entry<Port, SOAPAddress> entry = supported.entrySet().iterator().next();
        defaultPort = entry.getKey();
        soapAddress = entry.getValue();

        log.finest("Using the supported port: " + defaultPort);

        service = execution.getServiceOfPort(defaultPort);
      }

      if (service != null) {
        final Document document = BpelUtil.getServiceRefDocument(soapAddress, service.getQName());
        partnerRoleEndPointReference = document.getDocumentElement();
      }
    } catch (Exception e) {
      if (ports == null
          || defaultPort == null) {
        log.fine("You have to defined a defaut port for partnerRole, which bind the portType of the partnerLink");
      } else {
        // Always log exceptions !! Especially, since Exception covers all sort
        // of exceptions,
        // such as IllegalArgumentException, NullPointerException and the like.
        // TODO: Maybe, we should not catch Exception, but one subtype (e.g
        // OrchestraException?)
        log.log(Level.WARNING, "Exception thrown during initializePartnerRole()", e);
        throw new OrchestraException(e);
      }

    }
  }

  /**
   * Initialize myRole endpoint reference to the defaultPort value.
   *
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to
   */
  @SuppressWarnings("unchecked")
  public void initializeMyRole(BpelExecution execution) {
    Port defaultPort = null;
    WsdlsInfos wsdlsInfos = execution.getProcessDefinition().getWsdlInfos();
    try {
      defaultPort = wsdlsInfos.getPorts(partnerLinkDefinition.getMyRolePortTypeQName()).get(0);
      Service service = execution.getServiceOfPort(defaultPort);
      List<ExtensibilityElement> elements = (List<ExtensibilityElement>) defaultPort
      .getExtensibilityElements();
      SOAPAddress soapAddress = (SOAPAddress) BpelUtil
      .findExtensibilityElement(elements, SOAPAddress.class);
      myRoleEndPointReference = BpelUtil.getServiceRefDocument(soapAddress,
          service.getQName()).getDocumentElement();
    } catch (Exception e) {
      if (wsdlsInfos.getPorts(partnerLinkDefinition.getMyRolePortTypeQName()) == null || defaultPort == null) {
        if (log.isLoggable(Level.FINE)) {
          log
          .fine("You have to defined a defaut port myRole, which bind the portType of the partnerLink");
        }
      }
      // TODO : see if we have to signal the error
    }
  }

  public Element getMyRoleEndPointReference() {
    return this.myRoleEndPointReference;
  }

  public void setMyRoleEndPointReference(Element myRoleEndPointReference) {
    this.myRoleEndPointReference = myRoleEndPointReference;
  }

  public Element getPartnerRoleEndPointReference() {
    return this.partnerRoleEndPointReference;
  }

  public void setPartnerRoleEndPointReference(
      Element partnerRoleEndPointReference) {
    this.partnerRoleEndPointReference = partnerRoleEndPointReference;
  }

  public String getName() {
    return this.name;
  }

  public PartnerLink getPartnerLinkDefinition() {
    return this.partnerLinkDefinition;
  }

  public long getDbid() {
    return dbid;
  }

}
