package org.ow2.orchestra.services;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.env.Environment;
import org.jbpm.env.EnvironmentFactory;
import org.jbpm.pvm.Activity;
import org.jbpm.tx.Transaction;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.exception.BpelFaultException;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.impl.ReplierImpl;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.var.Message;

/**
 * base implementation for the services to be published for a deployed bpel
 * process.
 *
 * @author Tom Baeyens, Charles Souillard, Guillaume Porcher
 */
public class Receiver {

  private static class ExecElementToSignal {
    private BpelExecution bpelExecution;
    private ReceivingElement receivingElement;

    public BpelExecution getBpelExecution() {
      return bpelExecution;
    }
    public ReceivingElement getReceivingElement() {
      return receivingElement;
    }
    public ExecElementToSignal(BpelExecution bpelExecution,
        ReceivingElement receivingElement) {
      this.bpelExecution = bpelExecution;
      this.receivingElement = receivingElement;
    }
  }

  /** log. */
  private static Logger log = Logger.getLogger(Receiver.class.getName());

  protected long dbid;

  protected int dbversion;

  /**
   * maps ReceiveElementKeys(processQName, portTypeQName, operationName) to a
   * list of executions that are waiting for an incoming message.
   */
  protected Set<BpelExecution> waitingExecutions;

  /**
   * messages stored when there is no execution yet that matches this incoming
   * message.
   */
  protected List<PendingMessage> storedMessages;

  protected OperationKey operationKey;

  protected Receiver() {
    super();
  }

  public Receiver(OperationKey operationKey) {
    this.operationKey = operationKey;
  }

  /**
   * Handle incoming message for the given operation key. Use the given
   * {@link MessageCarrier} to store the reply.
   *
   * @param incomingMessage
   * @param operationKey
   * @param environmentFactory
   * @param messageCarrier
   * @return the execution to which the message was given. <code>null</code>
   *         if there is no execution waiting for this message (the message is
   *         stored).
   */
  public static BpelExecution handle(Message incomingMessage,
      OperationKey operationKey, EnvironmentFactory environmentFactory,
      MessageCarrier messageCarrier) {
    Environment environment = null;
    ExecElementToSignal execToSignal = null;
    // Step 1 : give the message to an existing instance or create a new
    // instance
    // This step is synchronized on the process
    synchronized (MutexRepository.getProcessMutex(operationKey
        .getProcessQName())) {
      try {
        environment = environmentFactory.openEnvironment();
        Publisher servicePublisher = (Publisher) environment
            .get(Publisher.class);
        Receiver receiver = (Receiver) servicePublisher
            .getReceiver(operationKey);
        if (receiver == null) {
          throw new OrchestraException("impossible to find a receiver corresponding to: " + operationKey);
        }
        execToSignal = receiver.handle(incomingMessage, environment,
            messageCarrier);
      } catch (Exception e) {
        if (environment != null) {
          Transaction t = environment.getTransaction();
          if (t != null) {
            t.setRollbackOnly();
          }
        }
        if (e instanceof BpelFaultException) {
          throw (BpelFaultException) e;
        }
        throw new OrchestraException(e);
      } finally {
        if (environment != null) {
          environment.close();
        }
      }
    }
    // Step 2 : continue execution
    // Synchronize on the instance to prevent concurrent modification in
    // separate transactions
    if (execToSignal != null) {
      BpelExecution exec = execToSignal.getBpelExecution();
      synchronized (MutexRepository.getInstanceMutex(exec)) {
        try {
          environment = environmentFactory.openEnvironment();
          exec = environment.get(Repository.class).getUpdatedExecution(exec);
          Map<String, Object> parameters = new HashMap<String, Object>();
          parameters.put("receivingElement",
              environment.get(Repository.class).getUpdatedReceivingElement(execToSignal.getReceivingElement()));
          exec.signal(parameters);
        } catch (Exception e) {
          if (environment != null) {
            Transaction t = environment.getTransaction();
            if (t != null) {
              t.setRollbackOnly();
            }
          }
          if (e instanceof BpelFaultException) {
            throw (BpelFaultException) e;
          }
          throw new OrchestraException(e);
        } finally {
          if (environment != null) {
            environment.close();
          }
        }
      }
      ExecElementToSignal execElt = execToSignal;
      // Step 3 : for each child execution, verify that there is no message
      // available.
      // This step is synchronized on the process as we modify the receiver
      // (shared between all instances of a process)
      while (execElt != null && !execElt.getBpelExecution().isEnded()) {
        synchronized (MutexRepository.getProcessMutex(operationKey
            .getProcessQName())) {
          synchronized (MutexRepository.getInstanceMutex(execElt.getBpelExecution())) {
            try {
              BpelExecution execution = execElt.getBpelExecution();
              environment = environmentFactory.openEnvironment();
              execution = environment.get(Repository.class).getUpdatedExecution(execution);
              if (execution != null) {
                execElt = findExecutionToSignal(execution.getProcessInstance());
              } else {
                execElt = null;
              }
            } catch (Exception e) {
              if (environment != null) {
                Transaction t = environment.getTransaction();
                if (t != null) {
                  t.setRollbackOnly();
                }
              }
              if (e instanceof BpelFaultException) {
                throw (BpelFaultException) e;
              }
              throw new OrchestraException(e);
            } finally {
              if (environment != null) {
                environment.close();
              }
            }
          }
        }
        if (execElt != null) {
          // Step 4: if a message can be associated with an execution, execute
          // the instance
          // (as for step 2, this is synchronized on the instance)
          BpelExecution execution = execElt.getBpelExecution();
          synchronized (MutexRepository.getInstanceMutex(execution)) {
            try {
              environment = environmentFactory.openEnvironment();
              execution = environment.get(Repository.class)
                  .getUpdatedExecution(execution);
              Map<String, Object> parameters = new HashMap<String, Object>();
              parameters.put("receivingElement",
                  environment.get(Repository.class).getUpdatedReceivingElement(execElt.getReceivingElement()));
              execution.signal(parameters);
            } catch (Exception e) {
              if (environment != null) {
                Transaction t = environment.getTransaction();
                if (t != null) {
                  t.setRollbackOnly();
                }
              }
              if (e instanceof BpelFaultException) {
                throw (BpelFaultException) e;
              }
              throw new OrchestraException(e);
            } finally {
              if (environment != null) {
                environment.close();
              }
            }
          }
        }
      }
      return execToSignal.getBpelExecution();
    }

    return null;
  }

  /**
   * Find if there is a stored message for one of the executions. If such an
   * execution is found, associate the stored message to this execution, and
   * return this execution.
   *
   * @param parent
   *            execution to check (recursive search in the child executions)
   * @return first execution found with a matching message.
   */
  private static ExecElementToSignal findExecutionToSignal(BpelExecution parent) {
    if (parent == null || parent.isEnded()) {
      return null;
    }
    Environment environment = Environment.getCurrent();
    BpelExecution instance = parent.getProcessInstance();
    ExecElementToSignal execToSignal = null;
    Set<BpelExecution> waitingExec = new HashSet<BpelExecution>(instance
        .getWaitingExecutions());
    if (waitingExec != null) {
      for (BpelExecution child : waitingExec) {
        Activity activity = child.getNode().getBehaviour();
        if (activity instanceof InboundMessageElement) {
          for (ReceivingElement receivingElement : ((InboundMessageElement)activity).getReceivingElements()) {
            OperationKey opKey = receivingElement.getOperationKey(child
                .getProcessDefinition().getQName());
            Receiver receiver = environment.get(Repository.class).getReceiver(
                opKey);
            receiver.addWaitingExecution(child);
            if (execToSignal == null) {
              Message message = receiver.getMessageFromStoredMessage(child,
                  receivingElement);
              if (message != null) {
                if (receiver.getWaitingExecution(child, receivingElement) == null) {
                  throw new OrchestraException("No such waiting execution : "
                      + child);
                }
                associateMessage(child, receivingElement, message);
                execToSignal = new ExecElementToSignal(child, receivingElement);
              }
            }
          }
        }
      }
    }
    return execToSignal;
  }

  /**
   * Add all executions of the given instance that are waiting on a receiving
   * activity to the list waiting activities of the matching receiver.
   *
   * @param parent
   *            the execution
   */
  private static void addWaitingExecutions(BpelExecution parent) {
    if (parent == null) {
      return;
    }
    Environment environment = Environment.getCurrent();
    BpelExecution instance = parent.getProcessInstance();
    Set<BpelExecution> waitingExec = new HashSet<BpelExecution>(instance
        .getWaitingExecutions());
    if (waitingExec != null) {
      for (BpelExecution child : waitingExec) {
        Activity activity = child.getNode().getBehaviour();
        if (activity instanceof InboundMessageElement) {
          InboundMessageElement inboundMessageElement = (InboundMessageElement) activity;
          for (ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
            OperationKey opKey = receivingElement.getOperationKey(child
                .getProcessDefinition().getQName());
            Receiver receiver = environment.get(Repository.class).getReceiver(
                opKey);
            receiver.addWaitingExecution(child);
          }
        }
      }
    }
  }

  /**
   * is called when a message arrives in the engine. Try to find an execution
   * which is waiting for this kind of message (it means operationKeys matches
   * and if the waiting execution activity has correlation check if correlations
   * are OK. If no matching execution was found, try to use the given message to
   * create a new instance of this process if this message can be given to a
   * startActivty being createInstance=true Else, store the message so that it
   * can be used later when an instance will need it. If the given replier is
   * not null, associate it to the corresponding runtimeScope (top level scope
   * if messageExchange is not specified, scopeRuntime declaring the used
   * messageExchange if there is one.
   *
   * @param incomingMessage
   * @param environment
   */

  private ExecElementToSignal handle(Message incomingMessage,
      Environment environment, MessageCarrier messageCarrier) {
    ExecElementToSignal execElementToSignal = null;

    Repository repository = environment.get(Repository.class);
    BpelProcess bpelProcess = repository.getProcess(operationKey
        .getProcessQName());

    // find a waiting execution using the correlations
    execElementToSignal = getWaitingExecution(incomingMessage, environment);
    // if an execution was waiting for such an incoming message
    if (execElementToSignal != null) {
      addReplier(execElementToSignal, environment, messageCarrier);
      return execElementToSignal;
    } else {
      ReceivingElement startElement = bpelProcess.getStartElement(operationKey);
      // if the incoming message matched a start activity
      if (startElement != null) {
        if (log.isLoggable(Level.FINE)) {
          log.fine("found start node " + startElement + " for incoming msg");
        }
        BpelExecution bpelExecution = (BpelExecution) bpelProcess.startExecution();
        addWaitingExecutions(bpelExecution);
        ExecElementToSignal receiveExecution = getWaitingExecution(bpelExecution,
            startElement);
        if (receiveExecution == null) {
          throw new OrchestraException(
              "No waiting execution found !, process = "
                  + bpelExecution.getProcessDefinition().getQName());
        }
        associateMessage(receiveExecution.getBpelExecution(), startElement, incomingMessage);
        addReplier(receiveExecution, environment, messageCarrier);
        return receiveExecution;
        // initiaceCS

      } else { // no start node was found
        storeIncomingMessage(incomingMessage, messageCarrier);
        return null;
      }
    }
  }

  /**
   * Associate the given replier to the corresponding runtimeScope (top level
   * scope if messageExchange is not specified, scopeRuntime declaring the used
   * messageExchange if there is one.
   */
  private void addReplier(ExecElementToSignal elementToSignal, Environment environment,
      MessageCarrier messageCarrier) {
    if (messageCarrier != null) {
      ReceivingElement receivingElement = elementToSignal.getReceivingElement();
      BpelExecution bpelExecution = elementToSignal.getBpelExecution();
      String messageExchange = receivingElement.getMessageExchange();
      ReplierKey replierKey = new ReplierKey(operationKey, messageExchange,
          bpelExecution.getScopeRuntime());
      Replier replier = createReplier(replierKey);
      MessageCarrierRepository.addMessageCarrier(replierKey, messageCarrier);
      environment.get(Repository.class).storeReplier(replier);
    }
  }

  /**
   * Search for an execution of the given process waiting for a message on the
   * given portType/operation. I one was found and the cs matches (if any),
   * associate this execution to the given message and return this execution.
   *
   * @param incomingMessage
   *            message received we want to give to an execution
   * @param operationKey
   *            operationKey
   * @param environment
   *            environment
   * @return the found execution or null if no one was found
   */
  private ExecElementToSignal getWaitingExecution(Message incomingMessage,
      Environment environment) {
    if (waitingExecutions == null) {
      return null;
    }
    for (BpelExecution bpelExecution : waitingExecutions) {
      // ReceivingElement receivingElement =
      // waitingExecution.getReceivingElement();
      InboundMessageElement inboundMessageElement = (InboundMessageElement) bpelExecution
          .getNode().getBehaviour();
      for (ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
        if (receivingElement.getOperationKey(bpelExecution.getProcessDefinition().getQName()).equals(this.operationKey)) {
          if (csMatch(incomingMessage, receivingElement, bpelExecution)) {
            if (!waitingExecutions.remove(bpelExecution)) {
              throw new OrchestraException("Impossible to remove candidate : "
                + bpelExecution);
            }
            associateMessage(bpelExecution, receivingElement, incomingMessage);
            return new ExecElementToSignal(bpelExecution, receivingElement);
          }
        }
      }
    }
    return null;
  }

  /**
   * Search for an execution of the given process instance waiting for a message
   * on the given node. This method is used to associate an incoming message
   * which creates a new instance to the newly created instance.
   *
   * @return the found execution or null if no one was found
   */
  private ExecElementToSignal getWaitingExecution(BpelExecution parent,
      ReceivingElement startElement) {
    if (waitingExecutions == null) {
      return null;
    }
    for (BpelExecution bpelExecution : waitingExecutions) {
      if (bpelExecution.getProcessInstance().equals(parent)) {
        Activity activity = bpelExecution.getNode().getBehaviour();
        if (activity instanceof InboundMessageElement) {
          InboundMessageElement inboundMessageElement =  (InboundMessageElement) activity;
          if (inboundMessageElement.getReceivingElements().contains(startElement)) {
            if (!waitingExecutions.remove(bpelExecution)) {
              throw new OrchestraException("Impossible to remove candidate : "
                  + bpelExecution);
            }
            return new ExecElementToSignal(bpelExecution, startElement);
          }
        }
      }
    }
    return null;
  }

  /**
   * Associate the given message to the given execution and startActivity.
   */
  private static void associateMessage(BpelExecution bpelExecution,
      ReceivingElement receivingElement, Message message) {
    bpelExecution.setHasPendingMessage(true);
    bpelExecution.getProcessInstance().getWaitingExecutions().remove(
        bpelExecution);
    receivingElement.receive(message, bpelExecution);

    if (log.isLoggable(Level.FINE)) {
      log.fine(Receiver.class + ".associateMessage, message = " + message);
    }
  }

  /**
   * Store the given message as a message that could be consumed by an activity
   * waiting for a message on the given portType/operation.
   *
   * @param message
   *            message to store
   * @param operationName
   *            operation
   * @param portType
   *            portType
   * @param environment
   *            environment
   *
   */
  private void storeIncomingMessage(Message message,
      MessageCarrier messageCarrier) {
    if (storedMessages == null) {
      storedMessages = new ArrayList<PendingMessage>();
    }
    PendingMessage pendingMessage = null;
    if (messageCarrier != null) {
      MessageCarrierRepository.addPendingMessageCarrier(messageCarrier);
      pendingMessage = new PendingMessage(message, messageCarrier.getUuid());
    } else {
      pendingMessage = new PendingMessage(message, null);
    }
    storedMessages.add(pendingMessage);
  }

  // ///////////////////////////////////////////////////////////////////////////////////////

  /**
   * Return true if the given message match the given startActivity for the
   * given execution. It takes in charge cs verification.
   *
   * @param messageToAnalyse
   *            messageToAnalyse
   * @param receivingElement
   *            startActivity
   * @param bpelExecution
   *            bpelExecution
   * @return true if the given message match the given startActivity for the
   *         given execution.
   */
  private boolean csMatch(Message messageToAnalyse,
      ReceivingElement receivingElement, BpelExecution bpelExecution) {
    if (receivingElement.getCorrelations() == null) {
      return true;
    }
    // we calcul the values for each correlationSet
    for (Correlation correlation : receivingElement.getCorrelations()) {
      if (!correlation.correlationOK(messageToAnalyse, receivingElement,
          bpelExecution)) {
        return false;
      }
    }
    return true;

  }

  /**
   * Search for a stored message matching the given parameters.
   *
   * @param bpelExecution
   *            bpelExecution
   * @param receivingElement
   *            startActivity
   * @return a stored message matching the given parameters. or null.
   */
  private Message getMessageFromStoredMessage(BpelExecution bpelExecution,
      ReceivingElement receivingElement) {
    if (storedMessages == null) {
      return null;
    }
    for (PendingMessage pendingMessage : storedMessages) {
      Message message = pendingMessage.getMessage();
      if (csMatch(message, receivingElement, bpelExecution)) {
        MessageCarrier messageCarrier = MessageCarrierRepository
            .removePendingMessageCarrier(pendingMessage.getMessageCarrierId());
        addReplier(new ExecElementToSignal(bpelExecution, receivingElement), Environment.getCurrent(),
            messageCarrier);
        storedMessages.remove(pendingMessage);
        return message;
      }
    }
    return null;
  }

  /**
   * Add a waiting execution in the the list.
   *
   * @param bpelExecution
   *            bpelExecution
   * @param receivingElement
   *            startActivity
   */
  private void addWaitingExecution(BpelExecution bpelExecution) {
    synchronized (MutexRepository.getProcessMutex(bpelExecution.getProcessDefinition()
        .getQName())) {
      if (waitingExecutions == null) {
        waitingExecutions = new HashSet<BpelExecution>();
      }
      waitingExecutions.add(bpelExecution);
    }
  }

  public void removeCancelledExecution(BpelExecution bpelExecution) {
    synchronized (MutexRepository.getProcessMutex(bpelExecution.getProcessDefinition()
        .getQName())) {
      if (waitingExecutions == null) {
        return;
      }
      waitingExecutions.remove(bpelExecution);
    }
  }

  public OperationKey getOperationKey() {
    return this.operationKey;
  }

  private Replier createReplier(ReplierKey replierKey) {
    return new ReplierImpl(replierKey);
  }

}
