/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.impl;

import java.util.List;

import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.jbpm.pvm.Execution;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Receive;
import org.ow2.orchestra.definition.element.OnMessage;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.ReplierKey;
import org.ow2.orchestra.services.itf.BpelPersistenceService;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.services.itf.Repository;

public class DbRepository implements Repository {

  /*
   * PROCESS REPOSITORY
   */
  public BpelProcess getProcess(QName processQName) {
    return getBpelPersistenceService().findBpelProcess(
        processQName.getNamespaceURI(), processQName.getLocalPart());
  }

  public List<BpelProcess> getProcesses() {
    return getBpelPersistenceService().getBpelProcesses();
  }

  public BpelProcess removeProcess(QName processQName) {
    MutexRepository.removeProcessMutex(processQName);
    BpelProcess process = getBpelPersistenceService().findBpelProcess(
        processQName.getNamespaceURI(), processQName.getLocalPart());
    if (process != null) {
      getBpelPersistenceService().delete(process);
    }
    return process;
  }

  public void storeProcess(BpelProcess process) {
    getBpelPersistenceService().save(process);
  }

  /*
   * INSTANCE REPOSITORY
   */
  public List<BpelExecution> getInstances() {
    return getBpelPersistenceService().getBpelInstances();
  }

  public List<BpelExecution> getInstances(QName processQName) {
    return getBpelPersistenceService().getBpelInstances(processQName);
  }

  public void removeInstance(BpelExecution instance) {
    removeInstance(instance, false); 
  }
  
  public void removeInstance(BpelExecution instance, boolean force) {
    instance = getBpelPersistenceService().get(BpelExecution.class,
        instance.getDbid());
    if (force) {
      List<Receiver> receivers = getReceivers(instance.getProcessDefinition().getQName());
      if (receivers != null) {
        for (Receiver receiver : receivers) {
          receiver.removeCancelledExecution(instance);
        }
      }
      removeRepliers(instance);
    }
    if (instance != null) {
      getBpelPersistenceService().delete(instance.getProcessInstance());
    }
  }

  public void storeInstance(BpelExecution instance) {
    getBpelPersistenceService().save(instance.getProcessInstance());
  }

  public BpelExecution getUpdatedExecution(Execution execution) {
    return getBpelPersistenceService().get(BpelExecution.class,
        execution.getDbid());
  }

  public ReceivingElement getUpdatedReceivingElement(
      ReceivingElement receivingElement) {
    if (receivingElement instanceof Receive) {
      return getBpelPersistenceService().get(Receive.class, ((Receive)receivingElement).getDbid());
    } else if (receivingElement instanceof OnMessage) {
      return getBpelPersistenceService().get(OnMessage.class, ((OnMessage)receivingElement).getDbid());
    }
    return receivingElement;
  }
  /*
   * RECEIVER REPOSITORY
   */
  public Receiver getReceiver(OperationKey operationKey) {
    return getBpelPersistenceService().getReceiver(operationKey);
  }

  public Receiver removeReceiver(OperationKey operationKey) {
    Receiver receiver = getReceiver(operationKey);
    if (receiver != null) {
      getBpelPersistenceService().delete(receiver);
    }
    return receiver;
  }

  public void storeReceiver(OperationKey operationKey, Receiver receiver) {
    getBpelPersistenceService().save(receiver);
  }

  public List<Receiver> getReceivers(QName processQName) {
    return getBpelPersistenceService().getReceivers(processQName);
  }

  /*
   * REPLIER REPOSITORY
   */
  public Replier getReplier(ReplierKey replierKey) {
    return getBpelPersistenceService().getReplier(replierKey);
  }

  public void removeRepliers(BpelExecution instance) {
    // TODO: see for child execution scopes
    List<ScopeRuntime> scopeRuntimes = instance.getScopeRuntimes();
    if (scopeRuntimes !=  null) {
      for (ScopeRuntime scopeRuntime : scopeRuntimes) {
        List<Replier> repliers = getBpelPersistenceService().getRepliers(scopeRuntime);
        if (repliers != null) {
          for (Replier replier : repliers) {
            getBpelPersistenceService().delete(replier);    
          }
        }
      }
    }
  }
  
  public Replier removeReplier(ReplierKey replierKey) {
    Replier replier = getReplier(replierKey);
    if (replier != null) {
      getBpelPersistenceService().delete(replier);
    }
    return replier;
  }

  public void storeReplier(Replier replier) {
    getBpelPersistenceService().save(replier);
  }

  /*
   * COMMON METHODS
   */
  private BpelPersistenceService getBpelPersistenceService() {
    return Environment.getCurrent().get(BpelPersistenceService.class);
  }

}
