/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Map.Entry;
import java.util.zip.Deflater;
import java.util.zip.DeflaterOutputStream;

import javax.wsdl.Part;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.pool.BasePoolableObjectFactory;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.StackObjectPool;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.exception.OrchestraException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;

/**
 * @author Charles Souillard
 */
public final class BpelUtil {

  private static DocumentBuilderFactory documentBuilderFactory;
  private static TransformerFactory transformerFactory;
  private static ObjectPool transformerPool = new StackObjectPool(new PoolTransformerFactory());;
  private static ObjectPool documentBuilderPool = new StackObjectPool(new PoolDocumentBuilderFactory());

  /**
   * Default private constructor to prevent this class to be instanciated.
   */
  private BpelUtil() { }

  private static DocumentBuilderFactory getDocumentBuilderFactory() {
    if (documentBuilderFactory == null) {
      documentBuilderFactory = DocumentBuilderFactory.newInstance();
    }
    return documentBuilderFactory;
  }
  private static TransformerFactory getTransformerFactory() {
    if (transformerFactory == null) {
      transformerFactory = TransformerFactory.newInstance();
    }
    return transformerFactory;
  }

  private static class PoolTransformerFactory extends BasePoolableObjectFactory {
    public Object makeObject() {
      try {
        return getTransformerFactory().newTransformer();
      } catch (TransformerConfigurationException e) {
        throw new OrchestraException(e);
      }
    }
    //nothing to do when passivating
    //public void passivateObject(Object obj);
  }
  private static class PoolDocumentBuilderFactory extends BasePoolableObjectFactory {
    public Object makeObject() {
      try {
        DocumentBuilderFactory myDocumentBuilderFactory = getDocumentBuilderFactory();
        myDocumentBuilderFactory.setNamespaceAware(true);
        myDocumentBuilderFactory.setExpandEntityReferences(false);
        return myDocumentBuilderFactory.newDocumentBuilder();
      } catch (ParserConfigurationException e) {
        throw new OrchestraException(e);
      }
    }
    public void passivateObject(Object obj) {
      ((DocumentBuilder)obj).reset();
    }
  }
  private static Transformer getTransformer() {
    try {
      return (Transformer) transformerPool.borrowObject();
    } catch (Exception e) {
      throw new OrchestraException(e);
    }
  }
  private static void returnTransformer(Transformer transformer) {
    if (transformer != null) {
      try {
        transformerPool.returnObject(transformer);
      } catch (Exception e) {
        throw new OrchestraException(e);
      }
    }
  }
  private static DocumentBuilder getDocumentBuilder() {
    try {
      return (DocumentBuilder) documentBuilderPool.borrowObject();
    } catch (Exception e) {
      throw new OrchestraException(e);
    }
  }
  private static void returnDocumentBuilder(DocumentBuilder documentBuilder) {
    if (documentBuilder != null) {
      try {
        documentBuilderPool.returnObject(documentBuilder);
      } catch (Exception e) {
        throw new OrchestraException(e);
      }
    }
  }
  public static Document getNewDocument() {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = getDocumentBuilder();
      return documentBuilder.newDocument();
    } catch (Exception e) {
      throw new OrchestraException(e);
    } finally {
      returnDocumentBuilder(documentBuilder);
    }
  }

  /**
   * This method writes a DOM document to a file.
   * @param doc - doc to write
   * @param file - dest file;
   */
  public static void writeXmlFile(final Document doc, final File file) {
    Transformer transformer = null;
    FileOutputStream fos = null;
    try {
      // Prepare the DOM document for writing
      Source source = new DOMSource(doc);
      fos = new FileOutputStream(file);
      Result result = new StreamResult(fos);

      // Write the DOM document to the file
      transformer = getTransformer();
      transformer.transform(source, result);
    } catch (Exception e) {
      throw new OrchestraException("Exception caught while writing xml file", e);
    } finally {
      if (fos != null) {
        try {
          fos.close();
        } catch (Exception e) {
          // nothing
        }
      }
      returnTransformer(transformer);
    }
  }

  /**
   * Return true if the given element has an attribute named createInstance and if
   * its value is 'yes'.
   * @param element dom element to analyse
   * @return true if the given element has an attribute named createInstance and if
   * its value is 'yes'.
   */
  public static boolean isCreateInstance(Element element) {
    String createInstance = XmlUtil.attribute(element, "createInstance");
    if (createInstance != null && createInstance.equals("yes")) {
      return true;
    }
    return false;
  }
  /**
   * Get all namespaces defined in the given node and all of its parents.
   * @param n node from where namespaces are taken
   * @return all namespaces defined in the given node and all of its parents.
   */
  public static Map<String, String> getAllNameSpaces(org.w3c.dom.Node n) {
    HashMap<String, String> result = new HashMap<String, String>();
    if (n == null) {
      return result;
    }
    //we look for the NameSpace in the parent Node
    if (n.getParentNode() != null) {
      Map<String, String> resultParent = getAllNameSpaces(n.getParentNode());
      for (Entry<String, String> ns : resultParent.entrySet()) {
        if (!result.containsKey(ns.getKey())) {
          result.put(ns.getKey(), ns.getValue());
        }
      }

    }
    NamedNodeMap attributes = n.getAttributes();
    if (attributes != null) {
      for (int i = 0; i < attributes.getLength(); i++) {
        if (attributes.item(i).getNodeName().contains("xmlns:")) {
          String nodeName = attributes.item(i).getNodeName();
          String prefix = nodeName.split(":")[1];
          if (!result.containsKey(prefix)) {
            result.put(prefix, attributes.item(i).getTextContent());
          }
        }
      }
    }
    result.put(XmlConstants.BPEL_FUNC_PREFIX, XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE);
    return result;
  }

  /**
   * Return a string representation of the given Node based on javax.xml.transform api.
   * @param n node to convert to a string
   * @return a string representation of the given Node based on javax.xml.transform api.
   */
  public static String getNodeAsString(Node n) {
    Transformer transformer = null;
    try {
      Source source = new DOMSource(n);
      StringWriter stringWriter = new StringWriter();
      Result res = new StreamResult(stringWriter);
      transformer = getTransformer();
      transformer.transform(source, res);
      //return stringWriter.getBuffer().toString();

      String resultAsString = stringWriter.getBuffer().toString();
      String prolog = "<?xml";
      if (resultAsString.startsWith(prolog)) {
        String prologEnd = "?>";
        int prologEndIndex = resultAsString.indexOf(prologEnd);
        return resultAsString.substring(prologEndIndex + prologEnd.length());
      }
      return resultAsString;
    } catch (Exception e) {
      throw new OrchestraException("Exception caugth when transforming node to string", e);
    } finally {
      returnTransformer(transformer);
    }
  }

  public static ByteArrayOutputStream getNodeAsBAOS(Node n) {
    Transformer transformer = null;
    try {
      Source source = new DOMSource(n);
      ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
      Result res = new StreamResult(byteStream);
      transformer = getTransformer();

      transformer.transform(source, res);

      return byteStream;
    } catch (ClassCastException e) {
      throw new OrchestraException(e);
    } catch (Exception e) {
      throw new OrchestraException("Exception caugth when transforming node to bytestream", e);
    } finally {
      returnTransformer(transformer);
    }
  }

  public static ByteArrayOutputStream getNodeAsBAOS(Node n, Integer deflateLevel) {
    Transformer transformer = null;
    OutputStream deflaterStream = null;
    try {
      Source source = new DOMSource(n);
      ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
      Result res = null;
      if (deflateLevel == null) {
        res = new StreamResult(byteStream);
      } else {
        Deflater deflater = new Deflater(deflateLevel.intValue());
        deflaterStream = new DeflaterOutputStream(byteStream, deflater);
        res = new StreamResult(deflaterStream);
      }

      transformer = getTransformer();
      transformer.transform(source, res);


      return byteStream;
    } catch (Exception e) {
      throw new OrchestraException("Exception caugth when transforming node to string", e);
    } finally {
      if (deflaterStream != null) {
        try {
          deflaterStream.close();
        } catch (Exception e) {
          // nothing
        }
      }
      returnTransformer(transformer);
    }
  }

  /**
   * Returns the most local NameSpace corresponding to given prefix.
   * @param n : the child from where to begin the search
   * @param prefix : prefix linking to searched namespaceURI
   * @return the most local NameSpace corresponding to given prefix
   */
  public static String getNamespaceURI(org.w3c.dom.Node n, String prefix) {
    Node prefixDeclaration = n.getAttributes().getNamedItem("xmlns:" + prefix);
    if (prefixDeclaration != null) {
      //we have found the good NameSpace
      return prefixDeclaration.getNodeValue();
    }
    // we have found the good NameSpace
    //we look for the NameSpace in the parent Node
    return getNamespaceURI(n.getParentNode(), prefix);
  }

  /**
   * remove the prefix of the String.
   * @param s the String which has to be cut
   * @return the string without the prefix
   */
  public static String removePrefix(String s) {
    int i = s.indexOf(":");
    return s.substring(++i);
  }

  /**
   * Look for an instance of a class in a list of ExtensibilityElement.
   * @param listOfExtensibilityElement the list of ExtensibilityElement.
   * @param classToFind : the class we llok for
   * @return If found, return the element found, else, return null.
   */
  @SuppressWarnings("unchecked")
  public static Object findExtensibilityElement(List listOfExtensibilityElement, Class classToFind) {
    for (Object extElem : listOfExtensibilityElement) {
      if (classToFind.isInstance(extElem)) {
        return extElem;
      }
    }
    return null;
  }
  /**
   * return true if the given ns is a xmlSchema.
   * @param ns ns to analyse
   * @return true if the given ns is a xmlSchema.
   */
  public static boolean isXmlSchemaNs(String ns) {
    return ns != null && (ns.equals(XmlConstants.XMLNS_XMLSCHEMA) || ns.equals(XmlConstants.XMLNS_XMLSCHEMAXSD));
  }
  /**
   * Return a dom document conformant with ServiceRef xsd based on the given address and serviceQName.
   * @param address address to use to build the serviceRef document
   * @param serviceQN serviceQN to use to build the serviceRef document
   * @return a dom document conformant with ServiceRef xsd based on the given address and serviceQName.
   */
  public static Document getServiceRefDocument(SOAPAddress address, QName serviceQN) {
    DocumentBuilder documentBuilder = getDocumentBuilder();
    Document document = documentBuilder.newDocument();
    returnDocumentBuilder(documentBuilder);

    Element serviceRef = (Element)
      document.createElementNS(XmlConstants.XMLNS_BPEL_2_0_SERVICEREF, XmlConstants.SERVICE_REF_TAG_NAME);
    document.appendChild(serviceRef);

    Element epr = (Element)
      document.createElementNS(XmlConstants.ADDRESSING_REF_NAMESPACE, XmlConstants.EPR_TAG_NAME);
    serviceRef.appendChild(epr);

    Element addressElement = (Element)
      document.createElementNS(XmlConstants.ADDRESSING_REF_NAMESPACE,XmlConstants.ADDRESS_TAG_NAME);
    addressElement.setTextContent(address.getLocationURI());
    epr.appendChild(addressElement);

    Element serviceNameElement = (Element)
      document.createElementNS(XmlConstants.ADDRESSING_REF_NAMESPACE,XmlConstants.SERVICE_NAME_TAG_NAME);
    serviceNameElement.setAttribute("xmlns:eprServicePrefix", serviceQN.getNamespaceURI());
    serviceNameElement.setTextContent("eprServicePrefix:" + serviceQN.getLocalPart());
    epr.appendChild(serviceNameElement);

    return document;
  }

  public static List<Element> getChildrenElements(Node node) {
    List<Element> elements = new ArrayList<Element>();
    NodeList children = node.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
      Node child = children.item(i);
      if (child instanceof Element) {
        elements.add((Element)child);
      }
    }
    return elements;
  }

  public static Document getDocumentFromString(String s) {
    String prolog = "<?xml";
    if (s.startsWith(prolog)) {
      String prologEnd = "?>";
      int prologEndIndex = s.indexOf(prologEnd);
      s = s.substring(prologEndIndex + prologEnd.length());
    }
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = getDocumentBuilder();
      return documentBuilder.parse(new InputSource(new StringReader(s)));
    } catch (Exception e) {
      throw new OrchestraException("Esception while building a document from string : " + s, e);
    } finally {
      returnDocumentBuilder(documentBuilder);
    }
  }

  public static Document getDocumentFromInputStream(InputStream is) {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = getDocumentBuilder();
      return documentBuilder.parse(is);
    } catch (Exception e) {
      throw new OrchestraException("Exception while building a document from string : " + is, e);
    } finally {
      returnDocumentBuilder(documentBuilder);
    }
  }


  public static List<Element> getElements(Node n) {
    NodeList nodeList = n.getChildNodes();
    List<Element> elements = new ArrayList<Element>();
    for (int i = 0;i < nodeList.getLength();i++) {
      if (nodeList.item(i) instanceof Element) {
        elements.add((Element)nodeList.item(i));
      }
    }
    return elements;
  }

  public static List<Text> getTexts(Node n) {
    NodeList nodeList = n.getChildNodes();
    List<Text> texts = new ArrayList<Text>();
    for (int i = 0;i < nodeList.getLength();i++) {
      if (nodeList.item(i) instanceof Text) {
        texts.add((Text)nodeList.item(i));
      }
    }
    return texts;
  }
  public static Document getDocumentFromFile(File xmlFile) {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = getDocumentBuilder();
      return documentBuilder.parse(xmlFile);
    } catch (Exception e) {
      throw new OrchestraException("Unable to parse successfully xml file : " + xmlFile, e);
    } finally {
      returnDocumentBuilder(documentBuilder);
    }
  }

  public static Document getDocumentFromURL(URL xmlUrl) {
    DocumentBuilder documentBuilder = null;
    try {
      documentBuilder = getDocumentBuilder();
      InputStream inStream = xmlUrl.openStream();
      return documentBuilder.parse(inStream);
    } catch (Exception e) {
      throw new OrchestraException("Unable to parse successfully xml url : " + xmlUrl, e);
    } finally {
      returnDocumentBuilder(documentBuilder);
    }
  }

  public static Element getDocumentWithOneElement(QName elementQName) {
    Document document = getNewDocument();
    Element rootElement = (Element) document.createElementNS(elementQName.getNamespaceURI(), elementQName.getLocalPart());
    document.appendChild(rootElement);
    return rootElement;
  }

  public static Element getPartElement(Element father, Part part) {
    String partName = part.getName();
    List<Element> childrenElements = getElements(father);
    for (Element child : childrenElements) {
      if (child.getLocalName().equals(partName)) {
        return child;
      }
    }
    return null;
  }

  public static int getRandomInt(long diff, int max) {
    Random generator = new Random(System.currentTimeMillis() + diff);
    return generator.nextInt(max);
  }

  public static Boolean booleanValue(String text, boolean defaultValue) {
    Boolean result = booleanValue(text);
    if (result == null) {
      return defaultValue;
    }
    return result;
  }

  public static Boolean booleanValue(String text) {
    if (text == null) {
      return null;
    }

    // if we have to check for value true
    if (("true".equalsIgnoreCase(text))
        || ("enabled".equalsIgnoreCase(text))
        || ("on".equalsIgnoreCase(text))
        || ("yes".equalsIgnoreCase(text))
    ) {
      return Boolean.TRUE;
    }
    if (("false".equalsIgnoreCase(text))
        || ("disabled".equalsIgnoreCase(text))
        || ("off".equalsIgnoreCase(text))
        || ("no".equalsIgnoreCase(text))
    ) {
      return Boolean.FALSE;
    }

    return null;
  }
}


