/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the JBPM BPEL PUBLIC LICENSE AGREEMENT as
 * published by JBoss Inc.; either version 1.0 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.orchestra.exception.BpelFaultException;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**
 * Utility methods for dealing with JAXP objects.
 * @author Alejandro Guizar
 * @version $Revision: 1.26 $ $Date: 2007/11/25 13:03:14 $
 */
public final class BpelXmlUtil {

  static final String QUALIFIED_VALUE_PREFIX = "valueNS";

  private static final Log LOG = LogFactory.getLog(BpelXmlUtil.class);
  private static final boolean TRACE_ENABLED = LOG.isTraceEnabled();

  // suppress default constructor, ensuring non-instantiability
  private BpelXmlUtil() {
  }

  /**
   * Tells whether the name of the specified node matches the given namespace URI and local name.
   * @param node the node to examine
   * @param namespaceURI the namespace URI to match
   * @param localName the local name to match
   * @return <code>true</code> if the namespace URI and local name match, <code>false</code>
   * otherwise
   */
  public static boolean nodeNameEquals(Node node, String namespaceURI, String localName) {
    return nodeNamespaceURIEquals(node, namespaceURI) && localName.equals(node.getLocalName());
  }

  /**
   * Tells whether the namespace URI of the specified node matches the given namespace URI.
   * @param node the node to examine
   * @param namespaceURI the namespace URI to match
   * @return <code>true</code> if the namespace URI matches, <code>false</code> otherwise
   */
  public static boolean nodeNamespaceURIEquals(Node node, String namespaceURI) {
    String nodeNamespaceURI = node.getNamespaceURI();
    return nodeNamespaceURI == null || nodeNamespaceURI.length() == 0 ? namespaceURI == null
        || namespaceURI.length() == 0 : nodeNamespaceURI.equals(namespaceURI);
  }

  public static void setObjectValue(Node node, Object value, boolean keepSrcElementName) {
    switch (node.getNodeType()) {
      case Node.ELEMENT_NODE:
        setObjectValue((Element) node, value, keepSrcElementName);
        break;
      case Node.DOCUMENT_NODE:
        setObjectValue(((Document) node).getDocumentElement(), value, keepSrcElementName);
        break;
      default:
        // BPEL-243 throw selectionFailure if the source is an EII with xsi:nil=true
        if (value instanceof Element) {
          String nil = ((Element) value).getAttributeNS(XmlConstants.NS_XML_SCHEMA_INSTANCE,
            XmlConstants.ATTR_NIL);
          if (Boolean.TRUE.equals(DatatypeUtil.parseBoolean(nil))) {
            throw new BpelFaultException(XmlConstants.FAULT_SELECTION_FAILURE);
          }
        }
        // replace content
        node.setNodeValue(DatatypeUtil.toString(value));
    }
  }

  public static void setObjectValue(Element elem, Object value, boolean keepSrcElementName) {
    if (value instanceof Node) {
      switch (((Node) value).getNodeType()) {
        case Node.ELEMENT_NODE:
          // replace element
          copy(elem, (Element) value, keepSrcElementName);
          break;
        case Node.DOCUMENT_NODE:
          // replace element
          copy(elem, ((Document) value).getDocumentElement(), keepSrcElementName);
          break;
        default:
          // replace content
          setStringValue(elem, ((Node) value).getNodeValue());
      }
    /*}
    else if (value instanceof EndpointReference) {
      // replace element
      ((EndpointReference) value).writeServiceRef(elem);
    }*/
    } else {
      // replace content
      setStringValue(elem, DatatypeUtil.toString(value));
    }
  }

  public static void setStringValue(Element elem, String value) {
    // remove jbpm:initialized
    elem.removeAttributeNS(XmlConstants.NS_VENDOR, XmlConstants.ATTR_INITIALIZED);

    Node firstChild = elem.getFirstChild();
    // if first child is a text node, reuse it
    if (firstChild instanceof org.w3c.dom.Text) {
      firstChild.setNodeValue(value);
    // otherwise, just create a new text node
    } else {
      firstChild = elem.getOwnerDocument().createTextNode(value);
    }
    // remove all children
    removeChildNodes(elem);
    // append text
    elem.appendChild(firstChild);
  }

  public static String toTraceString(Element elem) {
    String namespace = elem.getNamespaceURI();
    String localName = elem.getLocalName();

    // easy way out: no namespace
    if (namespace == null || namespace.length() == 0) {
      return localName;
    }
    StringBuffer traceBuffer = new StringBuffer(namespace.length() + localName.length());
    traceBuffer.append('{').append(namespace).append('}');

    String prefix = elem.getPrefix();
    if (prefix != null && prefix.length() > 0) {
      traceBuffer.append(prefix).append(':');
    }
    return traceBuffer.append(localName).toString();
  }

  public static void copy(Element target, Element source, boolean keepSrcElementName) {
    if (TRACE_ENABLED) {
      LOG.trace("copying from: " + toTraceString(source));
    }
    if (keepSrcElementName) {
      /*
      Node parentNode = target.getParentNode();
      Document doc = target.getOwnerDocument();

      Element newTarget = doc.createElementNS(source.getNamespaceURI(), source.getLocalName());
      copyVisibleNamespaces(newTarget, target);
      ensureOwnNamespaceDeclared(newTarget);
      parentNode.replaceChild(newTarget, target);
      target = newTarget;
  */
    }
    // attributes
    removeAttributes(target);
    copyAttributes(target, source);
    // all namespaces
    copyVisibleNamespaces(target, source);
    ensureOwnNamespaceDeclared(target);
    // child nodes
    removeChildNodes(target);

    copyChildNodes(target, source);
    if (TRACE_ENABLED) {
      LOG.trace("copied to: " + toTraceString(target));
    }
  }

  public static void copyVisibleNamespaces(final Element target, Element source) {
    // copy namespaces declared at source element
    copyNamespaces(target, source);
    // go up the element hierarchy
    for (Node parent = source.getParentNode(); parent instanceof Element; parent = parent.getParentNode()) {
      copyNamespaces(target, (Element) parent);
    }
  }

  public static void copyNamespaces(final Element target, Element source) {
    // easy way out: no attributes
    if (!source.hasAttributes()) {
      return;
    }
    // traverse attributes to discover namespace declarations
    NamedNodeMap attributes = source.getAttributes();
    for (int i = 0, n = attributes.getLength(); i < n; i++) {
      Node attribute = attributes.item(i);
      // is attribute a namespace declaration?
      if (!XmlConstants.NS_XMLNS.equals(attribute.getNamespaceURI())) {
        continue;
      }
      // namespace declaration format
      // xmlns:prefix="namespaceURI" | xmlns="defaultNamespaceURI"
      String namespaceURI = attribute.getNodeValue();
      String prefix = attribute.getLocalName();
      // default namespace declaration?
      if ("xmlns".equals(prefix)) {
        // BPEL-195: prevent addition matching visible declaration at target
        if ("".equals(getPrefix(namespaceURI, target))) {
          continue;
        }
        addNamespaceDeclaration(target, namespaceURI);
        if (TRACE_ENABLED) {
          LOG.trace("added default namespace declaration: " + namespaceURI);
        }
      } else {
        // BPEL-195: prevent addition matching visible declaration at target
        if (prefix.equals(getPrefix(namespaceURI, target))) {
          continue;
        }
        addNamespaceDeclaration(target, namespaceURI, prefix);
        if (TRACE_ENABLED) {
          LOG.trace("added namespace declaration: " + prefix + "->" + namespaceURI);
        }
      }
    }
  }

  public static void copyAttributes(Element target, Element source) {
    // easy way out: no attributes
    if (!source.hasAttributes()) {
      return;
    }
    // traverse attributes
    NamedNodeMap attributes = source.getAttributes();
    for (int i = 0, n = attributes.getLength(); i < n; i++) {
      Node sourceAttr = attributes.item(i);
      String namespaceURI = sourceAttr.getNamespaceURI();
      String name = sourceAttr.getNodeName();
      // namespace declaration?
      if (XmlConstants.NS_XMLNS.equals(namespaceURI)) {
        continue;
      }
      // unqualified?
      if (namespaceURI == null || namespaceURI.length() == 0) {
        target.setAttribute(name, sourceAttr.getNodeValue());
        if (TRACE_ENABLED) {
          LOG.trace("set attribute: " + name);
        }
      } else {
        // qualified
        Attr targetAttr = target.getOwnerDocument().createAttributeNS(namespaceURI, name);
        targetAttr.setValue(sourceAttr.getNodeValue());
        target.setAttributeNodeNS(targetAttr);
        ensureNamespaceDeclared(targetAttr, namespaceURI, sourceAttr.getPrefix());
        if (TRACE_ENABLED) {
          LOG.trace("set attribute: {" + namespaceURI + '}' + name);
        }
      }
    }
  }

  public static void copyChildNodes(Element target, Element source) {
    Document targetDoc = target.getOwnerDocument();
    for (Node child = source.getFirstChild(); child != null; child = child.getNextSibling()) {
      switch (child.getNodeType()) {
        case Node.ELEMENT_NODE:
          copyChildElement(target, (Element) child);
          break;
        case Node.TEXT_NODE:
        case Node.CDATA_SECTION_NODE:
          target.appendChild(targetDoc.createTextNode(child.getNodeValue()));
          if (TRACE_ENABLED) {
            LOG.trace("appended text: " + child.getNodeValue());
          }
          break;
        default:
          LOG.debug("discarding child: " + child);
      }
    }
  }

  public static void copyChildElement(Element parent, Element source) {
    Element target = (Element) parent.getOwnerDocument().importNode(source, true);
    parent.appendChild(target);
    // ensure namespace declared after appending in order to consider all visible declarations
    ensureOwnNamespaceDeclared(target);
    if (TRACE_ENABLED) {
      LOG.trace("appended element: " + toTraceString(target));
    }
  }

  /**
   * Ensures the given namespace URI is associated to the given prefix in the scope of the given
   * attribute. If the namespace URI is <em>not</em> associated to the prefix, a namespace
   * declaration is {@linkplain #addNamespaceDeclaration(Element, String, String) added} to the
   * attribute's owner element. Otherwise, this call has no effect.
   * @param attribute the attribute to verify
   * @param namespaceURI the namespace URI to ensure; cannot be <code>null</code> or empty
   * @param prefix the prefix to ensure; cannot be <code>null</code> or empty
   */
  public static void ensureNamespaceDeclared(Attr attribute, String namespaceURI, String prefix) {
    if (prefix == null || prefix.length() == 0) {
      throw new IllegalArgumentException("prefix cannot be empty");
    }
    if (namespaceURI == null || namespaceURI.length() == 0) {
      throw new IllegalArgumentException("namespaceURI cannot be empty");
    }
    Element element = attribute.getOwnerElement();
    if (!prefix.equals(getPrefix(namespaceURI, element))) {
      // prefix is associated with other/no URI, declare locally
      addNamespaceDeclaration(element, namespaceURI, prefix);
    }
  }

  /**
   * Ensures the given namespace URI is associated to the given prefix in the scope of the given
   * element. If the namespace URI is <em>not</em> associated to the prefix, a namespace
   * declaration is {@linkplain #addNamespaceDeclaration(Element, String, String) added} to the
   * element. Otherwise, this call has no effect.
   * @param element the element to verify
   * @param namespaceURI the namespace URI to ensure; cannot be <code>null</code> or empty unless
   * <code>prefix</code> is <code>null</code> or empty as well
   * @param prefix the prefix to ensure; if <code>null</code> or empty, the given namespace URI is
   * ensured to be the default
   * @throws IllegalArgumentException if <code>namespaceURI</code> is <code>null</code> or
   * empty, <em>unless</em> <code>prefix</code> is empty as well
   */
  public static void ensureNamespaceDeclared(Element element, String namespaceURI, String prefix) {
    if (prefix == null || prefix.length() == 0) {
      if (namespaceURI == null) {
        namespaceURI = "";
      }
      // verify the given URI is the default namespace
      if (!"".equals(getPrefix(namespaceURI, element))) {
        // the given URI is not the default namespace, declare locally
        addNamespaceDeclaration(element, namespaceURI);
      }
    } else {
      if (namespaceURI == null || namespaceURI.length() == 0) {
        throw new IllegalArgumentException("namespaceURI cannot be empty unless prefix is empty");
      }
      // verify given prefix is associated to given URI
      if (!prefix.equals(getPrefix(namespaceURI, element))) {
        // prefix is associated with other/no URI, declare locally
        addNamespaceDeclaration(element, namespaceURI, prefix);
      }
    }
  }

  public static void ensureOwnNamespaceDeclared(Element elem) {
    ensureNamespaceDeclared(elem, elem.getNamespaceURI(), elem.getPrefix());
  }

  public static void addNamespaceDeclaration(Element elem, String namespaceURI) {
    elem.setAttributeNS(XmlConstants.NS_XMLNS, "xmlns", namespaceURI);
  }

  public static void addNamespaceDeclaration(Element elem, String namespaceURI, String prefix) {
    elem.setAttributeNS(XmlConstants.NS_XMLNS, "xmlns:" + prefix, namespaceURI);
  }

  public static void removeAttributes(Element elem) {
    if (elem.hasAttributes()) {
      NamedNodeMap attributeMap = elem.getAttributes();
      // since node maps are live, hold attributes in a separate collection
      int n = attributeMap.getLength();
      Attr[] attributes = new Attr[n];
      for (int i = 0; i < n; i++) {
        attributes[i] = (Attr) attributeMap.item(i);
      }
      // now remove each attribute from the element
      for (int i = 0; i < n; i++) {
        elem.removeAttributeNode(attributes[i]);
      }
    }
  }

  public static void removeChildNodes(Node node) {
    for (Node current = node.getFirstChild(), next; current != null; current = next) {
      next = current.getNextSibling();
      node.removeChild(current);
    }
  }

  /**
   * Creates an element with the given qualified name. The returned element will be set as the
   * document element of its owner document.
   * @param name the qualified name of the element to create
   * @return a new element with the specified <code>name</code>
   */
  public static Element createElement(QName name) {
    String namespace = name.getNamespaceURI();
    String localName = name.getLocalPart();
    if (namespace.length() == 0) {
      return createElement(localName);
    }
    String prefix = name.getPrefix();
    return createElement(namespace, prefix.length() > 0 ? prefix + ':' + localName : localName);
  }

  /**
   * Creates an element with the given namespace URI and prefixed name. The returned element will be
   * set as the document element of its owner document.
   * @param namespaceURI the namespace URI of the element to create
   * @param prefixedName the prefixed name of the element to instantiate
   * @return a new element with the specified <code>namespaceURI</code> and
   * <code>prefixedName</code>.
   */
  public static Element createElement(String namespaceURI, String prefixedName) {
    Document doc = BpelUtil.getNewDocument();
    Element elem = doc.createElementNS(namespaceURI, prefixedName);
    doc.appendChild(elem);

    // some TrAX implementations do not fix up namespaces, declare namespace
    String prefix = elem.getPrefix();
    if (prefix != null) {
      addNamespaceDeclaration(elem, namespaceURI, prefix);
    } else {
      addNamespaceDeclaration(elem, namespaceURI);
    }
    return elem;
  }

  /**
   * Creates an element with the given local name and neither namespace URI nor prefix. The returned
   * element will be set as the document element of its owner document.
   * @param localName the local name of the element to create
   * @return a new element with the specified <code>localName</code>
   */
  public static Element createElement(String localName) {
    Document doc = BpelUtil.getNewDocument();
    Element elem = doc.createElementNS(null, localName);
    doc.appendChild(elem);
    return elem;
  }

  /**
   * Retrieves the prefix associated with a namespace URI in the given context node.
   * @param namespaceURI the namespace whose prefix is required
   * @param contextNode the node where to search for namespace declarations
   * @return the prefix associated with the namespace URI; the empty string indicates the default
   * namespace, while <code>null</code> indicates no association
   */
  public static String getPrefix(String namespaceURI, Node contextNode) {
    switch (contextNode.getNodeType()) {
      case Node.ATTRIBUTE_NODE:
        contextNode = ((Attr) contextNode).getOwnerElement();
        break;
      case Node.ELEMENT_NODE:
        break;
      default:
        contextNode = contextNode.getParentNode();
    }

    while (contextNode != null && contextNode.getNodeType() == Node.ELEMENT_NODE) {
      NamedNodeMap attributes = contextNode.getAttributes();
      for (int i = 0, n = attributes.getLength(); i < n; i++) {
        Node attr = attributes.item(i);

        // is attribute a namespace declaration and matches the given URI?
        if (XmlConstants.NS_XMLNS.equals(attr.getNamespaceURI())
            && namespaceURI.equals(attr.getNodeValue())) {
          String prefix = attr.getLocalName();
          return "xmlns".equals(prefix) ? "" : prefix;
        }
      }
      contextNode = contextNode.getParentNode();
    }
    return null;
  }


  public static Element createDefaultElement() {
    return createElementWithContent(null);
  }
  public static Element createElementWithContent(String data) {
    Document document = BpelUtil.getNewDocument();

    Element defaultElement = document.createElementNS(
        XmlConstants.ELT_DEFAULT_NS,
        XmlConstants.ELT_DEFAULT_PREFIX + ":" + XmlConstants.ELT_DEFAULT_ELEMENT);

    defaultElement.setAttributeNS("http://www.w3.org/2000/xmlns/",
        "xmlns:" + XmlConstants.ELT_DEFAULT_PREFIX, XmlConstants.ELT_DEFAULT_NS);
    document.appendChild(defaultElement);
    if (data != null) {
      defaultElement.setTextContent(data);
    }
    return defaultElement;
  }

  public static boolean isDefaultElement(Element element) {
    if (element == null) {
      return false;
    }
    return element.getNamespaceURI().equals(XmlConstants.ELT_DEFAULT_NS)
      && element.getLocalName().equals(XmlConstants.ELT_DEFAULT_ELEMENT);
  }

}
