/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;

import javax.wsdl.Part;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.MimeHeaders;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPConnection;
import javax.xml.soap.SOAPConnectionFactory;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;
import javax.xml.transform.dom.DOMSource;

import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.var.Message;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

public final class SOAPUtil {

  private static SOAPConnectionFactory conFactory;
  private static MessageFactory messagefactory;

  private SOAPUtil() {

  }

  private static synchronized SOAPConnectionFactory getSOAPConnectionFactory() {
    if (conFactory == null) {
      try {
        conFactory = SOAPConnectionFactory.newInstance();
      } catch (Exception e) {
        throw new OrchestraException(e);
      }
    }
    return conFactory;
  }

  private static synchronized MessageFactory getMessageFactory() {
    if (messagefactory == null) {
      try {
        messagefactory = MessageFactory.newInstance();
      } catch (Exception e) {
        throw new OrchestraException(e);
      }
    }
    return messagefactory;
  }

  private static SOAPMessage getSOAPMessageFromDocument(Document document) {
    SOAPMessage soapMessage = null;
    try {
      soapMessage = getMessageFactory().createMessage();
      SOAPPart soapPart = soapMessage.getSOAPPart();
      soapPart.setContent(new DOMSource(document));
    } catch (Exception e) {
      throw new OrchestraException("Exception when creating a SOAPMessage from document : " + document, e);
    }
    return soapMessage;
  }

  public static SOAPMessage buildSOAPMessage(Document document, String soapAction) {
    SOAPMessage soapMessage = null;
    try {
      soapMessage = getMessageFactory().createMessage();

      MimeHeaders hd = soapMessage.getMimeHeaders();
      hd.addHeader("SOAPAction", soapAction);
      SOAPPart soapPart = soapMessage.getSOAPPart();

      SOAPEnvelope enveloppe = soapPart.getEnvelope();
      enveloppe.addNamespaceDeclaration("env", "http://schemas.xmlsoap.org/soap/envelope/");

      SOAPBody body = enveloppe.getBody();

      body.addDocument(document);

      soapMessage.saveChanges();
    } catch (Exception e) {
      throw new OrchestraException("Exception while building a SOAPMessage", e);
    }
    return soapMessage;
  }

  public static SOAPMessage buildRpcSOAPMessage(String soapAction, Message message, String operationName) {
    Document document = buildRpcDocument(message, operationName);
    return buildSOAPMessage(document, soapAction);
  }

  public static Document buildRpcDocument(Message message, String operationName) {
    Document document = null;
    try {
      document = BpelUtil.getNewDocument();
      Element operation = document.createElement(operationName);

      Iterator<Entry<String, Element>> it = message.getParts().entrySet().iterator();
      while (it.hasNext()) {
        Entry<String, Element> entry = it.next();
        String partName = entry.getKey();
        Element partElement = entry.getValue();

        if (BpelXmlUtil.isDefaultElement(partElement)) {
          Element partEl = document.createElement(partName);
          partEl.setTextContent(partElement.getTextContent());
          operation.appendChild(partEl);
        } else {
          Node importedPart = document.importNode(partElement, true);
          operation.appendChild(importedPart);
        }
      }
      document.appendChild(operation);
    } catch (Exception e) {
      throw new OrchestraException("Exception while building Rpc SOAP Message", e);
    }
    return document;
  }

  public static SOAPMessage buildDocumentSOAPMessage(String soapAction, Message message) {
    Element part = message.getParts().values().iterator().next();
    return buildDocumentSOAPMessage(soapAction, part);
  }

  public static SOAPMessage buildDocumentSOAPMessage(String soapAction, Element part) {
    Document document = null;
    try {
      document = BpelUtil.getNewDocument();
      Node importedPart = document.importNode(part, true);
      document.appendChild(importedPart);

    } catch (Exception e) {
      throw new OrchestraException("Exception while building Rpc SOAP Message", e);
    }
    return buildSOAPMessage(document, soapAction);
  }

  public static SOAPMessage buildDocumentSOAPMessage(String soapAction, String part) {
    Element partEl = BpelUtil.getDocumentFromString(part).getDocumentElement();
    return buildDocumentSOAPMessage(soapAction, partEl);
  }

  public static SOAPMessage call(SOAPMessage soapMessage, String endpoint) {
    SOAPMessage response = null;
    SOAPConnection connection = null;
    try {
      connection = getSOAPConnectionFactory().createConnection();
      response = connection.call(soapMessage, endpoint);
    } catch (Exception e) {
      throw new OrchestraException("Exception caught while performing call", e);
    } finally {
      if (connection != null) {
        try {
          connection.close();
        } catch (Exception e) {
          // nothing
        }
      }
    }
    return response;
  }

  public static Message getMessageFromRpcElement(Element operationElement, Collection<Part> parts) {
    HashMap <String, Element> messageParts = new HashMap<String, Element>();
    Message message = new Message(messageParts);

    for (Part part : parts) {
      Element el = BpelUtil.getPartElement(operationElement, part);
      if (part.getTypeName() != null && BpelUtil.isXmlSchemaNs(part.getTypeName().getNamespaceURI())) {
        //it is a schema type
        // TODO : manage simpleTypes here
        String text = ((Text)el.getFirstChild()).getData();
        Element element = BpelXmlUtil.createElementWithContent(text);
        messageParts.put(part.getName(), element);
      } else {
        messageParts.put(part.getName(), el);
      }
    }
    return message;
  }

  public static Message buildMessageFromRpcSOAPBodyElement(Collection<Part> parts, Element soapBodyElement) {
    HashMap<String, Element> messageParts = new HashMap<String, Element>();
    for (Part part : parts) {
      Element el = BpelUtil.getPartElement(soapBodyElement, part);
      if (el == null) {
        throw new OrchestraException("Impossible to build a Message from the given rpc soap body element."
            + " One element with a localName equals to " + part.getName() + " is mandatory.");
      }
      if (part.getTypeName() != null && BpelUtil.isXmlSchemaNs(part.getTypeName().getNamespaceURI())) {
        //it is a schema type
        String text = null;
        Node child = el.getFirstChild();
        if (child != null) {
          text = ((Text)child).getData();
        } else {
          text = el.getTextContent();
        }

        Element element = BpelXmlUtil.createElementWithContent(text);
        messageParts.put(part.getName(), element);
      } else {
        messageParts.put(part.getName() , el);
      }
    }
    Message message = new Message(messageParts);
    return message;
  }
  public static Message buildMessageFromDocumentSOAPBodyElement(Collection<Part> parts, Element soapBodyElement) {
    HashMap <String, Element> messageParts = new HashMap<String, Element>();
    Part part = parts.iterator().next();
    String partName = part.getName();
    messageParts.put(partName, soapBodyElement);
    return new Message(messageParts);
  }

}

