/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.HashMap;
import java.util.Map;


/**
 * List of all the Static Analysis Faults
 *
 * @author Goulven Le Jeune
 *
 */
public final class StaticAnalysisFault {

  private StaticAnalysisFault() { }

  // Map of the code and description of the fault
  private static Map<String, String> saf = null;

  /**
   * Get the Static Analysis Fault description from its code.
   *
   * @param code : Code of the Static Analysis Fault
   * @return String with the description of the Fault
   */
  public static String getFault(String code) {
    if (saf == null) {
      initiateSAF();
    }
    return saf.get(code);
  }

  /**
   * Build the Fault Message based on its code and cause.
   *
   * @param code : Code of the Static Analysis Fault
   * @param cause : Explanation of the fault
   * @return a string with the definition of the Static Analysis fault and the explanation
   */
  public static String getMessage(String code, String cause) {
    String message = StaticAnalysisFault.getFault(code);
    message += System.getProperty("line.separator");
    message += cause;
    return message;
  }

  private static void initiateSAF() {
    saf = new HashMap<String, String>();
    saf.put("SA00001", "SA00001: A WS-BPEL processor MUST reject a WS-BPEL "
        + "that refers to solicit-response or notification operations portTypes.");
    saf.put("SA00002", "SA00002: A WS-BPEL processor MUST reject any WSDL "
        + "portType definition that includes overloaded operation names.");
    saf.put("SA00003", "");
    saf.put("SA00004", "");
    saf.put("SA00005", "SA00005: If the portType attribute is included for "
        + "readability, in a <receive>, <reply>, <invoke>, <onEvent> or "
        + "<onMessage> element, the value of the portType attribute MUST "
        + "match the portType value implied by the combination of the "
        + "specified partnerLink and the role implicitly specified by the activity.");
    saf.put("SA00006","");
    saf.put("SA00007","");
    saf.put("SA00008","");
    saf.put("SA00009","");
    saf.put("SA00010","");
    saf.put("SA00011","");
    saf.put("SA00012","");
    saf.put("SA00013","");
    saf.put("SA00014","");
    saf.put("SA00015","SA00015: To be instantiated, an executable business "
        + "process MUST contain at least one <receive> or <pick> activity annotated with a createInstance=yes attribute.");
    saf.put("SA00016","SA00016: A partnerLink MUST specify the myRole or the partnerRole, or both.");
    saf.put("SA00017","SA00017: The initializePartnerRole attribute MUST NOT "
        + "be used on a partnerLink that does not have a partner role.");
    saf.put("SA00018","SA00018: The name of a partnerLink MUST be unique "
        + "among the names of all partnerLinks defined within the same immediately enclosing scope.");
    saf.put("SA00019","");
    saf.put("SA00020","");
    saf.put("SA00021","");
    saf.put("SA00022","");
    saf.put("SA00023","SA00023: The name of a variable MUST be unique among "
        + "the names of all variables defined within the same immediately enclosing scope.");
    saf.put("SA00024","SA00024: Variable names are BPELVariableNames, that is,"
        + "NCNames (as defined in XML Schema specification) but in addition they MUST NOT contain the . character.");
    saf.put("SA00025","SA00025: The messageType, type or element attributes "
        + "are used to specify the type of a variable. Exactly one of these attributes MUST be used.");
    saf.put("SA00026","");
    saf.put("SA00027","");
    saf.put("SA00028","");
    saf.put("SA00029","");
    saf.put("SA00030","");
    saf.put("SA00031","");
    saf.put("SA00032","");
    saf.put("SA00033","");
    saf.put("SA00034","");
    saf.put("SA00035","");
    saf.put("SA00036","");
    saf.put("SA00037","");
    saf.put("SA00038","");
    saf.put("SA00039","");
    saf.put("SA00040","");
    saf.put("SA00041","");
    saf.put("SA00042","");
    saf.put("SA00043","");
    saf.put("SA00044","SA00044: The name of a <correlationSet> MUST be unique"
         + "among the names of all <correlationSet> defined within the same immediately enclosing scope.");
    saf.put("SA00045","");
    saf.put("SA00046","");
    saf.put("SA00047","");
    saf.put("SA00048","");
    saf.put("SA00049","");
    saf.put("SA00050","");
    saf.put("SA00051","SA00051: The inputVariable attribute MUST NOT be used on an "
          + "Invoke activity that contains <toPart> elements.");
    saf.put("SA00052","SA00052: The outputVariable attribute MUST NOT be used on an "
          + "<invoke> activity that contains a <fromPart> element.");
    saf.put("SA00053","");
    saf.put("SA00054","");
    saf.put("SA00055","SA00055: For <receive>, if <fromPart> elements are used on a "
          + "<receive> activity then the variable attribute MUST NOT be used on the same activity.");
    saf.put("SA00056","");
    saf.put("SA00057","");
    saf.put("SA00058","");
    saf.put("SA00059","SA00059: For <reply>, if <toPart> elements are used on a "
          + "<reply> activity then the variable attribute MUST NOT be used on the same activity.");
    saf.put("SA00060","");
    saf.put("SA00061","");
    saf.put("SA00062","");
    saf.put("SA00063","");
    saf.put("SA00064","SA00064: For <flow>, a declared link s name MUST be unique "
          + "among all <link> names defined within the same immediately enclosing <flow>.");
    saf.put("SA00065","SA00065: The value of the linkName attribute of <source> or "
          + "<target> MUST be the name of a <link> declared in an enclosing <flow> activity.");
    saf.put("SA00066","SA00066: Every link declared within a <flow> activity MUST "
          + "have exactly one activity within the <flow> as its source and exactly "
          + "one activity within the <flow> as its target.");
    saf.put("SA00067","");
    saf.put("SA00068","SA00068: An activity MAY declare itself to be the source of "
          + "one or more links by including one or more <source> elements. Each <source> element MUST use a distinct link name.");
    saf.put("SA00069","SA00069: An activity MAY declare itself to be the target of "
          + "one or more links by including one or more <target> elements. Each "
          + "<target> element associated with a given activity MUST use a link name "
          + "distinct from all other <target> elements at that activity.");
    saf.put("SA00070","");
    saf.put("SA00071","");
    saf.put("SA00072","");
    saf.put("SA00073","");
    saf.put("SA00074","");
    saf.put("SA00075","");
    saf.put("SA00076","");
    saf.put("SA00077","");
    saf.put("SA00078","");
    saf.put("SA00079","");
    saf.put("SA00080","");
    saf.put("SA00081","");
    saf.put("SA00082","");
    saf.put("SA00083","");
    saf.put("SA00084","");
    saf.put("SA00085","");
    saf.put("SA00086","");
    saf.put("SA00087","");
    saf.put("SA00088","");
    saf.put("SA00089","");
    saf.put("SA00090","");
    saf.put("SA00091","");
    saf.put("SA00092","");
    saf.put("SA00093","");
    saf.put("SA00094","");
    saf.put("SA00095","");
  }
}
