/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.var;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.wsdl.Part;
import javax.xml.namespace.QName;

import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.w3c.dom.Element;

/**
 * Models a WSDL message type variable.
 * 
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet
 *         De Boudemange
 */
public class Message {

  protected long dbid;

  protected Map<String, Element> parts;

  /**
   * Empty message constructor. Parts can be added with
   * {@link #addPart(MessagePart)}.
   */
  public Message() {
    this.parts = new HashMap<String, Element>();
  }

  /**
   * Create a multi-part message.
   * 
   * @param parts
   *            HashMap containing the message parts
   */
  public Message(Map<String, Element> parts) {
    this.parts = parts;
  }

  /**
   * Constructor based on the wsdl Message.
   * 
   * @param wsdlMessage
   *            wsdlMessage that specify this message
   */
  @SuppressWarnings("unchecked")
  public Message(javax.wsdl.Message wsdlMessage) {
    this();
    Map<String, Part> wsdlMessageParts = wsdlMessage.getParts();
    for (Iterator<Map.Entry<String, Part>> i = wsdlMessageParts.entrySet()
        .iterator(); i.hasNext();) {
      Part part = ((Map.Entry<String, Part>) i.next()).getValue();
      String partName = part.getName();
      // this.parts.put(partName, null);
      QName elementQName = part.getElementName();
      QName typeQName = part.getTypeName();
      if (typeQName != null) {
        this.parts.put(partName, BpelXmlUtil.createDefaultElement());
      } else {
        this.parts.put(partName, BpelXmlUtil.createElement(elementQName));
      }
    }
  }

  public Element getPartValue(String partName) {
    if (partName != null) {
      return parts.get(partName);
    }
    return parts.values().iterator().next();
  }

  /**
   * Return true if the part having the given name is initialized.
   * 
   * @param partName
   *            partName
   * @return true if the part having the given name is initialized.
   */
  public boolean isPartInitialized(String partName) {
    return parts.get(partName) != null;
  }

  /**
   * GetVariablePropertyFunctionTest if the message has part named partName
   * (given in parameter).
   * 
   * @param partName :
   *            name of the part.
   * @return true if the part has been found, false else.
   */
  public boolean hasPart(String partName) {
    if (partName != null) {
      return parts.containsKey(partName);
    }
    return false;
  }

  /**
   * Return a string representation of the message.
   * 
   * @return a String representation of the message
   */
  public String toString() {
    String s = "<message>";
    for (String part : parts.keySet()) {
      if (parts.get(part) != null) {
        s += "<" + part + ">";
        if (parts.get(part) == null) {
          s += "no_value_for_this_part";
        } else {
          s += XmlUtil.toString(parts.get(part));
        }
        s += "</" + part + ">";
      }
    }
    s += "</message>";
    return s;
  }

  /**
   * Make a deep copy of the Xml object.
   * 
   * @return The duplicate of Xml object
   */
  public Message duplicate() {
    Message duplicateMessage = new Message();
    // Get an iterable set of all part names, as HashMaps are not iterable.
    Set<String> partNameSet = this.parts.keySet();

    for (String partName : partNameSet) {
      Element duplicatePart = null;
      if (this.parts.get(partName) != null) {
        duplicatePart = (Element) this.parts.get(partName).cloneNode(true);
      }
      duplicateMessage.parts.put(partName, duplicatePart);
    }
    return duplicateMessage;
  }

  /**
   * Compare the Message to another one.
   * 
   * @param anotherMessage :
   *            the other Message to compare to
   * @return true if all parts are equals, else false
   */
  public boolean equals(Object anotherMessage) {
    if (anotherMessage == null) {
      return false;
    }
    if (!(anotherMessage instanceof Message)) {
      return false;
    }
    Message message = (Message) anotherMessage;
    Map<String, Element> otherParts = message.getParts();

    return otherParts.equals(parts);
  }

  /**
   * Returns a hash code value for the Message.
   * 
   * @return a hash code value for this Message.
   */
  public int hashCode() {
    int hashCode = 0;
    for (Element messagePart : parts.values()) {
      hashCode += messagePart.hashCode();
    }
    return hashCode;
  }

  public Map<String, Element> getParts() {
    return parts;
  }

  public void setParts(Map<String, Element> parts) {
    this.parts = parts;
  }

  public void setPart(String partName, Object value) {
    Element currentValue = getPartValue(partName);
    BpelXmlUtil.setObjectValue(currentValue, value, false);
  }
}
