/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 *
 * WSDeployer.java
 *
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet De Boudemange
 *
 * Created on : Mar 1, 2007
 *
 * $Id: WSDeployer.java 1380 2008-04-03 15:12:25Z souillac $
 **/
package org.ow2.orchestra.ws;

import com.ibm.wsdl.extensions.soap.SOAPConstants;
import com.ibm.wsdl.extensions.soap12.SOAP12Constants;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Logger;

import javax.wsdl.Binding;
import javax.wsdl.BindingFault;
import javax.wsdl.BindingInput;
import javax.wsdl.BindingOperation;
import javax.wsdl.BindingOutput;
import javax.wsdl.Fault;
import javax.wsdl.Input;
import javax.wsdl.Operation;
import javax.wsdl.OperationType;
import javax.wsdl.Output;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.extensions.ElementExtensible;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.extensions.soap.SOAPBinding;
import javax.wsdl.extensions.soap.SOAPBody;
import javax.wsdl.extensions.soap.SOAPFault;
import javax.wsdl.extensions.soap.SOAPOperation;
import javax.wsdl.extensions.soap12.SOAP12Address;
import javax.wsdl.extensions.soap12.SOAP12Binding;
import javax.wsdl.extensions.soap12.SOAP12Body;
import javax.wsdl.extensions.soap12.SOAP12Fault;
import javax.wsdl.extensions.soap12.SOAP12Operation;
import javax.xml.namespace.QName;

import org.jbpm.xml.Problem;
import org.ow2.orchestra.exception.OrchestraException;




/**
 *
 * This class will deploy a WebService on the choosen ws engine.
 */
public abstract class WSDeployer {

  private static Logger log = Logger.getLogger(WSDeployer.class.getName());

  /**
   * Ns of wsdl soap 1.1.
   */
  public static final String URI_WSDL11_SOAP11 = SOAPConstants.NS_URI_SOAP;

  /**
   * Ns of wsdl soap 1.2.
   */
  public static final String URI_WSDL11_SOAP12 = SOAP12Constants.NS_URI_SOAP12;

  /**
   * Default value for soap binding style.
   */
  public static final String DEFAULT_SOAP_BINDING_STYLE = "document";

  /**
   * soap over http uri used in soap binding.
   */
  public static final String SOAP_HTTP_TRANSPORT_URI = "http://schemas.xmlsoap.org/soap/http";

  /**
   * List of problems found in the wsdl file.
   */
  private List<Problem> problems;

  /**
   * A string which identify this ws : it could be the processName.
   */
  protected QName processQName;

  /**
   * Default constructor.
   *
   * @param wsId -
   *            wsId
   * @param orchestraDirectory
   *            Absolute path to the orchestra directory.
   */
  protected WSDeployer(final QName processQName) {
    this.processQName = processQName;
    this.problems = new ArrayList<Problem>();
  }

  /**
   * Deploys the Web Services classes of the wsdl representing the interface of
   * a BPEL process.
   *
   * @param wsdlDef
   *            the wsdl definition of the process
   *
   * @return true, if deployed
   */
  public boolean deploy(List<Service> services) {
    try {
      checkSupport(services);
      if (this.problems.size() > 0) {
        printProblems();
        return false;
      } else if (services.size() == 0) {
        /*
         * throw new Exception( "No Service having at least one port defining a
         * soap adress was found in the top level wsdl file. " + "To be
         * compliant with WSI BAsic Profile, used services must be declared in
         * the top level wsdl file.");
         */
        return false;
      } else {

        checkWSEngineIsAvailable();
        // deploys all services
        deployServices(services);

      }
      return true;
    } catch (Exception e) {
      throw new OrchestraException("Exception while deploying services", e);
    }
  }

  /**
   * This method will undeploy the ws from the choosen ws engine.
   *
   * @param wsdlDef -
   *            Definition object that represents the ws to undeploy.
   * @return true, if undeploy
   */
  public boolean undeploy(List<Service> services) {
    try {
      checkSupport(services);
      if (this.problems.size() == 0) {
        checkWSEngineIsAvailable();
        // undeploys all services
        undeployServices(services);
      } else {
        printProblems();
        return false;
      }
      return true;
    } catch (Exception e) {
      throw new OrchestraException("Exception while undeploying services", e);
    }
  }

  /**
   * Print problems list.
   *
   */
  protected void printProblems() {
    log.severe("***************************************");
    log.severe("***************************************");
    for (Problem problem : this.problems) {
      log.severe("- " + problem);
    }
    log.severe("***************************************");
    log.severe("***************************************");
  }

  /**
   * Utility method to add a problem.
   */
  protected void addProblem(final Binding binding, final String mess,
      final String severity) {
    if (mess != null) {
      addProblem("The following problem was found on binding : "
          + binding.getQName() + "\n" + mess, severity);
    }
  }

  /**
   * Utility method to add a problem.
   */
  protected void addProblem(final Binding binding,
      final BindingOperation bindingOperation, final String mess,
      final String severity) {
    if (mess != null) {
      StringBuffer stringBuffer = new StringBuffer();
      stringBuffer.append("The following problem was found on binding : "
          + binding.getQName());
      stringBuffer.append("\nThe problem was found in operation : "
          + bindingOperation.getName());
      stringBuffer.append("\n");
      stringBuffer.append(mess);
      addProblem(stringBuffer.toString(), severity);
    }
  }

  /**
   * Utility method to add a problem.
   */
  protected void addProblem(final Binding binding, final Operation operation,
      final String mess, final String severity) {
    if (mess != null) {
      StringBuffer stringBuffer = new StringBuffer();
      stringBuffer.append("The following problem was found on binding : "
          + binding.getQName());
      stringBuffer.append("\nThe problem was found in operation : "
          + operation.getName());
      stringBuffer.append("\n");
      stringBuffer.append(mess);
      addProblem(stringBuffer.toString(), severity);
    }
  }

  /**
   * Utility method to add a problem.
   */
  protected void addProblem(final String mess, final String severity) {
    this.problems.add(new Problem(mess, null, severity));
  }

  /**
   * This method will deploy the specified WS Service on the choosen ws engine.
   * This method must be overriden by each ws engine Deployer.
   *
   * @param def -
   *            Definition object that represents a WS Service to deploy.
   */
  protected abstract void deployServices(final List<Service> services);

  /**
   * This method will undeploy the specified WS Service from the choosen ws
   * engine. This method must be overriden by each ws engine Deployer.
   *
   * @param def -
   *            Definition object that represents a WS Service to undeploy.
   */
  protected abstract void undeployServices(final List<Service> services);

  /**
   * This method will add a problem if Ws Engine is not Available. This method
   * must be overriden by each ws engine Deployer.
   */
  protected abstract void checkWSEngineIsAvailable();

  /**
   * This method will add a problem if Ws Engine does not support the specified
   * operationType. This method must be overriden by each ws engine Deployer.
   *
   * @param operationType -
   *            operationType to check
   */
  protected abstract String checkOperationType(final OperationType operationType);

  /**
   * This method will add a problem if Ws Engine does not support this soap
   * binding version : soap 1.1 or soap 1.2 for example. This method must be
   * overriden by each ws engine Deployer.
   *
   * @param soapBinding -
   *            soapBinding to check
   */
  protected abstract String checkSoapVersion(
      final ExtensibilityElement soapBinding);

  /**
   * This method will add a problem if Ws Engine does not support this soap
   * binding transport. This method must be overriden by each ws engine
   * Deployer.
   *
   * @param soapBinding -
   *            soapBinding to check
   */
  protected abstract String checkTransport(
      final ExtensibilityElement soapBinding);

  /**
   * This method will add a problem if Ws Engine does not support this style
   * (rpc/document). This method must be overriden by each ws engine Deployer.
   *
   * @param soapOperation -
   *            soapOperation to check
   */
  protected abstract String checkOperationStyle(final String operationStyle);

  /**
   * This method will add a problem if Ws Engine does not support this use
   * (encoded/literal). This method must be overriden by each ws engine
   * Deployer.
   *
   * @param soapBody -
   *            soapBody to check
   */
  protected abstract String checkUse(final ExtensibilityElement soapBody);

  /**
   * This method will add a problem if Ws Engine does not support an attribute
   * of the soapBody. For examples, parts attribute. This method must be
   * overriden by each ws engine Deployer.
   *
   * @param soapBody -
   *            soapBody to check
   */
  protected abstract String checkSoapBody(final ExtensibilityElement soapBody);

  protected abstract String checkSoapFault(final ExtensibilityElement soapFault);

  @SuppressWarnings("unchecked")
  public List<Service> checkSupport(List<Service> services) {
    List<Service> soapServices = new ArrayList<Service>();
    for (Service service : services) {
      // get all ports of this service
      Collection<Port> ports = (Collection<Port>) service.getPorts().values();
      for (Port port : ports) {
        // if port has a soap address, then it could be a correct one
        if (getAddressFromPort(port) != null) {
          soapServices.add(service);
        }
        Binding binding = port.getBinding();
        checkBinding(binding);
      }
    }
    return soapServices;
  }

  @SuppressWarnings("unchecked")
  protected void checkBinding(final Binding binding) {
    // check that this binding is a soap binding
    // if yes, check soapversion used and transport
    ExtensibilityElement soapBinding = getSoapBindingFromBinding(binding);
    if (soapBinding == null) {
      addProblem(binding, "this binding is not a soap binding.",
          Problem.SEVERITY_ERROR);
    } else {
      String mess = checkSoapVersion(soapBinding);
      addProblem(binding, mess, Problem.SEVERITY_ERROR);
      mess = checkTransport(soapBinding);
      addProblem(binding, mess, Problem.SEVERITY_ERROR);

      String soapBindingStyle = getSOAPStyle(soapBinding);
      if (soapBindingStyle == null) {
        soapBindingStyle = DEFAULT_SOAP_BINDING_STYLE;
      }
      // check operations
      List<BindingOperation> operations = (List<BindingOperation>) binding
          .getBindingOperations();
      for (BindingOperation bindingOperation : operations) {
        checkOperation(binding, bindingOperation, soapBindingStyle);
      }
    }
  }

  @SuppressWarnings("unchecked")
  protected void checkOperation(final Binding binding,
      final BindingOperation bindingOperation, final String soapBindingStyle) {
    Operation operation = bindingOperation.getOperation();
    // check : operationType, style, use
    OperationType operationType = operation.getStyle();
    String mess = checkOperationType(operationType);
    addProblem(binding, operation, mess, Problem.SEVERITY_ERROR);
    ExtensibilityElement soapOperation = getSoapOperationFromBindingOperation(bindingOperation);
    String operationStyle = soapBindingStyle;
    if (soapOperation != null) {
      String operationStyle2 = getSOAPStyle(soapOperation);
      if (operationStyle2 != null) {
        operationStyle = operationStyle2;
      }
    }
    if (operationStyle.equals("document")) {
      verifyThereIsOnlyOnePart(binding, operation);
      /*
       * Part part = (Part)
       * operation.getInput().getMessage().getParts().values().iterator().next();
       * String partName = part.getName(); QName elementQName =
       * part.getElementName(); QName typeQName = part.getTypeName(); if
       * (typeQName != null) { addProblem(binding, operation, "this operation is
       * of style document and its part is using " + "type attribute :
       * forbidden", Problem.SEVERITY_FATALERROR); } if (elementQName == null) {
       * addProblem(binding, operation, "this operation is of style document and
       * its part is not using " + "element attribute : forbidden",
       * Problem.SEVERITY_FATALERROR); }
       */
    }
    mess = checkOperationStyle(operationStyle);
    addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    BindingInput input = bindingOperation.getBindingInput();
    if (input != null) {
      ExtensibilityElement soapBody = getSoapBodyFromElement(input);
      verifySoapBody(binding, bindingOperation, soapBody);
    }
    BindingOutput output = bindingOperation.getBindingOutput();
    if (output != null) {
      ExtensibilityElement soapBody = getSoapBodyFromElement(output);
      verifySoapBody(binding, bindingOperation, soapBody);
    }
    Collection<BindingFault> faults = (Collection<BindingFault>) bindingOperation
        .getBindingFaults().values();
    for (BindingFault fault : faults) {
      ExtensibilityElement soapBody = getSoapFaultFromElement(fault);
      verifySoapFault(binding, bindingOperation, soapBody);
    }
  }

  /**
   * For this soapBody, if it exists, check the use and verify that all defined
   * attributes are supported.
   *
   * @param binding -
   *            binding in which the soapBody is declared
   * @param bindingOperation -
   *            bindingOperation in which the soapBody is declared
   * @param soapBody -
   *            soapBody to analyse
   */
  protected void verifySoapBody(final Binding binding,
      final BindingOperation bindingOperation,
      final ExtensibilityElement soapBody) {
    String mess;
    if (soapBody != null) {
      mess = checkUse(soapBody);
      addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
      mess = checkSoapBody(soapBody);
      addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    } else {
      mess = "There is a missing soapBody";
      addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    }
  }

  protected void verifySoapFault(final Binding binding,
      final BindingOperation bindingOperation,
      final ExtensibilityElement soapFault) {
    String mess;
    if (soapFault != null) {
      mess = checkUse(soapFault);
      addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
      mess = checkSoapFault(soapFault);
      addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    } else {
      mess = "There is a missing soapFault";
      addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    }
  }

  /**
   * This method must be called only if the given operation is defined in
   * document style. In this case, only one part can be defined. Else it is not
   * supported.
   *
   * @param binding -
   *            binding in which the operation is declared
   * @param operation -
   *            operation in which verifications will be performed
   */
  @SuppressWarnings("unchecked")
  protected void verifyThereIsOnlyOnePart(final Binding binding,
      final Operation operation) {
    int nbOfParts;
    Input input = operation.getInput();
    if (input != null) {
      nbOfParts = input.getMessage().getParts().size();
      if (nbOfParts > 1) {
        addProblem(binding, operation,
            "This operation is used in document style and is using an input message "
                + "containing more than one part", Problem.SEVERITY_ERROR);
      }
    }

    Output output = operation.getOutput();
    if (output != null) {
      nbOfParts = output.getMessage().getParts().size();
      if (nbOfParts > 1) {
        addProblem(
            binding,
            operation,
            "This operation is used in document style and is using an output message containing more "
                + "than one part", Problem.SEVERITY_ERROR);
      }
    }

    Collection<Fault> faults = operation.getFaults().values();
    for (Fault fault : faults) {
      nbOfParts = fault.getMessage().getParts().size();
      if (nbOfParts > 1) {
        addProblem(
            binding,
            operation,
            "This operation is used in document style and is using a fault message containing more "
                + "than one part", Problem.SEVERITY_ERROR);
      }
    }
  }

  @SuppressWarnings("unchecked")
  public static <T> T getFromList(List< ? > list,
      Class< ? extends T>... types) {
    for (Object o : list) {
      for (Class< ? extends T> clazz : types) {
        if (clazz.isAssignableFrom(o.getClass())) {
          return (T) o;
        }
      }
    }
    return null;
  }

  @SuppressWarnings("unchecked")
  protected String getAddressFromPort(final Port p) {
    // Get the endpoint for a port
    List<ExtensibilityElement> extensibilityList = p.getExtensibilityElements();
    for (ExtensibilityElement obj : extensibilityList) {
      if (obj instanceof SOAPAddress) {
        return ((SOAPAddress) obj).getLocationURI();
      } else if (obj instanceof SOAP12Address) {
        return ((SOAP12Address) obj).getLocationURI();
      }
    }
    // didn't find it
    return null;
  }

  @SuppressWarnings("unchecked")
  public static ExtensibilityElement getSoapBindingFromBinding(
      final Binding binding) {
    return WSDeployer.getFromList((List<ExtensibilityElement>) binding.getExtensibilityElements(),
        SOAPBinding.class, SOAP12Binding.class);
  }

  @SuppressWarnings("unchecked")
  public static ExtensibilityElement getSoapOperationFromBindingOperation(
      final BindingOperation bindingOperation) {
    return WSDeployer
        .getFromList((List<ExtensibilityElement>) bindingOperation.getExtensibilityElements(),
            SOAPOperation.class, SOAP12Operation.class);
  }

  public static String getSOAPStyle(final ExtensibilityElement element) {
    if (element instanceof SOAPBinding) {
      return ((SOAPBinding) element).getStyle();
    } else if (element instanceof SOAP12Binding) {
      return ((SOAP12Binding) element).getStyle();
    } else if (element instanceof SOAPOperation) {
      return ((SOAPOperation) element).getStyle();
    } else if (element instanceof SOAP12Operation) {
      return ((SOAP12Operation) element).getStyle();
    }
    return null;
  }

  @SuppressWarnings("unchecked")
  protected ExtensibilityElement getSoapBodyFromElement(
      final ElementExtensible element) {
    return WSDeployer.getFromList((List<ExtensibilityElement>) element.getExtensibilityElements(),
        SOAPBody.class, SOAP12Body.class);
  }

  @SuppressWarnings("unchecked")
  protected ExtensibilityElement getSoapFaultFromElement(
      final ElementExtensible element) {
    return WSDeployer.getFromList((List<ExtensibilityElement>) element.getExtensibilityElements(),
        SOAPFault.class, SOAP12Fault.class);
  }

  /**
   * @return the processQName
   */
  public QName getProcessQName() {
    return processQName;
  }

  /**
   * @param processQName
   *            the processQName to set
   */
  public void setProcessQName(QName processQName) {
    this.processQName = processQName;
  }

}
