package org.ow2.orchestra.wsdl;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.wsdl.Message;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.UnknownExtensibilityElement;
import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.jbpm.xml.Parse;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.lang.PropertyQuery;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.Namespaces;
import org.ow2.orchestra.util.XmlConstants;
import org.w3c.dom.Element;

/**
 * parses the wsdl4j structures and translates them into the BpelProcess model.
 *
 * @author Tom Baeyens
 */
public class BpelWsdlParser {

  /** log. */
  private static Logger log = Logger.getLogger(BpelWsdlParser.class.getName());

  /**
   *
   * @param documentBaseUri
   *            documentBaseUri
   * @param wsdlInputSource
   *            wsdl inputSource to analyse
   * @param parse
   *            parse object : use it to add problems
   * @param processQueryLanguage
   *            queryLanguage attribute value of the process element
   * @param styleSheetRepository
   *            styleSheetRepository
   * @return a WsdlInfos object that contains all necessary wsdl details
   */
  @SuppressWarnings("unchecked")
  public Definition parse(URL wsdlURL, Parse parse,
      String processQueryLanguage, BpelProcess bpelProcess, WsdlsInfos wsdlInfos) {
    Definition wsdlDefinition = WsdlUtil.readWsdl(wsdlURL);
    return parse(wsdlDefinition, parse, processQueryLanguage, bpelProcess,
        wsdlInfos);
  }

  public Definition parse(Definition wsdlDefinition, Parse parse,
      String processQueryLanguage, BpelProcess bpelProcess, WsdlsInfos wsdlInfos) {
    try {
      Namespaces namespaces = parse.findObject(Namespaces.class);
      Set<String> addedPrefixes = namespaces.addNamespaces(wsdlDefinition
          .getNamespaces());

      List<Definition> allDefinitions = new ArrayList<Definition>();
      retriveAllDefinitions(wsdlDefinition, allDefinitions);

      try {

        // add the wsdl port types to the process definition
        wsdlInfos.addPortTypes(wsdlDefinition.getAllPortTypes());
        // add the wsdl messages to the process definition
        wsdlInfos.addMessages(getAllMessages(allDefinitions));
        // add the wsdl services to the process definition
        wsdlInfos.addServices(wsdlDefinition.getAllServices());
        // add the wsdl bindings to the process definition
        wsdlInfos.addBindings(wsdlDefinition.getAllBindings());

        // extract the partner link types from the extensibility elements.
        List<ExtensibilityElement> extensibilityElements = getAllExtensibilityElements(allDefinitions);
        if (extensibilityElements != null) {
          for (ExtensibilityElement extensibilityElement : extensibilityElements) {
            if (extensibilityElement instanceof UnknownExtensibilityElement) {
              UnknownExtensibilityElement unknownExtensibilityElement = (UnknownExtensibilityElement) extensibilityElement;
              QName elementQName = extensibilityElement.getElementType();

              if (isPartnerLinkType(elementQName)) {
                Element partnerLintTypeElement = unknownExtensibilityElement
                    .getElement();
                PartnerLinkType partnerLinkType = createPartnerLinkType(
                    partnerLintTypeElement,
                    wsdlDefinition.getTargetNamespace(), parse);
                wsdlInfos.addPartnerLinkType(partnerLinkType);
              } else if (isVarProperty(elementQName)) {
                addProperty(wsdlDefinition.getTargetNamespace(),
                    (UnknownExtensibilityElement) unknownExtensibilityElement,
                    wsdlInfos);
              } else if (isVarPropertyAlias(elementQName)) {
                addPropertyAlias(wsdlDefinition.getTargetNamespace(),
                    (UnknownExtensibilityElement) unknownExtensibilityElement,
                    processQueryLanguage, wsdlInfos, bpelProcess, parse
                        .findObject(Environment.class));
              } else if (isServiceRef(elementQName)) {
                // TODO : manage service-ref
                throw new OrchestraException("serviceRef not yet implemented !");
              }
            } else {
              if (log.isLoggable(Level.FINE)) {
                log.fine("unknown extensibility element: "
                    + extensibilityElement);
              }
            }
          }
        }
      } finally {
        namespaces.removePrefixes(addedPrefixes);
      }
    } catch (Exception e) {
      parse.addProblem("couldn't parse wsdl "
          + wsdlDefinition.getDocumentBaseURI(), e);
    }
    return wsdlDefinition;
  }

  /**
   * Add a property defined by the given prop parameter to the given wsdlInfos.
   *
   * @param targetNamespace
   *            tns of the property
   * @param prop
   *            prop element to analyse
   * @param wsdlInfos
   *            wsdlInfos on which the property will be added
   */
  private void addProperty(String targetNamespace,
      UnknownExtensibilityElement prop, WsdlsInfos wsdlInfos) {
    Element propertyElement = prop.getElement();
    String name = propertyElement
        .getAttribute(XmlConstants.VARPROP_PROPERTY_NAME);
    Property property = new Property(new QName(targetNamespace, name));
    property.setType(XmlUtil.attributeQName(propertyElement,
        XmlConstants.VARPROP_PROPERTY_TYPE));
    property.setElement(XmlUtil.attributeQName(propertyElement,
        XmlConstants.VARPROP_PROPERTY_ELEMENT));
    wsdlInfos.addProperty(property);
  }

  /**
   * Add a propertyAlias defined by the given propAlias parameter to the given
   * wsdlInfos.
   *
   * @param targetNamespace
   *            tns of the property
   * @param propAlias
   *            propAlias element to analyse
   * @param processQueryLanguage
   *            queryLanguage attribute set on the process element
   * @param wsdlInfos
   *            wsdlInfos on which the propertyAlias will be added
   * @param styleSheetRepository
   *            styleSheetRepository
   * @param environment
   *            environment
   */
  private void addPropertyAlias(String targetNamespace,
      UnknownExtensibilityElement propAlias, String processQueryLanguage,
      WsdlsInfos wsdlInfos, BpelProcess bpelProcess, Environment environment) {
    Element propertyAliasElement = propAlias.getElement();
    PropertyAlias propertyAlias = new PropertyAlias(XmlUtil.attributeQName(
        propertyAliasElement, XmlConstants.VARPROP_PROPERTYALIAS_PROPNAME));
    propertyAlias.setMessageType(XmlUtil.attributeQName(propertyAliasElement,
        XmlConstants.VARPROP_PROPERTYALIAS_MESSAGETYPE));
    String partAttributeValue = propertyAliasElement
        .getAttribute(XmlConstants.VARPROP_PROPERTYALIAS_PART);
    if (!"".equals(partAttributeValue)) {
      propertyAlias.setPart(partAttributeValue);
    }
    propertyAlias.setType(XmlUtil.attributeQName(propertyAliasElement,
        XmlConstants.VARPROP_PROPERTYALIAS_TYPE));
    propertyAlias.setElement(XmlUtil.attributeQName(propertyAliasElement,
        XmlConstants.VARPROP_PROPERTY_ELEMENT));

    List<Element> queries = XmlUtil.elements(propertyAliasElement,
        XmlConstants.XMLNS_BPEL_2_0_VARPROP,
        XmlConstants.VARPROP_PROPERTYALIAS_QUERY);
    if (queries.size() != 0) {
      Element queryElement = (Element) queries.get(0);

      String queryLanguage = XmlUtil.attribute(queryElement,
          XmlConstants.VARPROP_PROPERTYALIAS_QUERY_LANGUAGE);
      if (queryLanguage == null) {
        queryLanguage = processQueryLanguage;
      }
      /*
       * Query query = new Query(bpelProcess,
       * BpelUtil.getAllNameSpaces(queryElement), queryElement.getTextContent(),
       * null, null, queryLanguage); propertyAlias.setQuery(query);
       */
      PropertyQuery propertyQuery = new PropertyQuery();
      propertyQuery.setLanguage(queryLanguage);
      propertyQuery.setNamespaces(BpelUtil.getAllNameSpaces(queryElement));
      propertyQuery.setText(queryElement.getTextContent());
      propertyAlias.setQuery(propertyQuery);
    }
    wsdlInfos.addPropertyAlias(propertyAlias);
  }

  /**
   * return true if the given qn is conformant with a bpel plnk.
   *
   * @param qn
   *            qname to analyse
   * @return true if the given qn is conformant with a bpel plnk.
   */
  protected boolean isPartnerLinkType(QName qn) {
    return XmlConstants.XMLNS_BPEL_2_0_PLNKTYPE.equals(qn.getNamespaceURI())
        && qn.getLocalPart().equals("partnerLinkType");
  }

  /**
   * return true if the given qn is conformant with a bpel serviceRef.
   *
   * @param qn
   *            qname to analyse
   * @return true if the given qn is conformant with a bpel serviceRef.
   */
  protected boolean isServiceRef(QName qn) {
    return XmlConstants.XMLNS_BPEL_2_0_SERVICEREF.equals(qn.getNamespaceURI())
        && qn.getLocalPart().equals("service-ref");
  }

  /**
   * return true if the given qn is conformant with a bpel property.
   *
   * @param qn
   *            qname to analyse
   * @return true if the given qn is conformant with a bpel property.
   */
  protected boolean isVarProperty(QName qn) {
    return XmlConstants.XMLNS_BPEL_2_0_VARPROP.equals(qn.getNamespaceURI())
        && qn.getLocalPart().equals("property");
  }

  /**
   * return true if the given qn is conformant with a bpel propertyAlias.
   *
   * @param qn
   *            qname to analyse
   * @return true if the given qn is conformant with a bpel propertyAlias.
   */
  protected boolean isVarPropertyAlias(QName qn) {
    return XmlConstants.XMLNS_BPEL_2_0_VARPROP.equals(qn.getNamespaceURI())
        && qn.getLocalPart().equals("propertyAlias");
  }

  /**
   * Create a new PartnerLinkType base on the given partnerLinkTypeElement.
   *
   * @param partnerLinkTypeElement
   *            dom element to analyse
   * @param targetNamespace
   *            targetNamespace in which the plnk is defined
   * @param parse
   *            parse object to use to add problems
   * @return the created PartnerLinkType
   */
  private PartnerLinkType createPartnerLinkType(Element partnerLinkTypeElement,
      String targetNamespace, Parse parse) {
    String name = partnerLinkTypeElement.getAttribute("name");
    QName partnerLinkTypeName = new QName(targetNamespace, name);
    PartnerLinkType partnerLinkType = new PartnerLinkType(partnerLinkTypeName);
    List<Element> roleElements = XmlUtil.elements(partnerLinkTypeElement,
        "role");
    if (roleElements != null) {
      Namespaces namespaces = parse.findObject(Namespaces.class);
      for (Element roleElement : roleElements) {
        String roleName = roleElement.getAttribute("name");
        try {
          // NOTE it seems that the values in the WSDL have to be resolved
          // programmatically with the
          // javax.wsdl.Definition.getNamespaces().
          //
          // Regrettably, this is a different technique then the one in parsing
          // the bpel document, where
          // it is necessary to first set the namespace awareness to true and
          // then use the
          // org.w3c.dom.Element.lookupNamespaceURI(prefix);

          XmlUtil.NamespaceValue portTypeValue = XmlUtil
              .attributeNamespaceValue(roleElement, "portType");
          String uri = namespaces.getUri(portTypeValue.prefix);
          QName portTypeQName = new QName(uri, portTypeValue.localPart);
          partnerLinkType.addRole(roleName, portTypeQName);
        } catch (OrchestraException e) {
          parse.addProblem(e.getMessage());
        }
      }
    }
    return partnerLinkType;
  }

  private void retriveAllDefinitions(Definition def,
      List<Definition> allDefinitions) {
    allDefinitions.add(def);
    Map<String, List<Import>> importsMap = def.getImports();
    if (importsMap != null) {
      Collection<List<Import>> importLists = importsMap.values();
      if (importLists != null) {
        for (List<Import> importList : importLists) {
          if (importList != null) {
            for (Import i : importList) {
              allDefinitions.add(i.getDefinition());
              retriveAllDefinitions(i.getDefinition(), allDefinitions);
            }
          }
        }
      }
    }
  }

  private Map<QName, Message> getAllMessages(List<Definition> definitions) {
    Map<QName, Message> messages = new HashMap<QName, Message>();
    for (Definition definition : definitions) {
      messages.putAll(definition.getMessages());
    }
    return messages;
  }

  private List<ExtensibilityElement> getAllExtensibilityElements(
      List<Definition> definitions) {
    List<ExtensibilityElement> extensibilityElements = new ArrayList<ExtensibilityElement>();
    for (Definition definition : definitions) {
      extensibilityElements.addAll(definition.getExtensibilityElements());
    }
    return extensibilityElements;
  }
}
