/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.wsdl;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.xml.WSDLLocator;

import org.ow2.orchestra.exception.OrchestraException;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;

public class WSDLLocatorImpl implements WSDLLocator {

  private static Logger log = Logger.getLogger(WSDLLocatorImpl.class.getName());

  private EntityResolver entityResolver;

  private URL wsdlURL;

  private String latestImportURI;

  public WSDLLocatorImpl(EntityResolver entityResolver, URL wsdlFile) {
    if (wsdlFile == null) {
      throw new IllegalArgumentException("WSDL file argument cannot be null");
    }
    this.entityResolver = entityResolver;
    this.wsdlURL = wsdlFile;
  }

  public InputSource getBaseInputSource() {
    if (log.isLoggable(Level.FINE)) {
      log.fine("getBaseInputSource [wsdlUrl=" + wsdlURL + "]");
    }
    try {
      InputStream is = wsdlURL.openStream();
      return new InputSource(is);
    } catch (IOException e) {
      throw new OrchestraException("Cannot access wsdl from [" + wsdlURL
          + "], ", e);
    }
  }

  public String getBaseURI() {
    return wsdlURL.toExternalForm();
  }

  public InputSource getImportInputSource(String parent, String resource) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("getImportInputSource [parent=" + parent + ",resource="
          + resource + "]");
    }
    URL parentURL = null;
    try {
      parentURL = new URL(parent);
    } catch (MalformedURLException e) {
      if (log.isLoggable(Level.FINE)) {
        log.fine("Not a valid URL: " + parent);
      }
      return null;
    }
    String wsdlImport = null;
    String external = parentURL.toExternalForm();

    // An external URL
    if (resource.startsWith("http://") || resource.startsWith("https://")) {
      wsdlImport = resource;
    } else if (resource.startsWith("/")) {
      // Absolute path
      String beforePath = external.substring(0, external.indexOf(parentURL
          .getPath()));
      wsdlImport = beforePath + resource;
    } else {
      // A relative path
      String parentDir = external.substring(0, external.lastIndexOf("/"));

      // remove references to current dir
      while (resource.startsWith("./")) {
        resource = resource.substring(2);
      }

      // remove references to parentdir
      while (resource.startsWith("../")) {
        parentDir = parentDir.substring(0, parentDir.lastIndexOf("/"));
        resource = resource.substring(3);
      }

      wsdlImport = parentDir + "/" + resource;
    }

    try {
      if (log.isLoggable(Level.FINE)) {
        log.fine("Trying to resolve: " + wsdlImport);
        log.fine("Trying to resolve: entityResolver = " + entityResolver);
      }
      InputSource inputSource = entityResolver.resolveEntity(wsdlImport,
          wsdlImport);
      if (log.isLoggable(Level.FINE)) {
        log.fine("Trying to resolve: inputSource = " + inputSource);
      }
      if (inputSource != null) {
        latestImportURI = wsdlImport;
        if (log.isLoggable(Level.FINE)) {
          log.fine("Trying to resolve: latestImportURI = " + latestImportURI);
        }
      } else {
        throw new IllegalArgumentException("Cannot resolve imported resource: "
            + wsdlImport);
      }
      return inputSource;
    } catch (RuntimeException rte) {
      throw rte;
    } catch (Exception e) {
      throw new OrchestraException("Cannot access imported wsdl [" + wsdlImport
          + "], ", e);
    }
  }

  public String getLatestImportURI() {
    return latestImportURI;
  }

  public void close() {
    // TODO Auto-generated method stub

  }
}
