/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.wsdl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.wsdl.Binding;
import javax.wsdl.Message;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.ow2.orchestra.exception.OrchestraException;

/**
 *
 * @author Charles Souillard
 */
public class WsdlsInfos implements Serializable {

  /**   */
  private static final long serialVersionUID = 2688615599128088745L;

  protected long dbid;

  /** portTypes of the corresponding wsdl. */
  protected Map<QName, PortType> portTypes;

  /** messages of the corresponding wsdl. */
  protected Map<QName, Message> messages;

  /** services of the corresponding wsdl. */
  protected Map<QName, Service> services;

  /** binding of the corresponding wsdl. */
  protected Map<QName, Binding> bindings;

  /** partnerLinkTypes of the corresponding wsdl. */
  protected Map<QName, PartnerLinkType> partnerLinkTypes;

  /** properties of the corresponding wsdl. */
  protected Map<QName, Property> properties;

  /** propertyAliases of the corresponding wsdl. */
  protected List<PropertyAlias> propertyAliases;

  /**
   * Default constructor.
   */
  public WsdlsInfos() {
  }

  /**
   * Add a PartnerLinkType.
   *
   * @param partnerLinkType
   *            partnerLinkType to add
   */
  public void addPartnerLinkType(PartnerLinkType partnerLinkType) {
    if (partnerLinkTypes == null) {
      partnerLinkTypes = new HashMap<QName, PartnerLinkType>();
    }
    partnerLinkTypes.put(partnerLinkType.getName(), partnerLinkType);
  }

  /**
   * Return the partnerLinkType having the given QName.
   *
   * @param partnerLinkTypeName
   *            partnerLinkTypeName
   * @return the partnerLinkType having the given QName.
   */
  public PartnerLinkType getPartnerLinkType(QName partnerLinkTypeName) {
    if (partnerLinkTypes != null) {
      return partnerLinkTypes.get(partnerLinkTypeName);
    }
    return null;
  }

  /**
   * Add all portTypes.
   *
   * @param portTypesToAdd
   *            portTypes to add
   */
  public void addPortTypes(Map<QName, PortType> portTypesToAdd) {
    if (this.portTypes == null) {
      this.portTypes = portTypesToAdd;
    } else {
      this.portTypes.putAll(portTypesToAdd);
    }
  }

  /**
   * Return the portType having the given QName.
   *
   * @param portTypeQName
   *            portTypeName
   * @return the portType having the given QName.
   */
  public PortType getPortType(QName portTypeQName) {
    return portTypes.get(portTypeQName);
  }

  /**
   * Add all messages.
   *
   * @param messagesToAdd
   *            messages to add
   */
  public void addMessages(Map<QName, Message> messagesToAdd) {
    if (this.messages == null) {
      this.messages = messagesToAdd;
    } else {
      this.messages.putAll(messagesToAdd);
    }
  }

  /**
   * Return the message having the given QName.
   *
   * @param messageName
   *            messageName
   * @return the message having the given QName.
   */
  public Message getMessage(QName messageName) {
    return messages.get(messageName);
  }

  /**
   * Add all services.
   *
   * @param servicesToAdd
   *            services to add
   */
  public void addServices(Map<QName, Service> servicesToAdd) {
    if (this.services == null) {
      this.services = servicesToAdd;
    } else {
      this.services.putAll(servicesToAdd);
    }
  }

  /**
   * Return the service having the given QName.
   *
   * @param serviceName
   *            serviceName
   * @return the service having the given QName.
   */
  public Service getService(QName serviceName) {
    return services.get(serviceName);
  }

  /**
   * Return the service that contains the given port.
   *
   * @param port
   *            port
   * @return the service that contains the given port.
   */
  public Service getServiceOfPort(Port port) {
    String portName = port.getName();
    for (Service service : services.values()) {
      if (service.getPort(portName) != null) {
        return service;
      }
    }
    return null;
  }

  /**
   * Return the service that contains the given port.
   *
   * @param port
   *            port
   * @return the service that contains the given port.
   */
  public List<Service> getServicesOfPortType(QName portTypeQName) {
    List<Service> matchinServices = new ArrayList<Service>();
    for (Service service : services.values()) {
      for (Port port : (Collection<Port>) service.getPorts().values()) {
        if (port.getBinding().getPortType().getQName().equals(portTypeQName)) {
          matchinServices.add(service);
        }
      }
    }
    if (matchinServices.size() > 0) {
      return matchinServices;
    }
    return null;
  }

  /**
   * Return all ports implementing the given PortType.
   *
   * @param portTypeQName
   *            portType
   * @return all ports implementing the given PortType
   */
  @SuppressWarnings("unchecked")
  public List<Port> getPorts(QName portTypeQName) {
    List<Port> ports = new ArrayList<Port>();
    for (Service service : services.values()) {
      Map<QName, Port> servicePorts = (Map<QName, Port>) service.getPorts();
      for (Port port : servicePorts.values()) {
        if (port.getBinding().getPortType().getQName().equals(portTypeQName)) {
          ports.add(port);
        }
      }
    }
    return ports;
  }


  /**
   * Add all bindings.
   *
   * @param bindingsToAdd
   *            bindings to add
   */
  public void addBindings(Map<QName, Binding> bindingsToAdd) {
    if (this.bindings == null) {
      this.bindings = bindingsToAdd;
    } else {
      this.bindings.putAll(bindingsToAdd);
    }
  }

  /**
   * Return the binding having the given QName.
   *
   * @param bindingName
   *            bindingName
   * @return the binding having the given QName.
   */
  public Binding getBinding(QName bindingName) {
    return bindings.get(bindingName);
  }

  /**
   * Return the binding that contains the given port.
   *
   * @param port
   *            port
   * @return the binding that contains the given port.
   */
  public List<Binding> getBindingsOfPortType(QName portTypeQName) {
    List<Binding> matchinBindings = new ArrayList<Binding>();
    for (Binding binding : bindings.values()) {
      if (binding.getPortType().getQName().equals(portTypeQName)) {
        matchinBindings.add(binding);
      }
    }
    if (matchinBindings.size() > 0) {
      return matchinBindings;
    }
    return null;
  }

  /**
   * Add the given property.
   *
   * @param property
   *            property to add
   */
  public void addProperty(Property property) {
    if (this.properties == null) {
      this.properties = new HashMap<QName, Property>();
    }
    this.properties.put(property.getName(), property);
  }

  /**
   * Return the property having the given QName.
   *
   * @param propertyQName
   *            propertyQName
   * @return the property having the given QName.
   */
  public Property getProperty(QName propertyQName) {
    if (properties == null) {
      return null;
    }
    return properties.get(propertyQName);
  }

  /**
   * Add the given PrpertyAlias.
   *
   * @param propertyAlias
   *            propertyAlias to add
   */
  public void addPropertyAlias(PropertyAlias propertyAlias) {
    if (propertyAliases == null) {
      propertyAliases = new ArrayList<PropertyAlias>();
    }
    this.propertyAliases.add(propertyAlias);
  }

  /**
   * Get the propertyAlias associated with the propertyName and the messageType
   * given in parameter. Or null if there is no propertyAlias associated with
   * this propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param messageType :
   *            messageType of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithMessageType(QName propertyName,
      QName messageType) {
    for (PropertyAlias propertyAlias : propertyAliases) {
      if (propertyAlias.getMessageType() != null
          && propertyAlias.getMessageType().equals(messageType)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }

  /**
   * Return all propertyAlias using the given messageType.
   *
   * @param messageType
   *            messageType
   * @return all propertyAlias using the given messageType.
   */
  public List<PropertyAlias> getPropertyAliasesWithMessageType(QName messageType) {
    List<PropertyAlias> matchingPropertyAliases = new ArrayList<PropertyAlias>();
    for (PropertyAlias propertyAlias : propertyAliases) {
      if (propertyAlias.getMessageType() != null
          && propertyAlias.getMessageType().equals(messageType)) {
        matchingPropertyAliases.add(propertyAlias);
      }
    }
    return matchingPropertyAliases;
  }

  /**
   * Return all propertyAliases using the given messageType and properties.
   *
   * @param propertiesToScan
   *            properties we need to scan.
   * @param messageType
   *            messageType that returned propertyAlias should conform
   * @return all propertyAliases using the given messageType and properties.
   */
  public List<PropertyAlias> getPropertyAliases(
      Collection<Property> propertiesToScan, QName messageType) {
    List<PropertyAlias> result = new ArrayList<PropertyAlias>();
    for (Property property : propertiesToScan) {
      PropertyAlias propertyAlias = getPropertyAliasWithMessageType(property
          .getName(), messageType);
      if (propertyAlias == null) {
        throw new OrchestraException("propertyAlias is null, property = "
            + property);
      }
      result.add(propertyAlias);
    }
    return result;
  }

  /**
   * Get the propertyAlias associated with the propertyName and the element
   * given in parameter. Or null if there is no propertyAlias associated with
   * this propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param element :
   *            element of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithElement(QName propertyName,
      QName element) {
    for (PropertyAlias propertyAlias : propertyAliases) {
      if (propertyAlias.getElement() != null
          && propertyAlias.getElement().equals(element)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }

  /**
   * Get the propertyAlias associated with the propertyName and the type given
   * in parameter. Or null if there is no propertyAlias associated with this
   * propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param type :
   *            type of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithType(QName propertyName, QName type) {
    for (PropertyAlias propertyAlias : propertyAliases) {
      if (propertyAlias.getType() != null
          && propertyAlias.getType().equals(type)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }
}
