/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.test;

import java.io.File;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.wsdl.Operation;
import javax.wsdl.PortType;
import javax.xml.namespace.QName;

import junit.framework.TestCase;

import org.jbpm.env.Environment;
import org.jbpm.env.EnvironmentFactory;
import org.jbpm.env.session.PvmDbSession;
import org.jbpm.tx.Transaction;
import org.jbpm.util.ReflectUtil;
import org.jbpm.xml.Parse;
import org.jbpm.xml.XmlUtil;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.deployment.Deployer;
import org.ow2.orchestra.deployment.Deployment;
import org.ow2.orchestra.env.DefaultEnvXMLGenerator;
import org.ow2.orchestra.env.EnvFactoryRepository;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.parsing.BpelImportParser;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.PartnerLinkRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.MessageCarrier;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.itf.BpelPersistenceService;
import org.ow2.orchestra.services.itf.Invoker;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.services.itf.TestValidator;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.var.Message;
import org.w3c.dom.Element;
import org.w3c.dom.Text;

/**
 * @author Charles Souillard
 */
public abstract class BpelTestCase extends TestCase {

  /** Class used to generate the environment factory */
  public static Class< ? extends DefaultEnvXMLGenerator> envGeneratorClass = DefaultEnvXMLGenerator.class;

  static {
    try {
      String envGeneratorClassName = System.getProperty("env.generator");
      if (envGeneratorClassName != null) {
        Class< ? > clazz = ReflectUtil
            .loadClass(BpelTestCase.class.getClassLoader(),
                envGeneratorClassName);
        if (clazz != null) {
          envGeneratorClass = (Class<? extends DefaultEnvXMLGenerator>) clazz;
        }
      }
    } catch (Exception e) {
      throw new RuntimeException(
          "couldn't initialize environment generator properly", e);
    }
  }

  static {
    try {
      InputStream inputStream = BpelTestCase.class.getClassLoader()
          .getResourceAsStream("logging.properties");
      if (inputStream != null) {
        LogManager.getLogManager().readConfiguration(inputStream);
      }
    } catch (Exception e) {
      throw new RuntimeException("couldn't initialize logging properly", e);
    }
  }

  /** log. */
  private static Logger log = Logger.getLogger(BpelTestCase.class.getName());

  private QName processQName;

  public class CallResult {

    protected BpelExecution instance;

    protected MessageCarrier messageCarrier;

    protected List<VariableRuntime> variables;

    protected List<PartnerLinkRuntime> partnerLInks;

    protected List<String> events;

    protected long endTime;

    public CallResult(BpelExecution instance, MessageCarrier messageCarrier,
        List<VariableRuntime> variables, List<PartnerLinkRuntime> partnerLInks,
        List<String> events, long endTime) {
      this.instance = instance;
      this.messageCarrier = messageCarrier;
      this.variables = variables;
      this.partnerLInks = partnerLInks;
      this.events = events;
      this.endTime = endTime;
    }

    public MessageCarrier getMessageCarrier() {
      return messageCarrier;
    }

    public List<VariableRuntime> getVariables() {
      return variables;
    }

    public List<PartnerLinkRuntime> getPartnerLInks() {
      return partnerLInks;
    }

    public List<String> getEvents() {
      return events;
    }

    public long getEndTime() {
      return endTime;
    }

    public BpelExecution getInstance() {
      return Environment.getCurrent().get(Repository.class).getUpdatedExecution(instance);
    }

    public String toString() {
      return CallResult.class.getName() + "[instance: " + instance
          + ", messageCarrier: " + messageCarrier + ", variables: " + variables
          + ", partnerLInks: " + partnerLInks + ", events: " + events
          + ", endTime: " + endTime + "]";
    }

  }

  /*
   * protected BpelTestCase() { }
   */
  protected BpelTestCase(String processNamespace, String processName) {
    this.processQName = new QName(processNamespace, processName);
  }

  public EnvironmentFactory getEnvironmentFactory() {
    if (EnvFactoryRepository.get() == null) {
      try {
        DefaultEnvXMLGenerator envGenerator = envGeneratorClass.newInstance();
        envGenerator
            .setApplicationEntry(Invoker.DEFAULT_KEY, TestInvoker.class);
        envGenerator.setApplicationEntry(TestValidator.DEFAULT_KEY,
            TestValidatorImpl.class);

        EnvFactoryRepository.set(envGenerator.createEnvironmentFactory());
      } catch (Exception e) {
        throw new OrchestraException(e);
      }
    }
    return EnvFactoryRepository.get();
  }

  protected void deleteInstance(CallResult callResult) {
    BpelExecution instance = callResult.getInstance();
    if (instance != null) {
      Environment.getCurrent().get(Repository.class).removeInstance(instance);
      MutexRepository.removeInstanceMutex(instance.getProcessInstance()
          .getUuid());
    }
  }

  protected void deleteInstances(int nb) {
    Environment environment = null;
    try {
      environment = getEnvironmentFactory().openEnvironment();
      List<BpelExecution> execs = environment.get(Repository.class).getInstances(getProcessQName());
      assertEquals(nb, execs.size());
      for (int i = 0; i < nb; i++) {
        environment.get(Repository.class).removeInstance(execs.get(i), true);
      }
    } catch (Exception e) {
      e.printStackTrace();
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
  }
  /**
   * Deploy method using many wsdls.
   *
   * @param bpel
   *            bpel of the process
   * @param wsdls
   *            wsdl of the process + partner wsdls
   */
  protected void deploy(URL bpel, URL wsdl) {
    Environment environment = null;
    try {
      environment = getEnvironmentFactory().openEnvironment();

      Deployer deployer = new Deployer();
      assertExists("Bpel file : " + bpel + " doesn't exist", bpel);
      List<URL> wsdls = null;
      if (wsdl != null) {
        wsdls = new ArrayList<URL>();
        wsdls.add(wsdl);
      }
      Deployment bpelDeployment = BpelImportParser
          .parseBpelFromUrl(bpel, wsdls);

      BpelProcess bpelProcess = deployer.deploy(bpelDeployment, environment);
      assertNotNull(bpelProcess);
    } catch (Exception e) {
      if (environment != null) {
        Transaction t = environment.getTransaction();
        if (t != null) {
          t.setRollbackOnly();
        }
      }
      throw new OrchestraException(e);
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
  }

  public void undeploy() {
    Environment environment = null;
    try {
      environment = getEnvironmentFactory().openEnvironment();
      Deployer deployer = new Deployer();
      deployer.undeploy(processQName, environment);
    } catch (Exception e) {
      if (environment != null) {
        Transaction t = environment.getTransaction();
        if (t != null) {
          t.setRollbackOnly();
        }
      }
      throw new OrchestraException(e);
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
  }

  /**
   * Call the process on the endpoint corresponding to the given paramater and
   * don't wait for a returned message.
   *
   * @param waitForResponse
   *            if true wait until a reply message was received
   * @param op
   *            operation to call
   * @param parts
   *            parts of the message to send to initiate the call
   * @param plnkName
   *            partnerLink name to use for the call
   * @param portTypeQName
   *            portType defining the called operation
   * @return created instance
   */
  public CallResult call(HashMap<String, Element> parts, QName portTypeQName,
      String operationName) {
    Environment environment = null;
    MessageCarrier messageCarrier = null;
    Message message = new Message(parts);
    OperationKey operationKey = new OperationKey(processQName, portTypeQName,
        operationName);

    try {
      environment = getEnvironmentFactory().openEnvironment();

      BpelProcess bpelProcess = environment.get(Repository.class).getProcess(
          processQName);



      boolean requestResponse = isRequestResponse(bpelProcess, portTypeQName,
          operationName);
      if (requestResponse) {
        messageCarrier = new MessageCarrier();
      }
    } catch (Exception e) {
      if (environment != null) {
        Transaction t = environment.getTransaction();
        if (t != null) {
          t.setRollbackOnly();
        }
      }
      throw new OrchestraException(e);
    } finally {
      if (environment != null) {
        environment.close();
      }
    }

    BpelExecution bpelExecution = Receiver.handle(message, operationKey,
        getEnvironmentFactory(), messageCarrier);
    BpelExecution instance = null;
    if (bpelExecution != null) {
      instance = bpelExecution.getProcessInstance();
    }
    long endTime = System.currentTimeMillis();

    try {
      environment = getEnvironmentFactory().openEnvironment();
      TestValidator testValidator = (TestValidator) environment
          .get(TestValidator.DEFAULT_KEY);

      CallResult callResult = new CallResult(instance, messageCarrier,
          testValidator.getRecordedVariables(instance, processQName),
          testValidator.getRecordedPartnerLinks(instance, processQName),
          testValidator.getRecordedEvents(instance, processQName), endTime);

      testValidator.clean(instance);

      return callResult;
    } catch (Exception e) {
      if (environment != null) {
        Transaction t = environment.getTransaction();
        if (t != null) {
          t.setRollbackOnly();
        }
      }
      throw new OrchestraException(e);
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
  }

  protected Object getVariableValue(CallResult result, String variableName) {
    BpelPersistenceService persistenceService = Environment.getCurrent().get(
        BpelPersistenceService.class);
    for (VariableRuntime variableRuntime : result.variables) {
      if (persistenceService != null) {
        variableRuntime = persistenceService.get(VariableRuntime.class,
            variableRuntime.getDbid());
      }
      if (variableRuntime.getName().equals(variableName)) {
        return variableRuntime.getValue();
      }
    }
    return null;

  }

  protected PartnerLinkRuntime getPartnerLink(CallResult result,
      String partnerLinkName) {
    PvmDbSession persistenceService = Environment.getCurrent().get(
        PvmDbSession.class);
    for (PartnerLinkRuntime partnerLinkRuntime : result.partnerLInks) {
      if (persistenceService != null) {
        partnerLinkRuntime = persistenceService.get(PartnerLinkRuntime.class,
            partnerLinkRuntime.getDbid());
      }
      if (partnerLinkRuntime.getName().equals(partnerLinkName)) {
        return partnerLinkRuntime;
      }
    }
    return null;
  }

  protected List<String> getEvents(CallResult result) {
    return result.events;
  }

  /**
   * Return a file based on the given resource.
   *
   * @param resourcePath
   *            resourcePath of the file to return
   * @return a file based on the given resource.
   */
  protected File getFile(String resourcePath) {
    File f = null;
    try {
      f = new File(this.getClass().getResource(resourcePath).toURI().getPath());
    } catch (Exception e) {
      throw new OrchestraException(e);
    }
    return f;
  }

  /**
   * Return true if the operation corresponding to the given parameters is a
   * requestResponse op.
   *
   * @param portTypeQName
   *            portTypeName
   * @param operationName
   *            operationName
   * @return true if the operation corresponding to the given parameters is a
   *         requestResponse op.
   */
  @SuppressWarnings("unchecked")
  protected boolean isRequestResponse(BpelProcess bpelProcess,
      QName portTypeQName, String operationName) {
    PortType pt = bpelProcess.getWsdlInfos().getPortType(portTypeQName);
    if (pt == null) {
      throw new OrchestraException(
          "Process '"
              + bpelProcess.getName()
              + "', portType '"
              + portTypeQName
              + "' not found : cannot define if it is a request/response call or a oneWay call.");
    }
    List<Operation> operations = (List<Operation>) pt.getOperations();
    for (Operation op : operations) {
      if (op.getName().equals(operationName)) {
        return op.getOutput() != null;
      }
    }
    throw new OrchestraException(
        "Operation '"
            + operationName
            + "' not found in portType '"
            + portTypeQName
            + "' : cannot define if it is a request/response call or a oneWay call.");
  }

  /**
   * Assert that the given object is an Element having the given localName
   * definied in the process wsdl targetNamespace.
   *
   * @param o
   *            object to assert
   * @param rootElementLocalName
   *            expected Element rootElement local name
   * @return root element of the given object (if it is an Element)
   */
  protected Element assertTopElementNS(Object o, String rootElementLocalName) {
    assertNotNull(o);
    assertTrue("o is not an Element", o instanceof Element);
    Element root = (Element) o;
    assertEquals(rootElementLocalName, root.getLocalName());
    assertEquals(processQName.getNamespaceURI(), root.getNamespaceURI());
    return root;
  }

  /**
   * Assert that the given element has only one child having the given tagName
   * and the given text content.
   *
   * @param el
   *            element to check
   * @param tagName
   *            expected child tagName
   * @param textContent
   *            expected child text content
   * @return the found child (if any)
   */
  protected Element assertElementNSContentAndNoChild(Element el,
      String tagName, String textContent) {
    List<Element> children = XmlUtil.elements(el, processQName
        .getNamespaceURI(), tagName);
    assertEquals(1, children.size());
    Element child = (Element) children.get(0);
    assertEquals(textContent, child.getTextContent());
    assertNoElementChild(child);
    return child;
  }

  /**
   * assert true if the given element has either : no child either one which is
   * a Text.
   *
   * @param el
   *            element to check
   */
  protected void assertNoElementChild(Element el) {
    if (el.getChildNodes().getLength() == 1) {
      assertTrue(el.getChildNodes().item(0) instanceof Text);
    } else if (el.getChildNodes().getLength() == 2) {
      assertTrue(false);
    }
  }

  /**
   * String assertion.
   *
   * @param o
   *            object to assert
   * @param expected
   *            expected value
   * @return root element of the given object (if it is an Element)
   */
  protected Element assertString(Object o, String expected) {
    assertNotNull(o);
    assertTrue("o is not an Element", o instanceof Element);
    Element root = (Element) o;
    assertEquals(expected, root.getTextContent());
    return root;
  }

  /**
   * EndpointReference assertion.
   *
   * @param o
   *            serviceRef to assert
   * @param expectedAddress
   *            expectedAddress
   * @param expectedService
   *            expectedAddress
   */
  protected void assertServiceRefElement(Object o, String expectedAddress,
      QName expectedService) {
    assertNotNull(o);
    assertTrue("o is not an Element", o instanceof Element);
    Element serviceRef = (Element) o;
    System.out.println("XMLNS_BPEL_2_0_SERVICEREF    = '"
        + XmlConstants.XMLNS_BPEL_2_0_SERVICEREF);
    System.out.println("serviceRef.getNamespaceURI() = '"
        + serviceRef.getNamespaceURI());
    assertEquals(XmlConstants.XMLNS_BPEL_2_0_SERVICEREF, serviceRef
        .getNamespaceURI());
    assertEquals(XmlConstants.SERVICE_REF_TAG_NAME, serviceRef.getLocalName());
    assertEndpointReference(serviceRef, expectedAddress, expectedService);
  }

  /**
   * EndpointReference assertion.
   *
   * @param o
   *            serviceRef to assert
   * @param expectedAddress
   *            expectedAddress
   * @param expectedService
   *            expectedAddress
   */
  protected void assertServiceRefType(Object o, String expectedAddress,
      QName expectedService) {
    assertNotNull(o);
    assertTrue("o is not an Element", o instanceof Element);
    Element defaultElement = (Element) o;
    assertEquals(XmlConstants.ELT_DEFAULT_NS, defaultElement.getNamespaceURI());
    assertEquals(XmlConstants.ELT_DEFAULT_ELEMENT, defaultElement
        .getLocalName());
    assertEndpointReference(defaultElement, expectedAddress, expectedService);
  }

  /**
   * EndpointReference assertion.
   *
   * @param serviceRef
   *            serviceRef to assert
   * @param expectedAddress
   *            expectedAddress
   * @param expectedService
   *            expectedAddress
   */
  protected void assertEndpointReference(Element serviceRef,
      String expectedAddress, QName expectedService) {
    List<Element> serviceRefChildren = XmlUtil.elements(serviceRef,
        XmlConstants.ADDRESSING_REF_NAMESPACE, XmlConstants.EPR_TAG_NAME);
    assertEquals(1, serviceRefChildren.size());
    Element endpointReference = (Element) serviceRefChildren.get(0);

    List<Element> addressChildren = XmlUtil.elements(endpointReference,
        XmlConstants.ADDRESSING_REF_NAMESPACE, XmlConstants.ADDRESS_TAG_NAME);
    assertEquals(1, addressChildren.size());
    Element address = (Element) addressChildren.get(0);
    assertEquals(expectedAddress, address.getTextContent().trim());
    assertNoElementChild(address);

    List<Element> serviceNameChildren = XmlUtil.elements(endpointReference,
        XmlConstants.ADDRESSING_REF_NAMESPACE,
        XmlConstants.SERVICE_NAME_TAG_NAME);
    assertEquals(1, serviceNameChildren.size());
    Element serviceName = (Element) serviceNameChildren.get(0);
    String serviceNameContent = serviceName.getTextContent().trim();
    String[] tab = serviceNameContent.split(":");
    QName foundService = new QName(BpelUtil
        .getNamespaceURI(serviceName, tab[0]), tab[1]);
    assertEquals(expectedService, foundService);
    assertNoElementChild(serviceName);
  }

  protected void assertExists(URL url) {
    assertExists("Error: " + url + " does not exist", url);
  }

  protected void assertExists(String failMsg, URL url) {
    try {
      url.openStream();
    } catch (Exception e) {
      fail(failMsg);
    }
  }

  public String getProcessNamespace() {
    return processQName.getNamespaceURI();
  }

  public String getProcessName() {
    return processQName.getLocalPart();
  }

  public QName getProcessQName() {
    return this.processQName;
  }
}
