/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.test.perf.marketplace;

import java.util.HashMap;

import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.ow2.orchestra.exception.OrchestraException;
import org.ow2.orchestra.test.perf.AbstractPerfTestCase;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.var.Message;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard
 */
public class MarketplaceTest extends AbstractPerfTestCase {

  public MarketplaceTest() {
    super("http://orchestra.ow2.org/marketplace", "marketplace");
  }

  public void deploy() {
    deploy(this.getClass().getResource(getProcessName() + ".bpel"), this
        .getClass().getResource(getProcessName() + ".wsdl"));
  }

  public long launch() {
    String item = "item" + Thread.currentThread().getName();
    int buyerPrice = BpelUtil.getRandomInt(0, 1000);
    int sellerPrice = BpelUtil.getRandomInt(-45, 1000);
    BuyerThread buyerThread = new BuyerThread(item, buyerPrice);
    SellerThread sellerThread = new SellerThread(item, sellerPrice);
    buyerThread.start();
    sellerThread.start();

    try {
      buyerThread.join();
      sellerThread.join();
    } catch (Exception e) {
      throw new OrchestraException(
          "Exception while joining buyer and seller threads : "
              + e.getMessage(), e);
    }

    long endTime = System.currentTimeMillis();

    Environment environment = null;
    try {
      environment = getEnvironmentFactory().openEnvironment();
      CallResult buyerResult = buyerThread.callResult;
      CallResult sellerResult = sellerThread.callResult;
      Message buyerReply = buyerResult.getMessageCarrier().getMessage();
      Message sellerReply = sellerResult.getMessageCarrier().getMessage();

      if (buyerReply == null) {
        throw new OrchestraException("buyerReply is null");
      }
      if (sellerReply == null) {
        throw new OrchestraException("sellerReply is null");
      }
      Element buyerElement = buyerReply.getPartValue("outcome");
      if (buyerElement == null) {
        throw new OrchestraException("buyerReply.outcome is null");
      }
      Element sellerElement = sellerReply.getPartValue("outcome");
      if (sellerElement == null) {
        throw new OrchestraException("sellerReply.outcome is null");
      }
      String buyerOutcome = buyerElement.getTextContent();
      if (buyerOutcome == null) {
        throw new OrchestraException("buyerReply.outcome.textContent is null");
      }
      String sellerOutcome = sellerElement.getTextContent();
      if (sellerOutcome == null) {
        throw new OrchestraException("sellerReply.outcome.textContent is null");
      }
      chekOutcomes(buyerPrice, buyerOutcome, sellerPrice, sellerOutcome);
      deleteInstance(buyerResult);
    } finally {
      if (environment != null) {
        environment.close();
      }
    }
    return endTime;
  }

  public static void chekOutcomes(int buyerPrice, String buyerOutcome,
      int sellerPrice, String sellerOutcome) {
    String success = "Deal Successful";
    String fail = "Deal Failed";
    if (buyerPrice >= sellerPrice) {
      if (!buyerOutcome.equals(success)) {
        throw new OrchestraException(
            "Buyer price > Seller price and buyer outcome is not '" + success
                + "', buyerOutcome='" + buyerOutcome + "', sellerOutCome='"
                + sellerOutcome + "'" + ", buyerPrice=" + buyerPrice
                + ", sellerPrice=" + sellerPrice);
      }
      if (!sellerOutcome.equals(success)) {
        throw new OrchestraException(
            "Buyer price > Seller price and seller outcome is not '" + success
                + "', buyerOutcome='" + buyerOutcome + "', sellerOutCome='"
                + sellerOutcome + "'" + ", buyerPrice=" + buyerPrice
                + ", sellerPrice=" + sellerPrice);
      }
    } else {
      if (!buyerOutcome.equals(fail)) {
        throw new OrchestraException(
            "Buyer price < Seller price and buyer outcome is not '" + fail
                + "', buyerOutcome='" + buyerOutcome + "', sellerOutCome='"
                + sellerOutcome + "'" + ", buyerPrice=" + buyerPrice
                + ", sellerPrice=" + sellerPrice);
      }
      if (!sellerOutcome.equals(fail)) {
        throw new OrchestraException(
            "Buyer price < Seller price and seller outcome is not '" + fail
                + "', buyerOutcome='" + buyerOutcome + "', sellerOutCome='"
                + sellerOutcome + "'" + ", buyerPrice=" + buyerPrice
                + ", sellerPrice=" + sellerPrice);
      }
    }
  }

  public abstract class MPThread extends java.lang.Thread {

    protected String item;

    protected int price;

    protected CallResult callResult;

    public MPThread(String item, int price) {
      super();
      this.item = item;
      this.price = price;
    }

    public abstract HashMap<String, Element> getParts();

    public abstract String getOperationName();

    public abstract QName getPortTypeQName();

    public void run() {
      this.callResult = call(getParts(), getPortTypeQName(), getOperationName());
    }
  }

  public class BuyerThread extends MPThread {

    public BuyerThread(String item, int price) {
      super(item, price);
    }

    public HashMap<String, Element> getParts() {
      HashMap<String, Element> parts = new HashMap<String, Element>();
      parts.put("item", BpelXmlUtil.createElementWithContent(this.item));
      parts.put("offer", BpelXmlUtil.createElementWithContent(Integer
          .toString(this.price)));
      return parts;
    }

    public String getOperationName() {
      return "buyer";
    }

    public QName getPortTypeQName() {
      return new QName(getProcessNamespace(), "buyerPT");
    }
  }

  public class SellerThread extends MPThread {

    public SellerThread(String item, int price) {
      super(item, price);
    }

    public HashMap<String, Element> getParts() {
      HashMap<String, Element> parts = new HashMap<String, Element>();
      parts.put("inventoryItem", BpelXmlUtil
          .createElementWithContent(this.item));
      parts.put("askingPrice", BpelXmlUtil.createElementWithContent(Integer
          .toString(this.price)));
      return parts;
    }

    public String getOperationName() {
      return "seller";
    }

    public QName getPortTypeQName() {
      return new QName(getProcessNamespace(), "sellerPT");
    }
  }

  public void testMarketplace() {
    deploy();
    launch();
    undeploy();
  }

  public String getAlias() {
    return "class.market";
  }

}
